-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- TODO BFA:
-- - profiling -- add a list of functions, get the time for each function at the end of frame, print the top functions
-- - add health and mana bars so we can see them without turning on personal buffs?
-- X need to figure out how to fix the unhide in combat issue
--   - instead of hide and unhide we just have to enable / disable mouse
--   - This fixes issues with hidden bars usurping mouse input, but doesn't fix issues with buttons locked out during combat.
-- - incoming damage can be used to predict when the player will die and put up a LoH or DS icon.
-- - if we have any usable trinkets they can be added to the skill or buff bars
-- X show cooldown on rotation
-- - consider selecting rotation skills less frequently than every frame so things feel more stable.
-- - Ret
--   - don't show blessings in a raid or party unless we can check if anyone in the party has our buff
--   - change inquisition timer to a bar?
--   - Should Wake of Ashes show if you will only waste 1 hopo?
--     - if Avenging Wrath / Crusade is on, yes
--     X if Wake is available and we have a choice between skills that will take us to 3 HoPo or 3+ HoPo,
--       choose the skill that takes us to exactly 3 HoPo
--   - Show Inquisition on cooldowns.
--   X test Hammer of Wrath in rotation.
--     X BUG: Hammer of Wrath showing up when target has > 20% health.
--   X Add Shield of Vengeance to cooldowns.
--   - Add Flash of Light to cooldowns (with Selfless Healer)
--   X BUG: when Inquisition is not a selected talent, inquisition timer always shows
--   X Word of Glory support
--   X Justicar's Vengeance support
--   x Hammer of Wrath when Crusade / Avenging Wrath is active
--   - Inquisition timer text sometimes flickers to full alpha
-- - Protection:
--   X have yet to try pro with this addon in BFA
-- - notify when cleanse is necessary (non-raid only)
-- X BUG: when loading the addon as Prot, Buff 1 frame text shows.
-- X BUG: Wake of Ashes and Blade of Justice aren't showing in rotation
--   - bad variable name used when added Prot range checking... Lua honey-badgers unknown variables.
-- - disable addon when in a vehicle or controlling another charcter
-- X BUG: showing cooldowns frame toggle didn't save. Misspelled HoJ._showCooldownsFrame when writing saved variables
-- TODO WoD:
-- * instead of change skills and buffs, change to rotation and skills
--   - rotation would show current best skill and next available skill,
--     including situational skills like Rebuke and LoH.
--   - skills would show currently available skills, and skills on CD
-- X Show Hammer of Wrath on skills bar ... geez.
-- X Show Divine Storm and TV/FV on skills bar
-- X Sort skills not on CD by priority instead of cooldown
-- * Supplication!
-- X Show Divine Storm when Divine Purpose procs and Empowered Divine Storm procs
-- X Show Templar's Verdict / Final Verdict when Divine Purpose procs
--
-- TODO Mists:
-- * show hammer of wrath in skills bar
-- * show Divine Crusader in skills bar
-- * option for show global cooldowns on skills
-- X countdown timer on inquisition
-- * separate buffs window in to a buffs and skills window
-- * rebuke icon when target is casting an interruptible spell
-- TODO Cataclysm:
-- X inquisition timer
-- X dim when mounted
-- TODO:
-- * warn when Ret has Righteous Fury on in a raid or instance. Righteous Fury is
--   useful in some PvE situations like holding aggro on Kilnmasters on Timeless Isle.
-- * if the player's target doesn't have aggro, show the icon for Hand of Reckoning
--   - this must be configurable and shouldn't default to on in instance and raids
-- * customize interface more
--   - allow icons to be clickable!
--   - allow several different setups for buttons or make the button windows movable
--     - just make a specific window for each button?  this would speed up checking
--       for which buffs are being shown.
-- * add hammer of wrath icon when target is below 20% health and HoW is not on cooldown
-- * allow user to choose the time before an icon is off cooldown for it to appear
--   - i.e. show Hammer of Wrath 0.25 seconds before the cooldown is up
--   - Cooldown Early Warning
-- X add Art of War
-- * add Divine Plea
-- * when in combat, check that the paladin has his preferred buffs such as Blessing
--   of Might, and a seal.
-- * show Flash Heal and Excorcism when Art of War is up.
--   - it's only necessary to show these if they're clickable, otherwise a single
--     art of war icon lets us know we can cast either.
--     - problem with this is that exorcism could be on cooldown
-- * option to show Turn Evil
-- * option to show Holy Wrath
-- * show Righteous Fury if ret points in Improved Righteous Fury
--   - for ret this should turn off in raids and instances
-- * trinkets
-- * show Hand of Freedom when movement impaired.
-- * generalize...
--   - choose a skill
--   - choose rules for when to show the skill
--     - mounted / not mounted
--     - in combat
--     - hostile target
--     - another buff is active
--     - choose early warning or not
--     - target health is below some %
-- -----------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- CircularBuffer class
-- If I could figure out the assinine requirements for loading multiple modules
-- in WoW's Lua I wouldn't need to add this at the beginning of this file.
--------------------------------------------------------------------------------

-- Meta class
-- head and tail always increase with the exception being when they reach math.maxinteger
-- rather than wrapping each time head or tail exceeds the max size of the buffer.
-- by simply increasing these and only wrapping at maxinteger we simplify some of the logic:
-- - inserts always happen at head % size
-- - empty is only when tail == head (no modulus necessary)
-- - full is when:
--   - head - tail >= size
-- - we don't need to store a count because it's easily derivable from head - tail
-- - when we reach maxInteger, we just subtract maxInteger from both head and tail, no
--   wrapping. This works even if tail is negative because in Lua the % operator is
--   actual modulo, not just remainder and will always be positive.
local CircularBuffer = { }
CircularBuffer.__index = CircularBuffer
CircularBuffer.maxInteger = bit.lshift( 1, 31 )

-- Derived class method new
function CircularBuffer:print( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text );
	end
end

function CircularBuffer:new( maxSize_ )
	-- disallow trailing nil parameters
	if ( maxSize_ == nil ) then
		self:print( "maxSize_ == nil!")
		return nil
	end

	local b = setmetatable( { head = 0, tail = 0, maxSize = maxSize_ }, CircularBuffer )
	return b
end

function CircularBuffer:incIndex( index )
	-- only wrap if we are at maximum integer value already
	if ( index == self.maxInteger ) then
		index = 0
	else
		index = index + 1
	end
	return index
end

function CircularBuffer:boolToStr( value )
	if ( not value ) then
		return "false"
	else
		return "true"
	end
end

function CircularBuffer:nilToStr( value )
	if ( value == nil ) then
		return "nil"
	else
		return value
	end
end

function CircularBuffer:count()
	return self.head - self.tail
end

function CircularBuffer:isFull()
	return self:count() >= self.maxSize
end

function CircularBuffer:isEmpty()
	return self.head - self.tail == 0
end

function CircularBuffer:add( val )
	if ( self:isFull() ) then
		return false
	end

	self[self.head % self.maxSize] = val

	if ( self.head == self.maxInteger ) then
		self.head = self.head - self.maxInteger
		self.tail = self.tail - self.maxInteger
	end

	self.head = self.head + 1
	return true
end

function CircularBuffer:remove()
	if ( self:isEmpty() ) then
		return nil
	end

	local index = self.tail % self.maxSize
	local value = self[index]
	self.tail = self.tail + 1
	return value
end

function CircularBuffer:empty()
	local count = self:count()
	for i = 0, count do
		local index = ( self.tail + i ) % self.maxSize
		self[index] = nil
	end

	self.head = 0
	self.tail = 0
end

-- because WoW doesn't support "require"
-- return CircularBuffer

-- ============================================================================
-- Profiling

local ProfileTable = { }
local ProfileFrameNum = 0

-- ----------------------------------------------------------------------------
-- ProfileBegin
-- ----------------------------------------------------------------------------

function ProfileBegin( name_, key )
	if ( ProfileTable[key] == nil ) then
		ProfileTable[key] = { name = name_, startMs = 0.0, endMs = 0.0, timeMs = 0.0, count = 0 }
	end
	ProfileTable[key].startMs = debugprofilestop()
end

function ProfileBegin2( name )
	ProfileBegin( name, name )
end

-- ----------------------------------------------------------------------------
-- ProfileEnd
-- ----------------------------------------------------------------------------
function ProfileEnd( key, startTime, spam )
	local endMs = debugprofilestop()

	local entry = ProfileTable[key]
	entry.endMs = endMs
	entry.timeMs = entry.endMs - entry.startMs
	entry.count = entry.count + 1
	entry.frameNum = ProfileFrameNum
	if ( spam ~= nil ) then
		DEFAULT_CHAT_FRAME:AddMessage( "ProfileEnd( " .. ProfileFrameNum .. " ): " .. key .. " = " .. entry.timeMs .. " ms" );
	end
end

-- ----------------------------------------------------------------------------
-- ProfileFunction
-- ----------------------------------------------------------------------------

function ProfileFunction( name, func, ... )
	local startTime = ProfileBegin( name, name )
	func( ... )
	ProfileEnd( name, startTime )
end

-- ----------------------------------------------------------------------------
-- ProfileFunction
-- ----------------------------------------------------------------------------
function ProfileClear( name )
	local entry = ProfileTable[name]
	entry.timeMs = 0
end

-- ============================================================================

HoJ = { }

-- HoJ.ADDON_VERSION = 0.5; -- split hopo window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.6; -- split icon window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.81; -- major rewrite for Mists of Pandaria with low level pally fixes
-- HoJ.ADDON_VERSION = 0.82; -- ret skill bar for rebuke, blinding light, hammer of justice / fist of justice, added selfless healer to buffs, added stack count to buffs and skills. Did not implement skills for Prot yet.
-- HoJ.ADDON_VERSION = 0.9; -- Warlord's of Draenor. Removed skills that no longer exists (inquisition, blinding light)
-- HoJ.ADDON_VERSION = 0.91; -- Warlord's of Draenor, 6.1 skill priority adjustments.
-- HoJ.ADDON_VERSION = 0.92; -- Warlord's of Draenor, 6.2 priorities from Summonstone
HoJ.MAJOR_VERSION = 0 -- Battle for Azeroth, 8.1 Tides of Vengeance
-- HoJ.MINOR_VERSION = 93;
-- HoJ.BUILD_VERSION = 1; -- fixed one of the cooldown button timer texts not being set
-- HoJ.BUILD_VERSION = 2; -- fixed a bug where the crusader strike icon was incorrect if the add-on loaded while in prot spec.
-- HoJ.BUILD_VERSION = 3; -- resetting spell icons was resetting empyrean power to not show as divine storm
-- HoJ.BUILD_VERSION = 4; -- removed debug outputs, first beta
-- HoJ.BUILD_VERSION = 5; -- hammer of wrath fixes, added Shield of Vengeance
-- HoJ.BUILD_VERSION = 6; -- selfless healer cooldown
-- HoJ.BUILD_VERSION = 7; -- BUG: sometimes a rotation skill is shown when everything is on cooldown
-- HoJ.BUILD_TYPE = "beta";
HoJ.MINOR_VERSION = 95
--HoJ.BUILD_VERSION = 0; -- Very basic Protection rotation
-- HoJ.BUILD_VERSION = 1; -- Very basic Protection mitigation cooldowns
-- HoJ.BUILD_VERSION = 2; -- Amortized update logic to smooth out cost per frame.
-- HoJ.BUILD_VERSION = 3; -- renamed some frames, added cooldowns
-- HoJ.BUILD_VERSION = 4; -- added charge counter to rotation cooldown, saved variable versioning
-- HoJ.BUILD_VERSION = 5; -- fixed "predictAhead == nil!" spam during ret mode. That's why this is still in beta...
-- HoJ.BUILD_VERSION = 6; -- added data-driven rotation selection for Retribution
-- HoJ.BUILD_VERSION = 7; -- optimizations
-- HoJ.BUILD_VERSION = 8; -- fixed buff timer caching / inquisition never going away if talented.
-- HoJ.BUILD_VERSION = 9; -- Avenger's Shield will show up as a interrupt skill on the cooldowns list if Rebuke is on CD or out of range
-- HoJ.BUILD_VERSION = 10; -- Hammer of Justice should now show up correctly on cooldowns bar when Fist of Justice is talented
-- HoJ.BUILD_VERSION = 11; -- Fixed 10 yard range being calculated using Fist of Justice when FoJ talent was selected
-- HoJ.BUILD_VERSION = 1 -- 0.95.1 - Added frame alpha configuration to addon menu
-- HoJ.BUILD_VERSION = 2 -- 0.95.2 - tracking of blessing buffs on party members and (as much as possible) on non-party members
HoJ.BUILD_VERSION = 3 -- 0.95.3 - cleaned up an accidental nested loop in buff check, fixed issue with max party member being out of range

HoJ.BUILD_TYPE = "beta"

HoJ._savedVersionMajor = HoJ.MAJOR_VERSION
HoJ._savedVersionMinor = HoJ.MINOR_VERSION
HoJ._savedVersionBuild = HoJ.BUILD_VERSION

HoJ.SPELL_BUFF = 1			-- buffs a unit
HoJ.SPELL_SHAPESHIFT = 2	-- buffs self
HoJ.SPELL_SKILL = 3			-- not a buff, usually part of rotation
HoJ.SPELL_TALENT = 4		-- not a buff, usually part of rotation, but acquired via talents
HoJ.SPELL_TALENT_BUFF = 5	-- a buff that is a spell and talented (selfless healer)

HoJ.RANGE_NONE = 0
HoJ.RANGE_SELF = 1
HoJ.RANGE_MELEE = 2
HoJ.RANGE_10_YARDS = 3
HoJ.RANGE_12_YARDS = 4
HoJ.RANGE_20_YARDS = 5
HoJ.RANGE_30_YARDS = 6
HoJ.RANGE_40_YARDS = 7

HoJ.ROLE_HEALING = 1
HoJ.ROLE_TANK = 2
HoJ.ROLE_DPS = 3
HoJ.ROLE_UNKNOWN = 4
HoJ.ROLE_ANY = 5 		-- for talents usable by any role

--[[
The world state is used to prevents weird stuff from happening between the
time when the player leaves the world and returns (for instance, taking a portal).
I have seen the following happen:
- logged in on a character level 117
- used party sync and leveled to 120
- never logged out
- when taking a portal, on PLAYER_LEAVING_WORLD, UnitLevel( "player" ) returns 117
- after entering the world again, on PLAYER_ENTERING_WORLD, UnitLevel( "player" ) returns 120
- This was messing with blessing initialization.
--]]

HoJ.WORLD_STATE_UNKNOWN = 0
HoJ.WORLD_STATE_ENTERED = 1
HoJ.WORLD_STATE_LEFT = 2
HoJ._worldState = HoJ.WORLD_STATE_UNKNOWN

HoJ.REACTION_THRESHOLD = 5 -- 4 will only show skill bar for red mobs, 5 will show skill bar for yellow mobs

HoJ.SKILL_USE_PREDICTION_TIME = 0.01 -- look this far beyond the remaining GCD to predict a skill if we don't have one now

HoJ.ROLE_NAMES = { "HEALING", "TANK", "DPS", "UNKNOWN" }
HoJ.STR_TAB = "	"
HoJ.STR_WHITE = "|cffffffff"
HoJ.STR_YELLOW = "|cffffff00"
HoJ.STR_GREEN = "|cff00ff00"
HoJ.STR_PURPLE = "|cffff00ff"
HoJ.STR_RED = "|cffff0000"
HoJ.STR_ORANGE = "|cffff7f00"
HoJ.STR_LIGHTGRAY = "|cffcfcfcf"
HoJ.STR_GRAY = "|cff7f7f7f"
HoJ.STR_WHISPER = "|cffff7fff"
HoJ.STR_CHANNEL = "|cffffc0c0"

HoJ._needsReset = true		-- resets occur on role switches

HoJ.GROUP_NONE = 0
HoJ.GROUP_PARTY = 1
HoJ.GROUP_RAID = 2
HoJ._groupType = HoJ.GROUP_NONE
HoJ._numGroupMembers = 1

HoJ._enabledForRole = true

HoJ._role = HoJ.ROLE_UNKNOWN

HoJ.ADDON_NAME = HoJ.STR_YELLOW .. "Hammer of Justas" .. HoJ.STR_WHITE
HoJ.MSG_PREFIX = HoJ.STR_YELLOW .. "Hammer of Justas: " .. HoJ.STR_WHITE
HoJ.UPDATES_PER_SEC = 15
HoJ.RENDERS_PER_SEC = 60
HoJ.UPDATE_INTERVAL = 1.0 / HoJ.UPDATES_PER_SEC
HoJ.RENDER_INTERVAL = 1.0 / HoJ.RENDERS_PER_SEC

-- blessing update amortization
HoJ._partyUpdateIndex = -1 	-- amortizes party buff checks by cycling through party members
HoJ._nextUpdateBlessingFrame = 0
HoJ._updateBlessingsInterval = 1 --HoJ.UPDATES_PER_SEC * 5 -- after finishing a party update, wait this number of extra frames before starting the next

-- don't update everything on the same frame
HoJ.UPDATE_START = 1
HoJ.UPDATE_ROTATION = HoJ.UPDATE_START
HoJ.UPDATE_COOLDOWNS = 2
HoJ.UPDATE_BUFFS = 3
HoJ.UPDATE_MAX = HoJ.UPDATE_BUFFS
HoJ._updateType = HoJ.UPDATE_START

HoJ._nextUpdateTime = 0.0
HoJ._nextRenderTime = 0.0

HoJ._isPaladin = false
HoJ._inCombat = false
HoJ._level = 0
HoJ._holyPower = 0
HoJ._fakeHolyPower = 0 -- used to account for Divine Purpose giving us infinite HoPo
HoJ.MAX_HOLY_POWER = 5
HoJ._useHolyPower = false 	-- true when a holy power skill should be used
HoJ._playerName = nil
HoJ._playerRaceId = 0

HoJ._playerHealth = 1
HoJ._playerMaxHealth = 1
HoJ._playerHealthRatio = 1.0
HoJ._predictedHealth = 1
HoJ._predictedHealthTime = 0.0
HoJ._predictedDeathTime = 0.0
HoJ._predictedNearDeathTime = 0.0

HoJ._damagePerSec = 0
HoJ._healsPerSec = 0
HoJ._damagePerSecRatio = 0.0
HoJ._healsPerSecRatio = 0.0
HoJ._largestRecentDamageRatio = 0.0

HoJ._nextHealthTime = 0.0
HoJ._showHealthTime = 0.0	-- set to a time in the future to see health spam

HoJ.HEALTH_SAMPLE_SECONDS = 3 -- number of seconds we keep health history for
HoJ.HEALTH_SAMPLES_PER_SEC = 15	-- number of times per second that we sample health (make sure it's less than HoJ.UPDATE_INTERVAL)

HoJ._frameNum = 0 -- updated for every call to HoJ_OnUpdate
HoJ._updateFrameNum = 0 -- updated for every call to HoJ:Update
HoJ._renderFrameNum = 0 -- updated for every call to HoJ:UpdateRender
HoJ._enteredWorldUpdateFrameNum = 0 -- frame when player entered world, used to gate some udpates after entering world until some things settle
HoJ._updateElapsedTime = 0.0
HoJ._renderElapsedTime = 0.0
HoJ._updateBuffsTime = 0.0 -- when to next check buffs
HoJ._updateCooldownsTime = 0.0 -- when to next check cooldowns
--HoJ._divinePleaTriggerPercentage = 0.75;	-- trigger when we have less than this % of max mana
HoJ._blockedNames = { }	--["testname"] = "testname" };
--HoJ._blockedNames = { ["Mienyu"] = "Mienyu", ["Justas"] = "Justas" };
HoJ._ignoreHistory = { }
HoJ._silentIgnore = false
HoJ._stealthIgnore = true
HoJ._stealthIgnoreMessage = HoJ.STR_CHANNEL .. "[3. WorldDefense] : " .. HoJ.STR_YELLOW .. "Isle of Quel'Danas is under attack!"
HoJ._maxGlowAlpha = 0.5
HoJ._minGlowAlpha = 0.1
HoJ._glowAlpha = HoJ._minGlowAlpha
HoJ._glowScale = 1.025
HoJ._skillCDAlpha = 0.25;	-- alpha for spell icons on cooldown
HoJ._skillNormalAlpha = 0.75	-- alpha for skill icons not on cooldown
HoJ._events = { }
HoJ._HoPo = { nil, nil, nil, nil, nil }
HoJ._HoPoGlow = { nil, nil, nil, nil, nil }
HoJ._curTime = GetTime()
HoJ._gcdRemaining = 0.0
HoJ._gcdDuration = 0.0 -- we cache this because we can't just query it when a spell isn't active
HoJ._minHolyPowerForInquisition = 2
-- WODONLY
-- HoJ._sacredShieldWarningTime = 2

HoJ._targetExists = false
HoJ._targetCanAttack = false
HoJ._targetCanBeInterrupted = false
HoJ._targetHealth = 0.0
HoJ._targetHealthRatio = 0.0
HoJ._targetLevel = 1
HoJ._targetClassification = ""
HoJ._targetIsTough = false
HoJ._targetIsBoss = false -- "worldboss" or "rareelite"

HoJ._targetInRange = { }	-- one entry for each HoJ.RANGE_ type, indicating if the target is within that range

HoJ._showSkillCDs = true
HoJ._showCooldownsFrame = true
HoJ._showBuffsFrame = true
HoJ._showTemplarsVerdictFrame = true
HoJ._showHolyPowerFrame = true
HoJ._showInquisitionFrame = true
HoJ._showRotationFrame = true

HoJ._enableProfiling = false
HoJ._profileThreshold = 0.3 -- threshold in ms before showing CPU time
HoJ._profileFunction = nil

HoJ._enabled = true;

HoJ.INCOMING_DPS_CHECK_INTERVAL = 2.0
HoJ.AVENGING_WRATH_INCOMING_DPS_RATIO = -0.10	-- incoming DPS less than this ratio will trigger avenging wrath icon
HoJ.AVENGING_WRATH_HEALTH_THRESHOLD = 0.5
HoJ.PROT_GUARDIAN_INCOMING_DPS_RATIO = -0.2 -- incoming DPS less than this will trigger guardian of ancient kings icon

HoJ._buffs = { }	-- list of currently active buffs
HoJ._buffFrames = { }

HoJ._cooldowns = { }	-- list of skills to execute in order of priority / cooldown
HoJ._cooldownFrames = { }

HoJ._rotationSkills = { } -- list of best skills (only ever 1)
HoJ._rotationFrames = { }

HoJ._windowFrames = { }  -- list of all HoJ-specific window frames
HoJ._omniCC = false -- true if omniCC is enabled (it will show it's own timer)

HoJ._globalAlpha = 1.0 -- global alpha for HoJ UI frames
HoJ._rotationAlpha = 1.0
HoJ._cooldownsAlpha = 1.0
HoJ._buffsAlpha = 1.0
HoJ._InquisitionAlpha = 1.0

HoJ._debugSpellName = nil -- set to a spell name, e.g. "Wake of Ashes", to debug that spell

local HOJ_INQ_EXPIRED		= 0	-- in combat without inquisition on!
local HOJ_INQ_NEARLY_EXPIRED= 1	-- in combat but inquisition is expiring
local HOJ_INQ_EXPIRING 		= 2	-- in combat but inquisition is expiring
local HOJ_INQ_ON 			= 3	-- in combat and inquisition is on
local HOJ_INQ_NOT_IN_COMBAT	= 4	-- not in combat, so ignore inquisition
HoJ._inqState = HOJ_INQ_NOT_IN_COMBAT

local SCROLLBAR_WIDTH = 16
local SCROLLBAR_HEIGHT = 12
local EDGE_SIZE = 8

HoJ._throttledMessages = { }	-- list of throttled messages

-- -----------------------------------------------------------------------------
-- HoJ:GetVersionString
-- -----------------------------------------------------------------------------
function HoJ:GetVersionString()
	local version = "" .. HoJ.MAJOR_VERSION .. "." .. HoJ.MINOR_VERSION .. "." .. HoJ.BUILD_VERSION .. " " .. HoJ.BUILD_TYPE
	return version
end

-- -----------------------------------------------------------------------------
-- HoJ:DebugMsg
-- -----------------------------------------------------------------------------
function HoJ:DebugMsg( text )
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ThrottleMsg
-- -----------------------------------------------------------------------------
function HoJ:ThrottleMsg( interval, text )
	if ( DEFAULT_CHAT_FRAME ) then
--[[
		local msg = HoJ._throttledMessages[text]
		if ( msg == nil ) then
			msg = {
				lastTime = self._curTime
			}
			HoJ._throttledMessages[text] = msg
		else
			local timeSincePrinted = self._curTime - msg.lastTime
			if ( timeSincePrinted < interval ) then
				return
			end
		end

		msg.lastTime = self._curTime
--]]
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:FrameThrottleMsg
-- -----------------------------------------------------------------------------
function HoJ:FrameThrottleMsg( numFrames, text )
	if ( DEFAULT_CHAT_FRAME ) then
--[[
		local msg = HoJ._throttledMessages[text]
		if ( msg == nil ) then
			msg = {
				lastTime = self._curTime
			}
			HoJ._throttledMessages[text] = msg
		else
			local timeSincePrinted = self._curTime - msg.lastTime
			if ( timeSincePrinted < interval ) then
				return
			end
		end

		msg.lastTime = self._curTime
--]]
		if ( ( HoJ._frameNum % numFrames ) ~= 0 ) then
			return
		end
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Warn
-- -----------------------------------------------------------------------------
function HoJ:Warn( text )
	-- should be commented out for releases
	if ( DEFAULT_CHAT_FRAME ) then
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Print
-- -----------------------------------------------------------------------------
function HoJ:Print( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Msg
-- -----------------------------------------------------------------------------
function HoJ:Msg( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.MSG_PREFIX .. text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellName
-- -----------------------------------------------------------------------------
function HoJ:GetSpellName( spellID )
	name = GetSpellInfo( spellID );
	if ( name == nil ) then
		HoJ:DebugMsg( "Got nil spell name for id " .. spellID )
	end
	return name;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellRank
-- -----------------------------------------------------------------------------
function HoJ:GetSpellRank( spellID )
	_, rank = GetSpellInfo( spellID )
	return rank;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIcon
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIcon( spellID )
	_, _, icon = GetSpellInfo( spellID )
	return icon;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIndex
-- this tries to find the index of a spell in the spellbook, which is not the
-- same as the spell ID of a spell. For some reason the spell index is needed
-- to find if a target is in range of a spell, rather than the spell ID.
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIndex( spellInfo )
	-- tabs are the side tabs which are currently "General", "Retribution", "Holy", and "Protection"
	-- there is no way to know how many spells are on a single tab, but it won't be over 48 for paladins in 8.1
	local numSpellTabs = GetNumSpellTabs()
	--HoJ:DebugMsg( "Found " .. numSpellTabs .. " spellbook tabs" );
	local max = numSpellTabs * 96
	for i = 1, max do
		local spellName, spellSubName = GetSpellBookItemName( i, BOOKTYPE_SPELL )
		if ( spellName == nil ) then
			--HoJ:DebugMsg( "No spellbook info found for index " .. i );
			return -1;
	   	else
			-- use spellName and spellSubName here
			-- HoJ:DebugMsg( "Spell #" .. i .. ": '" .. spellName .. "'" )
			if ( spellInfo.name == spellName ) then
				--HoJ:DebugMsg( "Found spellbook index " .. i .. " for spell id " .. spellInfo.name );
				return i
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_InitSpellInfo
-- -----------------------------------------------------------------------------
function HoJ_InitSpellInfo()
	-- HoJ:DebugMsg( "InitSpellInfo" );
	for key, value in pairs( HoJ.SpellInfo ) do
		-- HoJ:DebugMsg( value.name );
		local name, rank, icon = GetSpellInfo( value.id )
		-- some spell infos use an alternate icon id -- for instance a buff that allows another
		-- spell to be cast (Empyrean Power), we want to show the spell that can be cast on the
		-- rotation frame, not the buff.
		if ( value.iconId ~= nil ) then
			-- HoJ:DebugMsg( "Getting iconId " .. value.iconId .. " for " .. value.name );
			_, _, icon = GetSpellInfo( value.iconId )
		end
		-- For some reason, if we query Crusader Strike while prot, we get a weird texture id
		-- so when we switch roles we go ahead and re-init all of the spell infos.
		-- HoJ:DebugMsg( "  id = " .. value.id );
		-- HoJ:DebugMsg( "  old icon = " .. value.icon .. ", new icon = " .. icon );
		value.name = name
		value.rank = rank
		value.icon = icon
		value.spellIndex = HoJ:GetSpellIndex( value )
	end
end

-- tracks which player the blessing was cast on, and keeps track of when it will time out
HoJ._blessingInfo = {
	[203538] = {
		name = "greaterBlessingOfKings",
		targetName = nil,
		expireTime = -1
	},
	[203539] = {
		name = "greaterBlessingOfWisdom",
		targetName = nil,
		expireTime = -1
	}
}

HoJ.SpellInfo = {
	["divinePurpose"] = {
		id = 223817,
		name = HoJ:GetSpellName( 223817 ),
		rank = HoJ:GetSpellRank( 223817 ),
		icon = HoJ:GetSpellIcon( 223817 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = true,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 7,
		talentCol = 1,
		talentRole = HoJ.ROLE_DPS
	},
	["divinePurposeBuff"] = {
		id = 223819,
		name = HoJ:GetSpellName( 223819 ),
		rank = HoJ:GetSpellRank( 223819 ),
		icon = HoJ:GetSpellIcon( 223819 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["inquisition"] = {
		id = 84963,
		name = HoJ:GetSpellName( 84963 ),
		rank = HoJ:GetSpellRank( 84963 ),
		icon = HoJ:GetSpellIcon( 84963 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT_BUFF,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 7,
		talentCol = 3,
		talentRole = HoJ.ROLE_DPS
	},
	["avengingWrath"] = {
		id = 31884,
		name = HoJ:GetSpellName( 31884 ),
		rank = HoJ:GetSpellRank( 31884 ),
		icon = HoJ:GetSpellIcon( 31884 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = true,
		text = nil,
		minLevel = 80,
		cooldownRemaining = 0,
		priority = 2000,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["crusade"] = {
		id = 231895,
		name = HoJ:GetSpellName( 231895 ),
		rank = HoJ:GetSpellRank( 231895 ),
		icon = HoJ:GetSpellIcon( 231895 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = true,
		text = nil,
		minLevel = 80,
		cooldownRemaining = 0,
		priority = 2000,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 7,
		talentCol = 2,
		talentRole = HoJ.ROLE_DPS
	},
	["divineProtection"] = {
		id = 498,
		name = HoJ:GetSpellName( 498 ),
		rank = HoJ:GetSpellRank( 498 ),
		icon = HoJ:GetSpellIcon( 498 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 26,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["ardentDefender"] = {
		id = 31850,
		name = HoJ:GetSpellName( 31850 ),
		rank = HoJ:GetSpellRank( 31850 ),
		icon = HoJ:GetSpellIcon( 31850 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["shieldOfTheRighteous"] = {
		id = 132403,
		name = HoJ:GetSpellName( 132403 ),
		rank = HoJ:GetSpellRank( 132403 ),
		icon = HoJ:GetSpellIcon( 132403 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 20,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["consecration"] = {
		id = 26573,
		name = HoJ:GetSpellName( 26573 ),
		rank = HoJ:GetSpellRank( 26573 ),
		icon = HoJ:GetSpellIcon( 26573 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 14,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 1,
		range = HoJ.RANGE_10_YARDS,
		duration = 12,
		role = HoJ.ROLE_TANK
	},
	-- the buff is apparently different than the spell?
	["protConsecrationBuff"] = {
		id = 188370,
		name = HoJ:GetSpellName( 188370 ),
		rank = HoJ:GetSpellRank( 188370 ),
		icon = HoJ:GetSpellIcon( 188370 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 14,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 1,
		range = HoJ.RANGE_10_YARDS,
		duration = 12,
		role = HoJ.ROLE_TANK
	},
	["retConsecration"] = {
		id = 205228,
		name = HoJ:GetSpellName( 205228 ),
		rank = HoJ:GetSpellRank( 205228 ),
		icon = HoJ:GetSpellIcon( 205228 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT, -- only a talent for ret
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 60,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 1,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 4,
		talentCol = 2,
		duration = 6,
		role = HoJ.ROLE_DPS
	},
	["avengersValor"] = {
		id = 197561,
		name = HoJ:GetSpellName( 197561 ),
		rank = HoJ:GetSpellRank( 197561 ),
		icon = HoJ:GetSpellIcon( 197561 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 60,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SEKF,
		duration = 15,
		role = HoJ.ROLE_TANK
	},
	["guardianOfAncientKings"] = {
		id = 86659,
		name = HoJ:GetSpellName( 86659 ),
		rank = HoJ:GetSpellRank( 86659 ),
		icon = HoJ:GetSpellIcon( 86659 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["divineShield"] = {
		id = 642,
		name = HoJ:GetSpellName( 642 ),
		rank = HoJ:GetSpellRank( 642 ),
		icon = HoJ:GetSpellIcon( 642 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 18,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["righteousFury"] = {
		id = 25780,
		name = HoJ:GetSpellName( 25780 ),
		rank = HoJ:GetSpellRank( 25780 ),
		icon = HoJ:GetSpellIcon( 25780 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 25780 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 12,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
-- WODONLY
--[[
	["sealOfTruth"] = {
		id = 31801,
		name = HoJ:GetSpellName( 31801 ),
		rank = HoJ:GetSpellRank( 31801 ),
		icon = HoJ:GetSpellIcon( 31801 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 31801 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 24,
		cooldownRemaining = 0,
		priority = 0
	},
	["sealOfRighteousness"] = {
		id = 20154,
		name = HoJ:GetSpellName( 20154),
		rank = HoJ:GetSpellRank( 20154 ),
		icon = HoJ:GetSpellIcon( 20154 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20154 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 42,
		cooldownRemaining = 0,
		priority = 0
	},
	["sealOfJustice"] = {
		id = 20164,
		name = HoJ:GetSpellName( 20164),
		rank = HoJ:GetSpellRank( 20164 ),
		icon = HoJ:GetSpellIcon( 20164 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20164 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0
	},
	["sealOfInsight"] = {
		id = 20165,
		name = HoJ:GetSpellName( 20165 ),
		rank = HoJ:GetSpellRank( 20165 ),
		icon = HoJ:GetSpellIcon( 20165 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20165 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 32,
		cooldownRemaining = 0,
		priority = 0
	},
	["blessingOfMight"] = {
		id = 19740,
		name = HoJ:GetSpellName( 19740 ),
		rank = HoJ:GetSpellRank( 19740 ),
		icon = HoJ:GetSpellIcon( 19740 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 19740 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 81,
		cooldownRemaining = 0,
		priority = 0
	},
--]]
--WODONLY
--[[
	["blessingOfKings"] = {
		id = 20217,
		name = HoJ:GetSpellName( 20217 ),
		rank = HoJ:GetSpellRank( 20217 ),
		icon = HoJ:GetSpellIcon( 20217 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0
	},
]]
	["blessingOfKings"] = { -- actually called "greater blessing of kings"
		id = 203538,
		name = HoJ:GetSpellName( 203538 ),
		rank = HoJ:GetSpellRank( 203538 ),
		icon = HoJ:GetSpellIcon( 203538 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		buffDurationSeconds = 3600,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 58,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["blessingOfWisdom"] = { -- actually called "greater blessing of wisdom"
		id = 203539,
		name = HoJ:GetSpellName( 203539 ),
		rank = HoJ:GetSpellRank( 203539 ),
		icon = HoJ:GetSpellIcon( 203539 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		buffDurationSeconds = 3600,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 65,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["blessingOfProtection"] = {
		id = 1022,
		name = HoJ:GetSpellName( 1022 ),
		rank = HoJ:GetSpellRank( 1022 ),
		icon = HoJ:GetSpellIcon( 1022 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["hammerOfWrath"] = {
		id = 24275,
		name = HoJ:GetSpellName( 24275 ),
		rank = HoJ:GetSpellRank( 24275 ),
		icon = HoJ:GetSpellIcon( 24275 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 24275 ),
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 38,
		cooldownRemaining = 0,
		priority = 800,
		hoPoGen = 1,
		range = HoJ.RANGE_30_YARDS,
		talentRow = 2,
		talentCol = 3,
		talentRole = HoJ.ROLE_DPS
	},
	["rebuke"] = {
		id = 96231,
		name = HoJ:GetSpellName( 96231 ),
		rank = HoJ:GetSpellRank( 96231 ),
		icon = HoJ:GetSpellIcon( 96231 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 96231 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 1,
		hoPoGen = 0,
		range = HoJ.RANGE_MELEE
	},
	["crusaderStrike"] = {
		id = 35395,
		name = HoJ:GetSpellName( 35395 ),
		rank = HoJ:GetSpellRank( 35395 ),
		icon = HoJ:GetSpellIcon( 35395 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 600,
		hoPoGen = 1,
		range = HoJ.RANGE_MELEE
	},
	["judgment"] = {
		id = 20271,
		name = HoJ:GetSpellName( 20271 ),
		rank = HoJ:GetSpellRank( 20271 ),
		icon = HoJ:GetSpellIcon( 20271 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 5,
		cooldownRemaining = 0,
		priority = 400,
		hoPoGen = 1,
		range = HoJ.RANGE_30_YARDS
	},
	["judgmentDebuff"] = {
		id = 197277,
		name = HoJ:GetSpellName( 197277 ),
		rank = HoJ:GetSpellRank( 197277 ),
		icon = HoJ:GetSpellIcon( 197277 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 5,
		cooldownRemaining = 0,
		priority = 400,
		hoPoGen = 1,
		range = HoJ.RANGE_30_YARDS
	},
	["templarsVerdict"] = {
		id = 85256,
		name = HoJ:GetSpellName( 85256 ),
		rank = HoJ:GetSpellRank( 85256 ),
		icon = HoJ:GetSpellIcon( 85256 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 85256 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 1400,
		hoPoGen = -3,
		range = HoJ.RANGE_MELEE
	},
	["finalVerdict"] = {
		id = 157048,
		name = HoJ:GetSpellName( 157048 ),
		rank = HoJ:GetSpellRank( 157048 ),
		icon = HoJ:GetSpellIcon( 157048 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 157048 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1400,
		hoPoGen = -3,
		range = HoJ.RANGE_MELEE
	},
	["finalVerdictBuff"] = {	-- this skill is also the buff...
		id = 157048,
		name = HoJ:GetSpellName( 157048 ),
		rank = HoJ:GetSpellRank( 157048 ),
		icon = HoJ:GetSpellIcon( 157048 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 157048 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 100,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
	["divineStorm"] = {
		id = 53385,
		name = HoJ:GetSpellName( 53385 ),
		rank = HoJ:GetSpellRank( 53385 ),
		icon = HoJ:GetSpellIcon( 53385 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 53385 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1300,
		hoPoGen = -3,
		range = HoJ.RANGE_12_YARDS
	},
	["executionSentence"] = {
		id = 163250,
		name = HoJ:GetSpellName( 163250 ),
		rank = HoJ:GetSpellRank( 163250 ),
		icon = HoJ:GetSpellIcon( 163250 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1500,
		hoPoGen = -3,
		range = HoJ.RANGE_20_YARDS,
		talentRow = 1,
		talentCol = 3,
		talentRole = HoJ.ROLE_DPS
	},
	["lightsHammer"] = {
		id = 114158,
		name = HoJ:GetSpellName( 114158 ),
		rank = HoJ:GetSpellRank( 114158 ),
		icon = HoJ:GetSpellIcon( 114158 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 114158 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1500,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["globalCooldown"] = {
		id = 61304, -- must be something on the GCD -- Crusader Strike for now
		name = HoJ:GetSpellName( 61304 ),
		rank = HoJ:GetSpellRank( 61304 ),
		icon = HoJ:GetSpellIcon( 61304 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
	["selflessHealer"] = {
		id = 85804,
		name = HoJ:GetSpellName( 85804 ),
		rank = HoJ:GetSpellRank( 85804 ),
		icon = HoJ:GetSpellIcon( 85804 ),
		spellToCast = 19750, -- cast as flash of light
		spellIndex = -1,  -- HoJ:GetSpellIndex( 85804 ),
		spellType = HoJ.SPELL_TALENT_BUFF,
		isProcd = true,
		showOnBar = true, 	-- set to false so we don't show on the buffs bar by default, because we show on cooldowns due to the special circumstances
		showCheckbox = true,
		showCharges = false,
		showAsBuff = false,
		text = nil,
		minLevel = 0,
		cooldownRemaining = 0,
		priority = 500,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE,
		talentRow = 6,
		talentCol = 1,
		maxBuffCount = 4,
		talentRole = HoJ.ROLE_DPS
	},
	["layOnHands"] = {
		id = 633,
		name = HoJ:GetSpellName( 633 ),
		rank = HoJ:GetSpellRank( 633 ),
		icon = HoJ:GetSpellIcon( 633 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 16,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["lightOfTheProtector"] = {
		id = 184092,
		name = HoJ:GetSpellName( 184092 ),
		rank = HoJ:GetSpellRank( 184092 ),
		icon = HoJ:GetSpellIcon( 184092 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 75,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["handOfTheProtector"] = {
		id = 213652,
		name = HoJ:GetSpellName( 213652 ),
		rank = HoJ:GetSpellRank( 213652 ),
		icon = HoJ:GetSpellIcon( 213652 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 75,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS,
		talentRow = 5,
		talentCol = 3,
		talentRole = HoJ.ROLE_TANK
	},
	["hammerOfJustice"] = {
		id = 853,
		name = HoJ:GetSpellName( 853 ),
		rank = HoJ:GetSpellRank( 853 ),
		icon = HoJ:GetSpellIcon( 853 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 7,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS
	},
	["fistOfJustice"] = {
		id = 234299,
		name = HoJ:GetSpellName( 234299 ),
		rank = HoJ:GetSpellRank( 234299 ),
		icon = HoJ:GetSpellIcon( 234299 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 45,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 3,
		talentCol = 1,
		talentRole = HoJ.ROLE_ANY
	},
	["blindingLight"] = {
		id = 115750,
		name = HoJ:GetSpellName( 115750 ),
		rank = HoJ:GetSpellRank( 115750 ),
		icon = HoJ:GetSpellIcon( 115750 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 45,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 3,
		talentCol = 3,
		talentRole = HoJ.ROLE_ANY
	},
	["flashOfLight"] = {
		id = 19750,
		name = HoJ:GetSpellName( 19750 ),
		rank = HoJ:GetSpellRank( 19750 ),
		icon = HoJ:GetSpellIcon( 19750 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 5,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["giftOfTheNaaru"] = {
		id = 59542,
		name = HoJ:GetSpellName( 59542 ),
		rank = HoJ:GetSpellRank( 59542 ),
		icon = HoJ:GetSpellIcon( 59542 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS,
		requiredRaceId = 11
	},
	["wakeOfAshes"] = {
		id = 255937,
		name = HoJ:GetSpellName( 255937 ),
		rank = HoJ:GetSpellRank( 255937 ),
		icon = HoJ:GetSpellIcon( 255937 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 5,
		range = HoJ.RANGE_12_YARDS,
		talentRow = 4,
		talentCol = 3,
		talentRole = HoJ.ROLE_DPS
	},
	["bladeOfJustice"] = {
		id = 184575,
		name = HoJ:GetSpellName( 184575 ),
		rank = HoJ:GetSpellRank( 184575 ),
		icon = HoJ:GetSpellIcon( 184575 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 20,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 2,
		range = HoJ.RANGE_12_YARDS
	},
	["empyreanPower"] = {
		id = 286390,
		name = HoJ:GetSpellName( 286390 ),
		rank = HoJ:GetSpellRank( 286390 ),
		iconId = 53385, -- shows the divine storm icon in the skill window instead of empyrean power
		icon = HoJ:GetSpellIcon( 53385 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 120,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_12_YARDS
	},
	["empyreanPowerBuff"] = {
		id = 286393,
		name = HoJ:GetSpellName( 286393 ),
		rank = HoJ:GetSpellRank( 286393 ),
		iconId = 53385, -- shows the divine storm icon in the skill window instead of empyrean power
		icon = HoJ:GetSpellIcon( 286393 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 120,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_12_YARDS
	},
	["wordOfGlory"] = {
		id = 210191,
		name = HoJ:GetSpellName( 210191 ),
		rank = HoJ:GetSpellRank( 210191 ),
		icon = HoJ:GetSpellIcon( 210191 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = -3,
		range = HoJ.RANGE_40_YARDS,
		talentRow = 6,
		talentCol = 3,
		talentRole = HoJ.ROLE_DPS
	},
	["justicarsVengeance"] = {
		id = 215661,
		name = HoJ:GetSpellName( 215661 ),
		rank = HoJ:GetSpellRank( 215661 ),
		icon = HoJ:GetSpellIcon( 215661 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = -5,
		range = HoJ.RANGE_40_YARDS,
		talentRow = 6,
		talentCol = 2,
		talentRole = HoJ.ROLE_DPS
	},
	["shieldOfVengeance"] = {
		id = 184662,
		name = HoJ:GetSpellName( 184662 ),
		rank = HoJ:GetSpellRank( 184662 ),
		icon = HoJ:GetSpellIcon( 184662 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["avengersShield"] = {
		id = 31935,
		name = HoJ:GetSpellName( 31935 ),
		rank = HoJ:GetSpellRank( 31935 ),
		icon = HoJ:GetSpellIcon( 31935 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["hammerOfTheRighteous"] = {
		id = 53595,
		name = HoJ:GetSpellName( 53595 ),
		rank = HoJ:GetSpellRank( 53595 ),
		icon = HoJ:GetSpellIcon( 53595 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["blessedHammer"] = {
		id = 204019,
		name = HoJ:GetSpellName( 204019 ),
		rank = HoJ:GetSpellRank( 204019 ),
		icon = HoJ:GetSpellIcon( 204019 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 15,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 1,
		talentCol = 3,
		talentRole = HoJ.ROLE_TANK
	},
	["forbearance"] = {
		id = 25771,
		name = HoJ:GetSpellName( 25771 ),
		rank = HoJ:GetSpellRank( 25771 ),
		icon = HoJ:GetSpellIcon( 25771 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 15,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	}
};

HoJ._protActiveBuffList = {
	[HoJ.SpellInfo.handOfTheProtector.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.lightOfTheProtector.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.shieldOfTheRighteous.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.giftOfTheNaaru.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.ardentDefender.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.guardianOfAncientKings.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.layOnHands.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.blessingOfProtection.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.divineShield.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.forbearance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.avengingWrath.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.protConsecrationBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.avengersValor.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 }
}

HoJ._retActiveBuffList = {
	[HoJ.SpellInfo.giftOfTheNaaru.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.layOnHands.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.blessingOfProtection.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.divineShield.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.forbearance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.shieldOfVengeance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.avengingWrath.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.crusade.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.retConsecration.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
--	[HoJ.SpellInfo.blessingOfKings.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
--	[HoJ.SpellInfo.blessingOfWisdom.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.divinePurposeBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.inquisition.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[HoJ.SpellInfo.empyreanPowerBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 }
}

HoJ._combatAuraSpellInfo = HoJ.SpellInfo.retributionAura; -- spell info for the aura we want to check for in combat

local sliderBackdrop  = {
	bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
	edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
	tile = true, tileSize = 8, edgeSize = 8,
	insets = { left = 3, right = 3, top = 6, bottom = 6 }
};

local listBackdrop = {
  -- path to the background texture
  bgFile = "",
  -- path to the border texture
  edgeFile = "Interface/Buttons/UI-SliderBar-Border",
  -- true to repeat the background texture to fill the frame, false to scale it
  tile = true,
  -- size (width or heighet) of the square repeating background tiles (in pixels)
  tileSize = 8,
  -- thickness of edge segments and square size of edge corners (in pixels)
  edgeSize = EDGE_SIZE,
  -- distance from the edges of the frame to those of the background texture (in pixels)
  insets = {
	left = 3,
	right = 3,
	top = 3,
	bottom = 3
  }
};

local SAVED_VARIABLES = "HoJSavedVars";

local DEFAULTS = {
	["blockedNames"] = { },
	["silentIgnore"] = false,
	["stealthIgnore"] = false,
	["showSkillCDs"] = true,
	["showSkillsBar"] = true,
	["showBuffsBar"] = true,
	["spellInfo"] = { }
};

HoJ.savedVars = { };

-- -----------------------------------------------------------------------------
-- HoJ:InitSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:InitSavedVariables()
	HoJ.savedVars = DEFAULTS;

	if ( _G[SAVED_VARIABLES] == nil ) then
		HoJ:DebugMsg( "Using defaults." );
	else
		HoJ:DebugMsg( "Using saved variables." );

		HoJ.savedVars = _G[SAVED_VARIABLES];
		for key, value in pairs( HoJ.SpellInfo ) do
			local savedSpellInfo = HoJ.savedVars["spellInfo"];
			if ( savedSpellInfo[key] == nil ) then
				HoJ:DebugMsg( "Adding saved var " .. key .. ", " .. value.name );
				savedSpellInfo[key] = { showOnBar = true };
			-- else
			--	HoJ:DebugMsg( "Found saved var " .. key .. " = " .. tostring( savedSpellInfo[key].showOnBar ) );
			end
		end
	end

	HoJ:ReadSavedVariables();

	_G[SAVED_VARIABLES] = HoJ.savedVars;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetDefaults
-- -----------------------------------------------------------------------------
function HoJ:SetDefaults()
	HoJ.savedVars = DEFAULTS;
	_G[SAVED_VARIABLES] = HoJ.savedVars;
	HoJ:ReadSavedVariables();
end

-- -----------------------------------------------------------------------------
-- HoJ:ReadVar
-- -----------------------------------------------------------------------------
function HoJ:ReadVar( name, default )
	if ( name == nil ) then
		return;
	end

	if ( HoJ.savedVars[name] ~= nil ) then
		-- HoJ:DebugMsg( "Read " .. name .. " = " .. tostring( HoJ.savedVars[name] ) );
		return HoJ.savedVars[name];
	else
		-- HoJ:DebugMsg( "Read " .. name .. " = nil, set to default of " .. tostring( default ) );
		return default;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ReadSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:ReadSavedVariables()
	--HoJ:DebugMsg( "ReadSavedVariables" );

	-- versioning
	HoJ._savedVersionMajor = HoJ:ReadVar( "VersionMajor", -1 )
	HoJ._savedVersionMinor = HoJ:ReadVar( "VersionMinor", -1 )
	HoJ._savedVersionBuild = HoJ:ReadVar( "VersionBuild", -1 )

	if ( HoJ.savedVars.spellInfo ~= nil ) then
		for key, value in pairs( HoJ.savedVars.spellInfo ) do
			--HoJ:DebugMsg( "Reading var " .. key );
			if ( HoJ.SpellInfo[ key ] ~= nil ) then -- don't error if there's an old var
				HoJ.SpellInfo[ key ].showOnBar = value.showOnBar;
			end
		end
	end

	-- ignore list
	HoJ:ReadIgnoreVariables();

	-- misc
	HoJ._showSkillCDs = HoJ:ReadVar( "showSkillCDs", true );
	HoJ._showCooldownsFrame = HoJ:ReadVar( "showCooldownsFrame", true );
	HoJ._showBuffsFrame = HoJ:ReadVar( "showBuffsFrame", true );
	HoJ._showRotationFrame = HoJ:ReadVar( "showRotationFrame", true );
	HoJ._showTemplarsVerdictFrame = HoJ:ReadVar( "showTemplarsVerdictFrame", true );
	HoJ._showHolyPowerFrame = HoJ:ReadVar( "showHolyPowerFrame", true );
	HoJ._showInquisitionFrame = HoJ:ReadVar( "showInquisitionFrame", true );

	HoJ._globalAlpha = HoJ:ReadVar( "globalAlpha", 1.0 )
	HoJ._rotationAlpha = HoJ:ReadVar( "rotationAlpha", 1.0 )
	HoJ._cooldownsAlpha = HoJ:ReadVar( "cooldownsAlpha", 1.0 )
	HoJ._buffsAlpha = HoJ:ReadVar( "buffsAlpha", 1.0 )
	HoJ._holyPowerAlpha = HoJ:ReadVar( "holyPowerAlpha", 1.0 )
	HoJ._inquisitionAlpha = HoJ:ReadVar( "inquisitionAlpha", 1.0 )

	if ( HoJ._savedVersionMajor < 0 or HoJ._savedVersionMinor < 0 or HoJ._savedVersionBuild < 0 ) then
		HoJ:DebugMsg( "Saved variables are not versioned. Some variables may be reset." )
		-- saved variable versioning was added in 0.94.4, anything with defaulted version
		-- numbers will need to have the Flash of Light showOnBar variable reset
		HoJ.SpellInfo.flashOfLight.showOnBar = true
		HoJ.SpellInfo.divinePurpose.showOnBar = true
		HoJ.SpellInfo.crusade.showOnBar = true
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:WriteVar
-- -----------------------------------------------------------------------------
function HoJ:WriteVar( name, var, default )
	if ( var == nil ) then
		HoJ:DebugMsg( "Boolean var " .. name .. " is nil!" );
		HoJ.savedVars[name] = default;
	else
		--HoJ:DebugMsg( "Wrote " .. name .. " = " .. tostring( var ) );
		HoJ.savedVars[name] = var;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:WriteSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:WriteSavedVariables()
	--HoJ:DebugMsg( "WriteSavedVariables" );

	-- versioning
	HoJ:WriteVar( "VersionMajor", HoJ.MAJOR_VERSION, HoJ.MAJOR_VERSION )
	HoJ:WriteVar( "VersionMinor", HoJ.MINOR_VERSION, HoJ.MINOR_VERSION )
	HoJ:WriteVar( "VersionBuild", HoJ.BUILD_VERSION, HoJ.BUILD_VERSION )

	for key, value in pairs( HoJ.SpellInfo ) do
		--HoJ:DebugMsg( "Writing var " .. key );
		local t = { showOnBar = value.showOnBar; }
		local v = HoJ.savedVars["spellInfo"];
		if ( v == nil ) then
			HoJ:DebugMsg( "v = nil" );
		end
		v[ key ] = t;
	end

	-- ignore list
	HoJ:WriteIgnoreVariables();

	-- misc
	HoJ:WriteVar( "showSkillCDs", HoJ._showSkillCDs, true );
	HoJ:WriteVar( "showSkillsFrame", HoJ._showCooldownsFrame, true );
	HoJ:WriteVar( "showBuffsFrame", HoJ._showBuffsFrame, true );
	HoJ:WriteVar( "showRotationFrame", HoJ._showRotationFrame, true );
	HoJ:WriteVar( "showTemplarsVerdictFrame", HoJ._showTemplarsVerdictFrame, true );
	HoJ:WriteVar( "showHolyPowerFrame", HoJ._showHolyPowerFrame, true );
	HoJ:WriteVar( "showInquisitionFrame", HoJ._showInquisitionFrame, true );

	HoJ:WriteVar( "globalAlpha", self._globalAlpha )
	HoJ:WriteVar( "rotationAlpha", self._rotationAlpha )
	HoJ:WriteVar( "cooldownsAlpha", self._cooldownsAlpha )
	HoJ:WriteVar( "buffsAlpha", self._buffsAlpha )
	HoJ:WriteVar( "holyPowerAlpha", self._holyPowerAlpha )
	HoJ:WriteVar( "inquisitionAlpha", self._inquisitionAlpha )
end

-----------------------------------------------------------------------------
-- HoJ:ReadIgnoreVariables
-- -----------------------------------------------------------------------------
function HoJ:ReadIgnoreVariables()
	if ( HoJ.savedVars["blockedNames"] ~= nil ) then
		HoJ._blockedNames = HoJ.savedVars["blockedNames"];
	end
	if ( HoJ.savedVars["silentIgnore"] ~= nil ) then
		HoJ._silentIgnore = HoJ.savedVars["silentIgnore"];
	end
	if ( HoJ.savedVars["stealthIgnore"] ~= nil ) then
		HoJ._stealthIgnore = HoJ.savedVars["stealthIgnore"];
	end
end

-----------------------------------------------------------------------------
-- HoJ:WriteIgnoreVariables
-- -----------------------------------------------------------------------------
function HoJ:WriteIgnoreVariables()
	HoJ.savedVars["blockedNames"] = HoJ._blockedNames;
	HoJ.savedVars["silentIgnore"] = HoJ._silentIgnore;
	HoJ.savedVars["stealthIgnore"] = HoJ._stealthIgnore;
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRole
-- -----------------------------------------------------------------------------
function HoJ:UpdateRole()
	local specId = GetSpecialization();

	-- HoJ:DebugMsg( "UpdateRole(): specId = " .. specId );
	if ( specId == 3 ) then
		HoJ._role = HoJ.ROLE_DPS;
	elseif ( specId == 2 ) then
		HoJ._role = HoJ.ROLE_TANK;
	elseif ( specId == 1 ) then
		HoJ._role = HoJ.ROLE_HEALING;
	else
		HoJ._role = HoJ.ROLE_UNKNOWN;
	end
end

-- -----------------------------------------------------------------------------
-- Slash command handling
-- -----------------------------------------------------------------------------
SLASH_HAMMEROFJUSTAS1, SLASH_HAMMEROFJUSTAS2 = '/hoj', '/hammerofjustas';
function SlashCmdList.HAMMEROFJUSTAS( msg, editbox )
	local command, rest = msg:match( "^(%S*)%s*(.-)$" );
	local name = rest:match( "^(%S*)%s*(%a-)" );

	-- HoJ:Print( "command = '" .. command .. "', rest = '" .. rest .. "'." );
	-- HoJ:Print( "name = '" .. name .. "'." );

	if ( command == "defaults" ) then
		HoJ:SetDefaults();
		HoJ:Msg( "Reset to defaults." );
	elseif ( command == "ignore" or command == "add" ) then
		HoJ._blockedNames[ name ] = name;
		HoJ.WriteBlockedNames();
		HoJ:Msg( "Added " .. name .. " to whisper ignore list." );
	elseif ( command == "unignore" or command == "remove" ) then
		HoJ._blockedNames[ name ] = nil;
		HoJ.WriteBlockedNames();
		HoJ:Msg( "Removed " .. name .. " from whisper ignore list." );
	elseif ( command == "list" ) then
		local count = 1;
		HoJ:Msg( "Ignore List:" );
		for key, value in pairs( HoJ._blockedNames ) do
			HoJ:Print( HoJ.STR_WHITE .. "  " .. count .. ": " .. value );
			count = count + 1;
		end
	elseif ( command == "silent" ) then
		if ( HoJ._silentIgnore == true ) then
			HoJ._silentIgnore = false;
			HoJ:Msg( "Silent Ignore is OFF." );
		else
			HoJ._silentIgnore = true;
			HoJ:Msg( "Silent Ignore is ON." );
		end
	elseif ( command == "stealth" ) then
		if ( HoJ._stealthIgnore == true ) then
			HoJ._stealthIgnore = false;
			HoJ:Msg( "Stealth Ignore is OFF." );
		else
			HoJ._stealthIgnore = true;
			HoJ:Msg( "Stealth Ignore is ON." );
		end
	elseif ( command == "history" ) then
		HoJ:Msg( "Ignore History:" );
		for key, value in pairs( HoJ._ignoreHistory ) do
			HoJ:Print( HoJ.STR_WHITE .. "From " .. key .. ": " );
			for i, historyMsg in ipairs( HoJ._ignoreHistory[ key ] ) do
				HoJ:Print( HoJ.STR_WHISPER .. "  " .. historyMsg );
			end
		end
	elseif ( command == "clear" ) then
		HoJ._ignoreHistory = { };
		HoJ:Print( "History cleared." );
	elseif ( command == "prof" or command == "profile" ) then
		if ( HoJ._enableProfiling ) then
			HoJ._enableProfiling = false;
			HoJ:Msg( "Profiling turned OFF." );
		else
			HoJ._enableProfiling = true;
			HoJ:Msg( "Profiling turned ON." );
		end
	elseif ( command == "thresh" or command == "threshold" ) then
		HoJ._profileThreshold = tonumber( name )
		if ( HoJ._profileThreshold == nil ) then
			HoJ._profileThreshold = 0.3
		end
		HoJ:Msg( "Profiling threshold set to " .. HoJ._profileThreshold );
	elseif ( command == "func" ) then
		HoJ._profileFunction = name
	elseif ( command == "disable" ) then
		HoJ._enabled = false;
		HoJ._needsReset = true;
		HoJ:Msg( "Update disabled." );
	elseif ( command == "enable" ) then
		HoJ._enabled = true;
		HoJ:Msg( "Update enabled." );
	else
		HoJ:Msg( "slash commands:" );
		HoJ:Print( HoJ.STR_WHITE .. "  ignore <name> - ignores whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  add <name> - same as ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  unignore <name> - stops ignoring whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  remove <name> - same as unignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  list - lists all currently ignored players." );
		HoJ:Print( HoJ.STR_WHITE .. "  silent - toggle silent ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  stealth - toggle stealth ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  profile - toggle addon profiling." );
		HoJ:Print( HoJ.STR_WHITE .. "  disable - disable the addon's update." );
		HoJ:Print( HoJ.STR_WHITE .. "  enable - enable the addon's update." );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnterCombat
-- -----------------------------------------------------------------------------
function HoJ:EnterCombat( ... )
	--HoJ:DebugMsg( "HoJ:EnterCombat" );
	HoJ._inCombat = true;

	-- intiate the inquisition cooldown because we may have turned on inquisition
	-- outside of combat.
--	HoJ:StartInquisitionCooldown();
end

-- -----------------------------------------------------------------------------
-- HoJ:LeaveCombat
-- -----------------------------------------------------------------------------
function HoJ:LeaveCombat( ... )
	--HoJ:DebugMsg( "HoJ:LeaveCombat" );
	HoJ._inCombat = false;
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIcon
-- -----------------------------------------------------------------------------
function HoJ:AddIcon( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		HoJ:RemoveIcon( list, spellInfo )
		return
	end

	-- HoJ:DebugMsg( "AddIcon( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconExclusive
-- -----------------------------------------------------------------------------
function HoJ:AddIconExclusive( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		self:DebugMsg( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		self:RemoveIcon( list, spellInfo )
		return
	end

	for i = 1, #list do
		if ( list[i].name == spellInfo.name ) then
			return
		end
	end

	-- self:DebugMsg( "AddIconExclusive( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconDebug
-- just adds debug text where the cooldown text would be
function HoJ:AddIconDebug( list, spellInfo, debugText )
	HoJ:AddIcon( list, spellInfo, "" )
end

-- -----------------------------------------------------------------------------
-- HoJ:RemoveIcon
-- -----------------------------------------------------------------------------
function HoJ:RemoveIcon( list, spellInfo )
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:RemoveIcon - Icon list is NIL" );
	end
	-- HoJ:DebugMsg( "RemoveIcon( " .. spellInfo.name .. " )" );
	for i, v in pairs( list ) do
		if ( v == spellInfo ) then
			table.remove( list, i );
			break;
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ClearIconList
-- -----------------------------------------------------------------------------
function HoJ:ClearIconList( list )
	for i, v in pairs( list ) do
		list[ i ] = nil;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SetIconFrame
-- -----------------------------------------------------------------------------
function HoJ:SetIconFrame( v, spellInfo )
	-- self:DebugMsg( "SetIconFrame for " .. spellInfo.name .. ", frame " .. v.frame:GetName() )
	if ( v == nil ) then
		HoJ:DebugMsg( "WTF" )
		return -- there are more skills than frames
	end

	v.frame:EnableMouse( true )
	--v.frame:Enable();
	v.frame:Show()
	-- HoJ:DebugMsg( "Enable mouse for " .. v.frame:GetName() );

	v.spellInfo = spellInfo

	v.texture:SetTexture( spellInfo.icon )
	v.texture:Show()

	local hideText = true
	if ( v.text ~= nil ) then
		if ( spellInfo.text ~= nil and spellInfo.text ~= "" ) then
			local fontFile = GameFontHighlightLarge:GetFont();
			v.text:SetFont( fontFile, 12, "THICKOUTLINE" );
			v.text:SetText( spellInfo.text );
			v.text:Show();
			if ( spellInfo.spellType == HoJ.SPELL_SKILL or spellInfo.spellType == HoJ.SPELL_TALENT ) then
				if ( v.texture ~= nil ) then
					v.texture:SetAlpha( HoJ._skillCDAlpha );
				end
			end
		end
	end
	if ( hideText ) then
		v.text:Hide();
		if ( v.texture ~= nil ) then
			v.texture:SetAlpha( HoJ._skillNormalAlpha );
		end
	end

	-- if the frame has a cooldown, set it
	if ( v.cooldown ~= nil ) then
		local hideCooldown = true
		local start, duration, enabled = GetSpellCooldown( spellInfo.id )
		if ( start > 0 and duration > 0 ) then
			v.cooldown:Show()
			v.cooldown:SetCooldown( start, duration )
			hideCooldown = false
		end
		if ( hideCooldown ) then
			v.cooldown:Hide()
		end
--	end

--	if ( v.charges ~= nil ) then
		local hideCharges = true
		local fontFile = GameFontHighlightSmallOutline:GetFont()
		v.charges:SetFont( fontFile, 12, "THICKOUTLINE" )
		local currentCharges, maxCharges = GetSpellCharges( spellInfo.id )
		if ( currentCharges ~= nil and maxCharges > 1 ) then
			v.charges:SetText( tostring( currentCharges ) )
			v.charges:Show()
			hideCharges = false
		end
		if ( hideCharges ) then
			v.charges:Hide()
		end
	end
--[[
	-- we cannot use SetAttribute() on a secure frame while in combat
	if ( not HoJ._inCombat ) then
		-- HoJ:DebugMsg( "spell " .. spellInfo.name );
		v.frame:SetAttribute( "type", "spell" );
		v.frame:SetAttribute( "unit", "player" );
		if ( v.spellToCast ~= nil ) then
			-- some buffs are meant to affect another spell (i.e. Selfless Healer modifies cast time of Flash of Light)
			-- though, due to combat restrictions, this isn't actually castable if it is added to the bar during combat (which it usually is).
			v.frame:SetAttribute( "spell", HoJ:GetSpellName( spellInfo.spellToCast ) );
		else
			v.frame:SetAttribute( "spell", spellInfo.name );
		end
	end
]]
end

function SkillCompare( tbl, a, b )
	local cdra = tbl[a].cooldownRemaining;
	local cdrb = tbl[b].cooldownRemaining;

	if ( cdra <= 0.00001 ) then
		if ( cdrb <= 0.00001 ) then
			return ( tbl[a].priority > tbl[b].priority );
		else
			return true;
		end
	elseif ( cdrb <= 0.00001 ) then
		-- tbl[a] must be on cooldown or we'd have gone into the first if
		return false;
	else
		-- both are on cooldown, sort by cooldown remaining time
		return ( cdra < cdrb );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateIcons
-- -----------------------------------------------------------------------------
local SortedKeyLists = {
	{ 1 },
	{ 1, 2 },
	{ 1, 2, 3 },
	{ 1, 2, 3, 4 },
	{ 1, 2, 3, 4, 5 }
}

function HoJ:UpdateIcons( list, frameList, sortList )
	-- clear all icon frames

	local setIconFrame = HoJ.SetIconFrame
	local frameNum = 0

	local size = #list
	if ( size > 1 and sortList ) then
		local sortIconsStartMs = ProfileBegin2( "SortIcons" )

		-- use a table lookup to initialize the sorted keys list to the correct size
		local sortedKeys = SortedKeyLists[size]
		table.sort( sortedKeys, function( a, b ) return SkillCompare( list, a, b ) end )

		ProfileEnd( "SortIcons", sortIconsStartMs )

		for i = 1, size do
			setIconFrame( self, frameList[i], list[ sortedKeys[i] ] )
			frameNum = i
		end
	else
		--local sortIconFramesStartMs = ProfileBegin2( "SetIconFramesSorted" )

		-- using #list is slightly faster in this case than ipairs()
		for i = 1, size do
			setIconFrame( self, frameList[i], list[i] )
			frameNum = i
		end

		-- ProfileEnd( "SetIconFramesSorted", sortIconFramesStartMs )
	end

	--local enableMouseStartMs = ProfileBegin2( "EnableMouse" )
	for i = (frameNum + 1), #frameList do
		local frame = frameList[ i ].frame
		frame:EnableMouse( false )
		frame:Hide()
	end
	--ProfileEnd( "EnableMouse", enableMouseStartMs )
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitBuffForName
-- -----------------------------------------------------------------------------
function HoJ:UnitBuffForName( unit, buffName )
	local i = 1;
	local name, icon, count = UnitAura( unit, i );
	while name do
		if ( name == buffName ) then
			return name, icon, count;
		end
		i = i + 1;
		name, icon, count = UnitBuff( unit, i );
	end
	return name, icon, count;
end

-- -----------------------------------------------------------------------------
-- HoJ:FindActiveBuffs
-- table of flags for each buff, indexed by spell id
--	{
--		isActive,		-- whether or not the buff is active
--		start,			-- time buff started
--		duration,		-- duration of the buff
--		timeRemaining, 	-- how long the buff has before expiration
--		stacks,			-- number of stacks of the buff
--		spellId,		-- spell Id of the buff
--	}
-- -----------------------------------------------------------------------------
function HoJ:FindActiveBuffs( unit, activeBuffs )
	-- set all active skills active flag to false
	for key, buff in pairs( activeBuffs ) do
		buff.isActive = false
		buff.start = 0.0
		buff.duration = 0.0
		buff.timeRemaining = 0.0
		buff.stacks = 0
	end

	local i = 1;
	local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i )
	while name do
		-- HoJ:DebugMsg( "Buff " .. name .. " is active, spellId = " .. spellId )
		if ( spellId ~= nil ) then
			-- HoJ:DebugMsg( "Buff " .. name .. " is active, spellId = " .. spellId )
			local activeBuff = activeBuffs[spellId]
			if ( activeBuff ~= nil ) then
				-- HoJ:DebugMsg( "Found " .. name .. " in activeBuffs" )
				activeBuff.isActive = true
				if ( duration ~= nil and expirationTime ~= nil ) then
					activeBuff.start = expirationTime - duration
					activeBuff.duration = duration
					activeBuff.timeRemaining = expirationTime - self._curTime
					if ( activeBuff.timeRemaining < 0.0 ) then
						activeBuff.timeRemaining = 0.0
					end
				end
				if ( count ~= nil ) then
					activeBuff.stacks = count
				end
			-- else
			--		HoJ:DebugMsg( "Buff " .. name .. " (id " .. spellId .. ") not in activeBuffs list" )
			end
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBuffActive
-- -----------------------------------------------------------------------------
function HoJ:IsBuffActive( spellId )
	local buff = nil
	if ( self._role == HoJ.ROLE_DPS ) then
		buff = self._retActiveBuffList[spellId]
	elseif ( self._role == HoJ.ROLE_TANK ) then
		buff = self._protActiveBuffList[spellId]
	else
		self:DebugMsg( "No active buff list for role " .. self._role )
		return false;
	end
	if ( buff == nil ) then
		return false
	end
	return buff.isActive
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuff
-- -----------------------------------------------------------------------------
function HoJ:GetBuff( activeBuffList, spellId )
	return activeBuffList[spellId]
end

-- -----------------------------------------------------------------------------
-- HoJ:GetUnitBuffTimeRemaining
-- returns:
--	float 	buffTimeRemaining
--	int 	stacks
-- -----------------------------------------------------------------------------
function HoJ:GetUnitBuffTimeRemaining( unit, inSpellId, filter )
	for i = 1, 40 do
		local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
		if ( name == nil ) then
			return 0.0, 0
		end
		-- self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
		if ( spellId == inSpellId ) then
			--self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
			return expirationTime - HoJ._curTime, count
		end
	end
	return 0.0, 0
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingWithPrediction
-- gets the cooldown remaining, including GCD if GCD is in progress
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingWithPrediction( spellId, predictAhead )
	local start, duration, enabled = GetSpellCooldown( spellId );

	-- we don't really care about enabled... that just tells us if the spell / buff is active
	if ( start > 0 and duration > 0 ) then
		-- we "look into the future" to just past the time when the current GCD is done,
		-- to see what will be usable then
		local remaining = ( start + duration ) - ( HoJ._curTime + predictAhead );
		if ( remaining < 0.0 ) then
			remaining = 0.0;
		end
		return remaining;
	end
	return 0.0;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingAfterGCD
-- For predicting what use next, we don't really care about the GCD period, we
-- just want to know if a spell will be usable once the GCD ends.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingAfterGCD( spellId )
	-- we "look into the future" to just past the time when the current GCD is done,
	-- to see what will be usable then
	return HoJ:GetCDRemainingWithPrediction( spellId, HoJ._gcdRemaining );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingIgnoringGCD
-- normally when we check for a spell cooldown, all spells return the GCD cooldown,
-- i.e., even if the spell itself is not on cooldown, if the GCD is, then it will
-- appear that the spell is on cooldown.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingIgnoringGCD( spellId )
	local start, duration, enabled = GetSpellCooldown( spellId );

	if ( start <= 0 or duration <= 0 ) then
		return 0.0
	end

	local endTime = start + duration
	local remaining = endTime - HoJ._curTime

	if ( self._gcdRemaining ) then
		-- we're in GCD...
		if ( self._gcdRemaining >= remaining ) then
			-- this is just a GCD cooldown, not the actual spell cooldown
			return 0.0
		end
	end
	return remaining
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffTimeRemaining
-- returns:
--   time remaining in seconds
--   start time
--   duration in seconds
-- -----------------------------------------------------------------------------
function HoJ:GetBuffTimeRemaining( spellId )
	local buff = nil
	if ( self._role == HoJ.ROLE_DPS ) then
		buff = self._retActiveBuffList[spellId]
	elseif ( self._role == HoJ.ROLE_TANK ) then
		buff = self._protActiveBuffList[spellId]
	end

	if ( buff == nil ) then
		return 0.0, 0.0, 0.0
	end

	return buff.timeRemaining, buff.start, buff.duration
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellIndex
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellIndex( spellIndex, name, unit, debugName )
	if ( spellIndex < 0 or spellIndex == nil ) then
		HoJ:DebugMsg( "Invalid spellIndex " .. spellIndex .. " for " .. debugName )
		return false
	end
	return IsSpellInRange( spellIndex, name, unit )
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellName
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellName( spellName, unit, debugName )
	if ( spellName == nil ) then
		HoJ:DebugMsg( "Invalid spell name " .. spellName .. " for " .. debugName )
		return false
	end
	return IsSpellInRange( spellName, unit )
end

-- -----------------------------------------------------------------------------
-- HoJ:InRange
-- -----------------------------------------------------------------------------
function HoJ:InRange( spellInfo, unit, debugName )
	local inRange = false;
	if ( spellInfo.spellType == HoJ.SPELL_TALENT ) then
		inRange = HoJ:SpellInRange_SpellIndex( spellInfo.spellIndex, "spell", unit, debugName );
	else
		inRange = HoJ:SpellInRange_SpellName( spellInfo.name, unit, debugName );
	end

	if ( inRange == nil ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = nil" );
		return false;
	elseif ( inRange == 0 ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = 0" );
		return false;
	end
	--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = true" );
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:BoolToStr
-- -----------------------------------------------------------------------------
function HoJ:BoolToStr( boolVar )
	if ( boolVar ) then
		return "true";
	end
	return "false";
end

-- -----------------------------------------------------------------------------
-- HoJ:WillOverflowHoPo
-- -----------------------------------------------------------------------------
function HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo )
	local overflow = ( curHoPo + spellInfo.hoPoGen ) - maxHoPo;
	-- HoJ:DebugMsg( spellInfo.name .. ": overflow = " .. overflow );
	return overflow;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsOffCooldown
-- -----------------------------------------------------------------------------
function HoJ:IsOffCooldown( spellInfo, predictAhead )
	local cdRemaining = HoJ:GetCDRemainingAfterGCD( spellInfo.id );
	-- HoJ:DebugMsg( "spell " .. spellInfo.name .. ": cdRemaining = " .. cdRemaining .. ", GCD Remaining = " .. HoJ._gcdRemaining );
	if ( cdRemaining == nil ) then
		HoJ:DebugMsg( "cdRemaining == nil!" )
		return false;
	elseif ( predictAhead == nil ) then
		return cdRemaining <= 0.0
	end
	if ( cdRemaining <= predictAhead ) then
		return true;
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsSkillUsable
-- -----------------------------------------------------------------------------
function HoJ:IsSkillUsable( spellInfo, predictAhead )
	if ( not HoJ._targetInRange[spellInfo.range] ) then
		-- HoJ:DebugMsg( spellInfo.name .. " is not in range" )
		return false;
	end
	if ( not HoJ:IsOffCooldown( spellInfo, predictAhead ) ) then
		return false;
	end
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsSkillUsableWithoutOverflow
-- -----------------------------------------------------------------------------
function HoJ:IsSkillUsableWithoutOverflow( spellInfo, predictAhead, curHoPo, maxHoPo )
	if ( not HoJ:IsSkillUsable( spellInfo, predictAhead ) ) then
		return false;
	end
	if ( HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo ) > 0 ) then
		return false;
	end
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckProcdBuffs
-- -----------------------------------------------------------------------------
function HoJ:CheckProcdBuffs()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end

	if ( self:IsBuffActive( self.SpellInfo.divinePurposeBuff.id ) ) then
		self:AddIcon( self._buffs, self.SpellInfo.divinePurpose )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ShowSkillsList
-- -----------------------------------------------------------------------------
function HoJ:ShowSkillsList( prefix, list )
	local skillsOrder = "";
	for i, v in pairs( list ) do
		if ( i ~= 1 ) then
			skillsOrder = skillsOrder .. ", ";
		end
		if ( v.text == nil ) then
			skillsOder = skillsOrder .. "nil";
		else
			skillsOrder = skillsOrder .. v.text;
		end
	end
	HoJ:DebugMsg( prefix .. skillsOrder );
end

-- -----------------------------------------------------------------------------
-- HoJ:IsTalentSelected
-- -----------------------------------------------------------------------------
function HoJ:IsTalentSelected( spellInfo )
	if ( spellInfo.talentRole ~= HoJ._role and spellInfo.talentRole ~= HoJ.ROLE_ANY ) then
		if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " does not match role " .. HoJ._role .. "!" )
		end
		return false
	end

	if ( spellInfo.talentRow == nil or spellInfo.talentCol == nil ) then
		if ( spellInfo.spellType ~= HoJ.SPELL_TALENT ) then
			HoJ:DebugMsg( "IsTalentSelected() called on non-talent " .. spellInfo.name );
		else
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " missing row and column" );
		end
		return true;
	end

	local _, _, _, selected, available = GetTalentInfo( spellInfo.talentRow, spellInfo.talentCol, 1 );
	if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
		HoJ:DebugMsg( "Talent " .. spellInfo.name .. ", available = " .. self:BoolToStr( available ) .. ", selected = " .. self:BoolToStr( selected ) )
	end
	return ( available and selected );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetHolyPower
-- -----------------------------------------------------------------------------
function HoJ:GetHolyPower()
	return HoJ._fakeHolyPower
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckInquisition
-- -----------------------------------------------------------------------------
function HoJ:CheckInquisition()
	if ( HoJ._level < HoJ.SpellInfo.inquisition.minLevel or not HoJ._inCombat or not HoJ:IsTalentSelected( HoJ.SpellInfo.inquisition ) ) then
		HoJ._inqState = HOJ_INQ_NOT_IN_COMBAT;
		return 0.0, 0.0, 0.0
	end
	if ( C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.inquisition.id ) ) then
		return 0.0, 0.0, 0.0
	end

	local inqTimeRemaining, inqStart, inqDuration = HoJ:GetBuffTimeRemaining( HoJ.SpellInfo.inquisition.id );

	local cdText = string.format( "%.1f", inqTimeRemaining );
	HoJ_InqTimer:SetText( cdText );

	if ( inqTimeRemaining <= 0 ) then
		--HoJ:DebugMsg( "HOJ_INQ_EXPIRED" );
		HoJ._inqState = HOJ_INQ_EXPIRED;
		HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );
		HoJ_InqTimer:SetText( "0.0" );
		HoJ:AddIcon( HoJ._cooldowns, HoJ.SpellInfo.inquisition, "0" );
	elseif ( inqTimeRemaining <= 5 ) then
		--HoJ:DebugMsg( "HOJ_INQ_NEARLY_EXPIRED" );
		HoJ._inqState = HOJ_INQ_NEARLY_EXPIRED;
		HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );
	elseif ( inqTimeRemaining <= 10 ) then
		--HoJ:DebugMsg( "HOJ_INQ_EXPIRING" );
		HoJ._inqState = HOJ_INQ_EXPIRING;
		HoJ_InqTimer:SetTextColor( 1.0, 1.0, 0.0 );
	else
		--HoJ:DebugMsg( "HOJ_INQ_ON" );
		HoJ_InqTimer:SetTextColor( 1.0, 1.0, 1.0 );
		HoJ._inqState = HOJ_INQ_ON;
	end

	return inqStart, inqDuration, inqTimeRemaining;
end

-- -----------------------------------------------------------------------------
-- HoJ:StartInquisitionCooldown
-- -----------------------------------------------------------------------------
function HoJ:StartInquisitionCooldown()
	if ( HoJ._inqState == HOJ_INQ_NOT_IN_COMBAT ) then
		return;
	end

	local inqStart, inqDuration, inqTimeRemaining = HoJ:CheckInquisition();
	--HoJ:DebugMsg( "inqStart = " .. inqStart .. ", inqDuration = " .. inqDuration );
	CooldownFrame_SetTimer( HoJ_InqFrameCooldown, inqStart, inqDuration, 1.0 );

	-- find the OmniCC timer frame and hide it so it doesn't overlay our timer
	if ( HoJ._omniCC ) then
		local frame = EnumerateFrames();
		while frame do
			if ( frame:GetParent() == HoJ_InqFrame ) then
				if ( frame:GetName() == nil ) then
					--HoJ:DebugMsg( "A unnamed frame is parented to HoJ_InqFrame." );
					frame:SetAlpha( 0.0 );	-- go away, omniCC
					break
				end
			end
			frame = EnumerateFrames( frame );
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateInquisitionFrames
-- -----------------------------------------------------------------------------
function HoJ:UpdateInquisitionFrames( self )
	--HoJ:DebugMsg( "UpdateInquisitionFrames" );
	if ( HoJ._showInquisitionFrame and HoJ._role == HoJ.ROLE_DPS and HoJ._enabledForRole ) then
		if ( HoJ._inqState == HOJ_INQ_NOT_IN_COMBAT ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_NOT_IN_COMBAT" );
			HoJ_InqFrameCooldown:SetAlpha( 0.0 );
			HoJ_InqTexture:SetAlpha( 0.0 );
			HoJ_InqTimer:SetAlpha( 0.0 );
		elseif ( HoJ._inqState == HOJ_INQ_EXPIRED ) then
			-- animate inquisition texture alpha
			local inqAlpha = finalAlpha + 0.25;
			if ( inqAlpha > 1.0 ) then
				inqAlpha = 1.0;
			end
			--HoJ:DebugMsg( "Render HOJ_INQ_EXPIRED " .. string.format( "%.2f", inqAlpha ) );
			HoJ_InqFrameCooldown:SetAlpha( inqAlpha );
			HoJ_InqTexture:SetAlpha( inqAlpha );
			HoJ_InqTimer:SetAlpha( inqAlpha );
		elseif ( HoJ._inqState == HOJ_INQ_ON ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_ON" );
			HoJ_InqFrameCooldown:SetAlpha( 0.1 );
			HoJ_InqTexture:SetAlpha( 0.1 );
			HoJ_InqTimer:SetAlpha( 0.2 );
		elseif ( HoJ._inqState == HOJ_INQ_NEARLY_EXPIRED ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_*" );
			HoJ_InqFrameCooldown:SetAlpha( 1.0 );
			HoJ_InqTexture:SetAlpha( 1.0 );
			HoJ_InqTimer:SetAlpha( 1.0 );
		elseif ( HoJ._inqState == HOJ_INQ_EXPIRING ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_*" );
			HoJ_InqFrameCooldown:SetAlpha( 0.4 );
			HoJ_InqTexture:SetAlpha( 0.4 );
			HoJ_InqTimer:SetAlpha( 0.5 );
		end
	else
		HoJ_InqFrameCooldown:SetAlpha( 0.0 );
		HoJ_InqTexture:SetAlpha( 0.0 );
		HoJ_InqTimer:SetAlpha( 0.0 );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetHealthRatio
-- -----------------------------------------------------------------------------
function HoJ:GetHealthRatio( unitName )
	local maxHealth = UnitHealthMax( unitName );
	local curHealth = UnitHealth( unitName );
	return curHealth / maxHealth;
end

-- -----------------------------------------------------------------------------
-- HoJ:SelectMitigationSkill
-- expects an array of table items of format:
--	{
--		spellInfo,			-- info about the spell
--		replacedByTalent,	-- optional, can be nil, otherwise specifies the spellInfo of the talent that replaces this spell
--      physicalOnly,		-- optional, can be nil
--		excludedByBuff,		-- optional, if not nil, then this buff prevents the skill from working (Forbearance)
--	}
-- NOTE: the list MUST be in priority order because the loop short-circuits once a
-- mitigation is found, so the order IS the priority
-- -----------------------------------------------------------------------------
function HoJ:SelectMitigationSkill( mitigationSkills, listName )
	if ( listName ~= nil ) then
		self:DebugMsg( HoJ.STR_GREEN .. "Checking mitigations: " .. listName )
	end

	for i, mitigation in ipairs( mitigationSkills ) do
		-- self:DebugMsg( "Checking mitigation skill " .. mitigation.spellInfo.name )
		if ( mitigation.spellInfo.requiredRaceId == nil or mitigation.spellInfo.requiredRaceId == self._playerRaceId ) then
			if ( not self._protActiveBuffList[mitigation.spellInfo.id].isActive ) then
				if ( ( mitigation.replacedByTalent == nil or not self:IsTalentSelected( mitigation.replacedByTalent ) )
					and ( mitigation.excludedByBuff == nil or not self:IsBuffActive( mitigation.excludedByBuffId ) ) ) then
					local isSelected = mitigation.spellInfo.type ~= HoJ.SPELL_TALENT or self:IsTalentSelected( mitigation.spellInfo )
					if ( not C_SpellBook.IsSpellDisabled( mitigation.spellInfo.id ) ) then
						if ( ( mitigation.physicalOnly ~= nil and HoJ._isTakingPhysicalDamage ) or mitigation.physicalOnly == nil ) then
							if ( isSelected ) then
								if ( self:IsOffCooldown( mitigation.spellInfo, 0.0 ) ) then
									-- self:DebugMsg( HoJ.STR_GREEN .. "Selected mitigation: " .. mitigation.spellInfo.name )
									return mitigation
								end
							end
						end
					end
				end
			end
		end
	end
end

HoJ._lowHealthMitigationList = {
	{ spellInfo = HoJ.SpellInfo.handOfTheProtector },
	{ spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector }
}

-- MODERATE MITIGATION SKILL PRIORITY
HoJ._moderateMitigationList = {
	{ spellInfo = HoJ.SpellInfo.shieldOfTheRighteous },
	{ spellInfo = HoJ.SpellInfo.giftOfTheNaaru },
	{ spellInfo = HoJ.SpellInfo.handOfTheProtector },
	{ spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector },
	{ spellInfo = HoJ.SpellInfo.ardentDefender },
	{ spellInfo = HoJ.SpellInfo.guardianOfAncientKings }
}

-- EMERGENCY MITIGATION SKILL PRIORITY
-- in situations where we're about to die very soon, prefer to
--
-- 1) heal using Hand / Light of the Protector
-- 2) heal using Lay on Hands
-- 3) cast ardent defender (because this will prevent a death and heal to 20%)
-- 4) cast guardian of ancient kings
-- 4) blessing of protection (assuming incoming damage is physical)
-- 5) divine shield (last resort, should be followed by a taunt-and-cancel if tanking)
--
-- not really sure about these priorities... in some situations where I'm pretty sure
-- I'm not going to die I prefer to to use Ardent Defender every two minutes as a
-- damage-reducer only.
HoJ._emergencyMitigationList = {
	{ spellInfo = HoJ.SpellInfo.layOnHands, excludedByBuffId = HoJ.SpellInfo.forbearance.id },
	{ spellInfo = HoJ.SpellInfo.handOfTheProtector },
	{ spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector },
	{ spellInfo = HoJ.SpellInfo.ardentDefender },
	{ spellInfo = HoJ.SpellInfo.guardianOfAncientKings },
	{ spellInfo = HoJ.SpellInfo.blessingOfProtection },
	{ spellInfo = HoJ.SpellInfo.divineShield, excludedByBuffId = HoJ.SpellInfo.forbearance.id }
}

-- -----------------------------------------------------------------------------
-- HoJ:CheckMitigationSkills
-- -----------------------------------------------------------------------------

function HoJ:CheckMitigationSkills()
	if ( not self._showCooldownsFrame ) then
		return;
	end

	if ( not self._inCombat and not self._targetCanAttack ) then
		return;
	end

	-- if < 50% health, show Light of the Protector
	if ( self._playerHealthRatio < 0.5 ) then
		local lowHealthMitigation = self:SelectMitigationSkill( self._lowHealthMitigationList )
		if ( lowHealthMitigation ) then
			self:AddIcon( self._cooldowns, lowHealthMitigation.spellInfo )
		end
	end

	if ( self:IsBuffActive( self.SpellInfo.divineShield.id ) ) then
		-- self:DebugMsg( "Divine Shield is on" );
		return
	end

	-- moderate damage mitigation
	if ( self._damagePerSecRatio > 0.1 or self._damagePerSecRatio * 5.0 > self._playerHealthRatio ) then
		local moderateMitigation = self:SelectMitigationSkill( self._moderateMitigationList )
		if ( moderateMitigation ) then
			self:AddIconExclusive( self._cooldowns, moderateMitigation.spellInfo )
		end
	end

	-- emergency mitigation
	if ( self._damagePerSecRatio * 2.0 > self._playerHealthRatio or ( self._predictedDeathTime > 0.0 and self._predictedDeathTime < 5.0 ) ) then
		local emergencyMitigation = self:SelectMitigationSkill( self._emergencyMitigationList )
		if ( emergencyMitigation ) then
			self:AddIconExclusive( self._cooldowns, emergencyMitigation.spellInfo )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckHolyAvenger
-- -----------------------------------------------------------------------------
function HoJ:CheckShieldOfVengeance()
	if ( not self._showBuffsFrame ) then
		return
	end

	if ( C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.shieldOfVengeance.id ) ) then
		return
	end

	if ( not self:IsOffCooldown( self.SpellInfo.shieldOfVengeance, 0.1 ) ) then
		return
	end

	-- don't bother showing Shield of Vengeance if we're not taking much damage
	if ( self._damagePerSecRatio < 0.1 ) then
		return
	end
	local shieldOfVengeanceOn = self:IsBuffActive( self.SpellInfo.shieldOfVengeance.id );
	if ( not self._inCombat ) then
		if ( not shieldOfVengeanceOn and self._targetCanAttack ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.shieldOfVengeance );
		end
	else
		if ( not isOn ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.shieldOfVengeance );
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckAvengingWrath
-- -----------------------------------------------------------------------------
function HoJ:CheckAvengingWrath()
	if ( not HoJ._showCooldownsFrame ) then
		return;
	end

	if ( HoJ._level < HoJ.SpellInfo.avengingWrath.minLevel or not HoJ._inCombat ) then
		return;
	end

	if ( C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.avengingWrath.id ) ) then
		return
	end

	local spell = HoJ.SpellInfo.avengingWrath;
	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.crusade ) ) then
		spell = HoJ.SpellInfo.crusade;
	end

	local awIsOn = HoJ:IsBuffActive( spell.id );
	if ( awIsOn ) then
		-- don't bother showing an icon if Avenging Wrath is currently active
		return;
	end

	local cooldownRemaining = HoJ:GetCDRemainingAfterGCD( spell.id );
	local charges, maxCharges = GetSpellCharges( spell.id );

	if ( charges == nil or maxCharges == 1 ) then
		-- without tier 18 2-piece armor bonus
		if ( cooldownRemaining < 2.5 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell );
		end
	else
		-- happens with tier 18 2-piece armor bonus
		if ( charges > 0 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell, string.format( "%i", charges ) );
		elseif ( cooldownRemaining < 2.5 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell, string.format( "%.1f", awCD ) );
		end
	end
end

HoJ._retActiveBlessingsList = {
	[HoJ.SpellInfo.blessingOfKings.id] = {
		spellInfo = HoJ.SpellInfo.blessingOfKings,
		targetName = nil,
		start = 0.0,
		duration = 0.0,
		updateIndex = -1
	},
	[HoJ.SpellInfo.blessingOfWisdom.id] = {
		spellInfo = HoJ.SpellInfo.blessingOfWisdom,
		targetName = nil,
		start = 0.0,
		duration = 0.0,
		updateIndex = -1
	}
}

-- -----------------------------------------------------------------------------
-- HoJ:GetGroupType
-- -----------------------------------------------------------------------------
function HoJ:GetGroupType()
	if ( IsInRaid() ) then
		return HoJ.GROUP_RAID
	elseif ( IsInGroup() ) then
		return HoJ.GROUP_PARTY
	end
	return HoJ.GROUP_NONE
end

-- -----------------------------------------------------------------------------
-- HoJ:MaxMemberForGroupType
-- The player doesn't count as a partyN unit id, so if there are 3 party members,
-- only "party1" and "party2" are valid unit ids.
-- If the player is in a raid, however, this is not true. If there are 3 raid
-- members, then "raid1", "raid2", and "raid3" are all valid and one of them will
-- represent the "player".
-- To account for this, we subtract one from the number of group members when
-- the player is in a party, but not when in a raid.
-- -----------------------------------------------------------------------------
function HoJ:MaxMemberForGroupType( groupType )
	if ( groupType == HoJ.GROUP_PARTY ) then
		return GetNumGroupMembers() - 1
	end
	return GetNumGroupMembers()
end

-- -----------------------------------------------------------------------------
-- HoJ:PrefixForGroupType
-- -----------------------------------------------------------------------------
function HoJ:PrefixForGroupType( groupType )
	if ( IsInRaid() ) then
		return "raid"
	elseif ( IsInGroup() ) then
		return "party"
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitIDForName
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForName( name )
	if ( name == HoJ._playerName ) then
		return "player"
	end

	local prefix = HoJ:PrefixForGroupType( HoJ._groupType )
	if ( prefix == nil ) then
		return nil
	end

	local max = HoJ:MaxMemberForGroupType( HoJ:GetGroupType() )
	for i = 1, max do
		local unitID = prefix .. i
		local memberName = UnitName( unitID )
		if ( memberName == name ) then
			return unitID
		end
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessing
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessing( blessing )
	blessing.targetName = nil
	blessing.start = 0.0
	blessing.duration = 0.0
	blessing.updateIndex = -1
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessings
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessings( blessings )
	for spellId, blessing in pairs( blessings) do
		--HoJ:DebugMsg( "Resetting " .. blessing.spellInfo.name .. "..." )
		HoJ:ResetBlessing( blessing )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessingFromUnitID
-- returns true if the unitID matches the name, in which case we can definitively
-- know that the unit is the one that was cast on.
-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessingFromUnitID( blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	if ( unitID == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: unitID == nil!" )
		return false
	end

	local unitName = UnitName( unitID )
	if ( unitName == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: UnitName() == nil for unit id '" .. unitID .. "'!" )
		return false
	end

	if ( blessing.targetName ~= unitName ) then
		return false
	end

	--HoJ:DebugMsg( "Blessing target is " .. unitID .. "( " .. unitName .. " )" )

	local found = false
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Updating " .. blessing.spellInfo.name .. " from " .. unitName )
			found = true
			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	-- not found on the unit it was cast on, so it's expired or been canceled
	if ( not found ) then
		--HoJ:DebugMsg( blessing.spellInfo.name .. " was not found. Resetting." )
		HoJ:ResetBlessing( blessing )
	end

	return true
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessingOnInitID
-- -----------------------------------------------------------------------------
function HoJ:FindBlessingOnUnitID( blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Found " .. blessing.spellInfo.name .. " on " .. unitID )
			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
			return true
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForUpdateIndex( index, groupType )
	if ( index < 0 ) then
		return "player"
	elseif ( index == 0 ) then
		return "target"
	end

	local max = HoJ:MaxMemberForGroupType( groupType )
	local prefix = HoJ:PrefixForGroupType( groupType )
	if ( prefix == nil ) then
		return nil
	end

	local i = index
	if ( i > max ) then
		i = max
	end

	local unitID = prefix .. i
	return unitID
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:FindBlessing( blessing )
	--HoJ:DebugMsg( "FindBlessing " .. blessing.spellInfo.name )

	local unitID = HoJ:UnitIDForUpdateIndex( blessing.updateIndex, self._groupType, self._numGroupMembers )
	if ( unitID == nil ) then
		return
	end

	if ( HoJ:FindBlessingOnUnitID( blessing, unitID ) ) then
		-- store the name so we can use the direct lookup
		blessing.targetName = UnitName( unitID )
		return
	end

	-- start over from the beginning
	blessing.updateIndex = blessing.updateIndex + 1
	if ( blessing.updateIndex > self._numGroupMembers ) then
		blessing.updateIndex = -1
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessings
-- For our purposes here a blessing is a buff that can only be on a single target
-- at a time and must be tracked for that target.

-- We can't rely on a blessing being set on a target immediately after cast due
-- to network lag.
-- We can't rely on a blessing being set on the player immeediately after they
-- enter the worl due to network lag.
-- We also cannot rely on being able to check the buff on the a player the buff
-- has been cast on if that player is not in the player's party.
-- For this reason, if we don't find a blessing on a target we have to treat it
-- differently based on whether or not the buff target is in the player's party.

-- To deal with all of the above, when a blessing is cast, we record the name
-- of the target it was cast on (not the unitID it was cast on).

-- Each time we check for blessings, if the blessing has a recorded targetName,
-- we cycle through unit IDs ("player",
-- "target" or and "partyN"/"raidN" for our current group) to try to find the
-- one that matches the recorded name.

-- If we don't find the a unit ID with a matching name, we just check for
-- buff expiration using the stored state.

-- If we do find a unit ID with a matching name, we search the unit's auras for
-- the blessing. If we don't find it, we know it's expired or canceled and we
-- can confidently show the buff icon to indicate it has expired.
--
-- If the buff has no recorded name, then we search through auras on the "player",
-- "target", and "partyN" or "raidN" unitIDs to find a matching buff cast by
-- the player. If we find it, we then store the name of that unit. This handles
-- initializing the buff tracking at any time after the player enters the world.
-- If any player buff is on another player that can be inspected via a unit ID
-- then it will start being tracked. This search through auras on all unit IDs
-- could be very expensive to do all at once for a large raid, so this test is
-- amortized and will check one unit ID per update frame (currently @ 15 fps).
-- This means it could take up to 40 / 15 seconds to find a buff after logging
-- in.

-- table of flags for each buf indexed by spell id
--	{
--		targetName,	-- name of the target the spell was cast on
--		start,		-- time buff started
--		duration,	-- duration of the buff
--		updateIndex	-- index of the current party member to check auras for
--	}
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessings( blessings )
	--HoJ:DebugMsg( "UpdateBlessings: " .. self._curTime )

	local groupType = HoJ:GetGroupType()
	local numGroupMembers = GetNumGroupMembers()
	if ( groupType ~= self._groupType ) then
		-- group changed so restart the buff checks
		--HoJ:DebugMsg( "Resetting update due to group change!" )
		for spellId, blessing in pairs( blessings ) do
			blessing.updateIndex = -1
		end
	elseif ( numGroupMembers ~= self._numGroupMembers ) then
		-- number of group members changed, so restart the buff checks if we're beyond the number of group members
		HoJ:DebugMsg( "Resetting update due to group size change!" )
		local max = HoJ:MaxMemberForGroupType( self._groupType )
		for spellId, blessing in pairs( blessings ) do
			if ( blessing.updateIndex >= max ) then
				blessing.updateIndex = -1
			end
		end
	end

	self._groupType = groupType
	self._numGroupMembers = numGroupMembers

	for spellId, blessing in pairs( blessings ) do
		if ( blessing.targetName ~= nil ) then
			--HoJ:DebugMsg( "Blessing " .. blessing.spellInfo.name .. " targets " .. blessing.targetName )
			if ( not HoJ:UpdateBlessingFromUnitID( blessing, "player" ) ) then
				if ( not HoJ:UpdateBlessingFromUnitID( blessing, "target" ) ) then
					if ( HoJ._groupType ~= HoJ.GROUP_NONE ) then

						local unitID = HoJ:UnitIDForName( blessing.targetName )
						if ( unitID ~= nil ) then
							if ( HoJ:UpdateBlessingFromUnitID( blessing, unitID ) ) then
								found = true
								break
							end
						end
					end
				end
			end
			-- check for expiration
			local expireTime = blessing.start + blessing.duration
			local timeRemaining = expireTime - self._curTime
			if ( timeRemaining <= 0.0 ) then
				HoJ:ResetBlessing( blessing )
			end
		else
			HoJ:FindBlessing( blessing )
		end
	end
	HoJ._nextUpdateBlessingFrame = self._updateFrameNum + self._updateBlessingsInterval
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckBlessings
-- -----------------------------------------------------------------------------
function HoJ:CheckBlessings()
	if ( not HoJ._showBuffsFrame ) then
		return
	end

	-- this has to be called every time we check blessings to update non-party member blessings
	if ( self._updateFrameNum >= self._nextUpdateBlessingFrame ) then
		HoJ:UpdateBlessings( HoJ._retActiveBlessingsList )
	end

	--HoJ:DebugMsg( "CheckBlessings )

	-- dont show any blessing buffs for a second after entering world
	if ( self._updateFrameNum < self._enteredWorldUpdateFrameNum + HoJ.UPDATES_PER_SEC ) then
		return
	end


	if ( HoJ._level >= HoJ.SpellInfo.blessingOfKings.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.blessingOfKings.id ) ) then
			local kingsBlessing = HoJ._retActiveBlessingsList[HoJ.SpellInfo.blessingOfKings.id]
			if ( kingsBlessing.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Kings icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfKings )
			end
		end
	end

	if ( HoJ._level >= HoJ.SpellInfo.blessingOfWisdom.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.blessingOfWisdom.id ) ) then
			local wisdomBlessing = HoJ._retActiveBlessingsList[HoJ.SpellInfo.blessingOfWisdom.id]
			if ( wisdomBlessing.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Wisdom icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfWisdom )
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckSelfHealing
-- -----------------------------------------------------------------------------
function HoJ:CheckSelfHealing()
	--if ( not self._inCombat or self._targetCanAttack ) then
	--	return
	--end

	local forbearanceActive = self:IsBuffActive( self.SpellInfo.forbearance.id )
	local emergencyLoH = false
	if ( not forbearanceActive
		and ( self._playerHealthRatio <= 0.2 or ( self._playerHealthRatio <= 0.5 and self._damagePerSecRatio >= 0.2 ) ) ) then
		self:AddIcon( self._cooldowns, self.SpellInfo.layOnHands )
		emergencyLoH = true
	end

	local shouldHeal = self._playerHealthRatio <= 0.6 or ( self._playerHealthRatio <= 0.8 and self._damagePerSecRatio > 0.1 )

	if ( shouldHeal ) then
		-- selfless healer
		if ( self:IsTalentSelected( self.SpellInfo.selflessHealer ) and self:IsOffCooldown( self.SpellInfo.selflessHealer ) ) then
			local name = nil
			local icon = nil
			local count = 0

			name, icon, count = self:UnitBuffForName( "player", self.SpellInfo.selflessHealer.name )
			if ( name ~= nil ) then
				if ( count >= self.SpellInfo.selflessHealer.maxBuffCount or self._playerHealthRatio <= 0.5 ) then
					self:AddIcon( self._cooldowns, self.SpellInfo.selflessHealer, string.format("%i", count ) )
					return
				end
			end
		elseif ( self:GetHolyPower() >= 3 and self:IsTalentSelected( self.SpellInfo.wordOfGlory ) and self:IsOffCooldown( self.SpellInfo.wordOfGlory ) ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.wordOfGlory )
			return
		elseif ( self:GetHolyPower() >= 5 and HoJ:IsTalentSelected( HoJ.SpellInfo.justicarsVengeance )
				and self:IsOffCooldown( self.SpellInfo.justicarsVengeance ) ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.justicarsVengeance )
			return
		end

		-- if we get here we need to heal but either we did not have any Selfless Healer stacks or we don't
		-- have enough holy power for Word of Glory or Justicar's Vengeance
		if ( self:IsOffCooldown( self.SpellInfo.flashOfLight ) ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.flashOfLight )
		elseif ( not emergencyLoH
				and not forbearanceActive
				and ( self._damagePerSecRatio * 2.0 >= self._playerHealthRatio or self._playerHealthRatio <= 0.3 )
				and self:IsOffCooldown( self.SpellInfo.layOnHands ) ) then
			-- if we have no other healing options and we're going to die within 2 seconds or we have less than 30% health, Lay on Hands
			self:AddIcon( self._cooldowns, self.SpellInfo.layOnHands )
		elseif ( not forbearanceActive and self:IsOffCooldown( self.SpellInfo.divineShield ) ) then
			-- we choose Divine Shield first if off cooldown because we don't currently know if we were taking only physical damage
			self:AddIcon( self._cooldowns, self.SpellInfo.divineShield )
		elseif ( self:IsOffCooldown( self.SpellInfo.blessingOfProtection ) ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.blessingOfProtection )
		end
	end
end

-- -----------------------------------------------------------------------------
-- Data-driven Skill Selection
-- -----------------------------------------------------------------------------
local HOJ_OP_NONE = -999

-- boolean comparison operators
local HOJ_OP_LESS_EQUAL = -2
local HOJ_OP_LESS = -1
local HOJ_OP_EQUAL = 0
local HOJ_OP_GREATER = 1
local HOJ_OP_GREATER_EQUAL = 2

-- logic operators
local HOJ_OP_AND = 1000
local HOJ_OP_OR = 1001

-- range operator
local HOJ_OP_IN_RANGE_INCLUSIVE = 1002 -- requires a const array { min, max }, true if >= min and <= max
local HOJ_OP_IN_RANGE_MIN_EX_MAX_IN = 1003 -- requires a const array { min, max }, true if > min and <= max
local HOJ_OP_IN_RANGE_MIN_IN_MAX_EX = 1004 -- requires a const array { min, max }, true if >= min and < max
local HOJ_OP_IN_RANGE_MIN_EX_MAX_EX = 1005 -- requires a const array { min, max }, true if > min and < max

-- TODO: need a way to group operations

-- constant types
local HOJ_TYPE_CONSTANT = 0 	-- use the constant specified by aValue or bValue

-- result types
local HOJ_TYPE_RESULT = 1

-- variable types
local HOJ_TYPE_VARIABLE_START					= 1000
local HOJ_TYPE_HOLY_POWER 						= HOJ_TYPE_VARIABLE_START 	-- get player's current holy power
local HOJ_TYPE_PLAYER_HEALTH_RATIO				= 1001	-- get the target's health ratio
local HOJ_TYPE_TARGET_HEALTH_RATIO				= 1002	-- get the target's health ratio
local HOJ_TYPE_BUFF_ACTIVE						= 1003	-- true if the buff specified by the const parm is active
local HOJ_TYPE_TARGET_RANGE						= 1004	-- get the target's range with const as the range type
local HOJ_TYPE_TALENT_SELECTED					= 1005 	-- true if the talent specified by the const parm (spellInfo) is selected
local HOJ_TYPE_TARGET_PLAYER_DEBUFF_REMAINING	= 1006	-- time remaining for a debuff applied by the player
local HOJ_TYPE_PLAYER_BUFF_REMAINING			= 1007	-- time remaining for a buff on the player

-- -----------------------------------------------------------------------------
-- HoJ:BooleanOp
-- compares two values using the specified operation constant
-- returns true or false
-- -----------------------------------------------------------------------------
function HoJ:BooleanOp( op, a, b )
	if ( op == HOJ_OP_LESS_EQUAL ) then
		return a <= b
	elseif ( op == HOJ_OP_LESS ) then
		return a < b
	elseif ( op == HOJ_OP_EQUAL ) then
		return a == b
	elseif ( op == HOJ_OP_GREATER ) then
		return a > b
	elseif ( op == HOJ_OP_GREATER_EQUAL ) then
		return a >= b
	elseif ( op == HOJ_OP_AND ) then
		return a and b
	elseif ( op == HOJ_OP_OR ) then
		return a or b
	elseif ( op == HOJ_OP_IN_RANGE_INCLUSIVE ) then
		return a >= b[1] and a <= b[2]
	elseif ( op == HOJ_OP_IN_RANGE_MIN_EX_MAX_IN ) then
		return a > b[1] and a <= b[2]
	elseif ( op == HOJ_OP_IN_RANGE_MIN_EX_MAX_IN ) then
		return a >= b[1] and a < b[2]
	elseif ( op == HOJ_OP_IN_RANGE_MIN_EX_MAX_EX ) then
		return a > b[1] and a < b[2]
	else
		self:DebugMsg( "Unknown op: " .. op )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetValue
-- returns a value based on the type and/or constant
-- -----------------------------------------------------------------------------
function HoJ:GetValue( type, constant, results )
	if ( type == HOJ_TYPE_CONSTANT ) then
		return constant
	elseif ( type >= HOJ_TYPE_VARIABLE_START ) then
		if ( type == HOJ_TYPE_HOLY_POWER ) then
			return self:GetHolyPower()
		elseif ( type == HOJ_TYPE_PLAYER_HEALTH_RATIO ) then
			return self._playerHealthRatio
		elseif ( type == HOJ_TYPE_TARGET_HEALTH_RATIO ) then
			return self._targetHealthRatio
		elseif ( type == HOJ_TYPE_BUFF_ACTIVE ) then
			return self:IsBuffActive( constant )
		elseif ( type == HOJ_TYPE_TARGET_RANGE ) then
			return self._targetInRange[constant]
		elseif ( type == HOJ_TYPE_TALENT_SELECTED ) then
			return self:IsTalentSelected( constant )
		elseif( type == HOJ_TYPE_TARGET_PLAYER_DEBUFF_REMAINING ) then
			return self:GetUnitBuffTimeRemaining( "target", constant, "PLAYER|HARMFUL" )
		elseif( type == HOJ_TYPE_PLAYER_BUFF_REMAINING ) then
			return self:GetBuffTimeRemaining( constant )
		else
			self:DebugMsg( "Unknown variable type: " .. type )
		end
	elseif ( type == HOJ_TYPE_RESULT ) then
		if ( results == nil ) then
			self:DebugMsg( "type is a result but results list == nil!" )
			return 0
		elseif ( constant == nil ) then
			self:DebugMsg( "type is a result but constant == nil! constant  must specify a result index!" )
		else
			return results[constant]
		end
	else
		self:DebugMsg( "Unknown value type: " .. type )
		return 0
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Compare
-- compares two values specified by a type constant and or a value using the
-- specified operation and returns a bool
-- -----------------------------------------------------------------------------
function HoJ:Compare( aType, aConst, bType, bConst, op, results )
	local a = self:GetValue( aType, aConst, results )
	local b = self:GetValue( bType, bConst, results )
	return self:BooleanOp( op, a, b )
end

-- -----------------------------------------------------------------------------
-- HoJ:DoRotSkillTest
function HoJ:DoRotSkillTest( test, results )
	if ( test == nil ) then
		return true 	-- no test, so always pass
	end
	local result = self:Compare( test.aType, test.aConst, test.bType, test.bConst, test.op, results )
	return result
end

-- -----------------------------------------------------------------------------
-- HoJ:DoRotSkillTests
-- runs the encoded tests and returns true or false
-- -----------------------------------------------------------------------------
function HoJ:DoRotSkillTests( rotSkill )
	if ( rotSkill.tests == nil ) then
		self:DebugMsg( "rotSkill.tests == nil" )
		return true
	end

	if ( C_SpellBook.IsSpellDisabled( rotSkill.spellInfo.id ) ) then
		return false
	end

	local debug = ( rotSkill.debugName ~= nil );

	if ( debug ) then
		self:DebugMsg( "rotSkill = " .. rotSkill.spellInfo.name )
	end

	local results = { }

	for i, test in ipairs( rotSkill.tests ) do
		local result = self:DoRotSkillTest( test, results )
		if ( debug ) then
			self:DebugMsg( "test[" .. i .. "] = " .. self:BoolToStr( result ) )
		end
		results[i] = result
	end
	-- return the last result
	local lastResult = results[#results]
	if ( lastResult == nil ) then
		self:DebugMsg( "lastResult = nil!" )
	end
	return lastResult
end

--[[
test = {
	aType	-- type of the first argument
	aConst	-- constant value of the first argument
	bType	-- type of the second argument
	bConst	-- constant value of the second argument
	op 		-- the operator to use ( a op b )
}
]]

HoJ._retRotationRules = {
	{
		spellInfo = HoJ.SpellInfo.inquisition,
		--debugName = "inquisition",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 2, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_PLAYER_BUFF_REMAINING, aConst = HoJ.SpellInfo.inquisition.id, bType = HOJ_TYPE_CONSTANT, bConst = { 0.0, 2.0 }, op = HOJ_OP_IN_RANGE_MIN_EX_MAX_IN },
			--{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.inquisition.id, bType = HOJ_TYPE_CONSTANT, bConst = false, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.executionSentence,
		--debugName = "executionSentence",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 3, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.executionSentence.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	-- if we have Divine Purpose up and it's about to run out, use it
	{
		spellInfo = HoJ.SpellInfo.templarsVerdict,
		--debugName = "TV",
		tests = {
			{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.divinePurposeBuff.id, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_PLAYER_BUFF_REMAINING, aConst = HoJ.SpellInfo.divinePurposeBuff.id, bType = HOJ_TYPE_CONSTANT, bConst = { 0.0, 2.0 }, op = HOJ_OP_IN_RANGE_MIN_EX_MAX_IN },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.templarsVerdict.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 4, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.wakeOfAshes,
		-- debugName = "wakeOfAshes",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 1, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.wakeOfAshes.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.empyreanPower,
		--debugName = "empyreanPower",
		tests = {
			{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.empyreanPowerBuff.id, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.empyreanPower.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	-- If judgment is off cooldown, we have 4 or 5 holy power, we're in judgment range, and we don't have judgment on our target,
	-- cast judgment so we get the 25% damage bonus when we use TV next.
	-- If already at 5 HoPo, this wastes 1 HoPo to get the 25% increase and its use is debatable -- I'm not sure what that 1 wasted
	-- HoPo amounts to in DPS.
	-- At 4 HoPo, this immediately applies Judgment to bring us to 5 HoPo where the next skill will be TV/DS
	-- Without this rule and TV == 4 we would normally skip TV since HoPo < 5, but use HoW since it will give us 4 + 1 == 5 TV
	-- If judgment is off cooldown, then that wastes an opportunity to get the 25% damage bonus on the next HoPo spender
	-- We could alternatively modify the HoW test to not fire if Judgment is off CD, or just prioritize Judgement over HoW, which some sites do.
	-- It's tempting to just move HoW priority after BoJ and Judgment, but it does have a higher overall DPS when it's up.
	-- If were were selecting this manually it would be much harder to coordinate when to priorities a 4-HoPo Judgement vs a 4-HoPo HoW
	-- so it might still make sense to just have HoW the higher priority.
	{
		spellInfo = HoJ.SpellInfo.judgment,
		--debugName = "judgment_before_TV",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.judgment.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND },
			{ aType = HOJ_TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = HoJ.SpellInfo.judgmentDebuff.id, bType = HOJ_TYPE_CONSTANT, bConst = 0.0, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 4, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.templarsVerdict,
		--debugName = "TV",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 5, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.templarsVerdict.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	-- also use TV if we have Judgment on the target that is about to expire
	{
		spellInfo = HoJ.SpellInfo.templarsVerdict,
		--debugName = "TV_judgment_bonus_expiring",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 3, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.templarsVerdict.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND },
			{ aType = HOJ_TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = HoJ.SpellInfo.judgmentDebuff.id, bType = HOJ_TYPE_CONSTANT, bConst = { 0.0, 2.0 }, op = HOJ_OP_IN_RANGE_MIN_EX_MAX_IN },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 4, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.hammerOfWrath,
		--debugName = "hammerOfWrath",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.hammerOfWrath.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND },
			{ aType = HOJ_TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 0.2, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 4, op = HOJ_OP_AND },

		}
	},
	{
		spellInfo = HoJ.SpellInfo.hammerOfWrath,
		--debugName = "hammerOfWrath_Crusade",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.hammerOfWrath.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND },
			{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.crusade.id, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.avengingWrath.id, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 4, bType = HOJ_TYPE_RESULT, bConst = 5, op = HOJ_OP_OR },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 6, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.bladeOfJustice,
		--debugName = "bladeOfJustice",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 3, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.bladeOfJustice.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.judgment,
		--debugName = "judgment",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.judgment.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.retConsecration,
		--debugName = "retConsecration",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_LESS_EQUAL }
		}
	},
	{
		spellInfo = HoJ.SpellInfo.crusaderStrike,
		--debugName = "crusaderStrike",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 4, op = HOJ_OP_LESS_EQUAL },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.crusaderStrike.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_AND }
		}
	},
	-- if no HoPo generators are available and we have >= 3 Holy Power or Divine Purpose, use TV / DS
	{
		spellInfo = HoJ.SpellInfo.templarsVerdict,
		--debugName = "templarsVerdict_DivinePurpose",
		tests = {
			{ aType = HOJ_TYPE_HOLY_POWER, aConst = 0, bType = HOJ_TYPE_CONSTANT, bConst = 3, op = HOJ_OP_GREATER_EQUAL },
			{ aType = HOJ_TYPE_BUFF_ACTIVE, aConst = HoJ.SpellInfo.divinePurpose.id, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 1, bType = HOJ_TYPE_RESULT, bConst = 2, op = HOJ_OP_OR },
			{ aType = HOJ_TYPE_TARGET_RANGE, aConst = HoJ.SpellInfo.templarsVerdict.range, bType = HOJ_TYPE_CONSTANT, bConst = true, op = HOJ_OP_EQUAL },
			{ aType = HOJ_TYPE_RESULT, aConst = 3, bType = HOJ_TYPE_RESULT, bConst = 4, op = HOJ_OP_AND }
		}
	},
}

-- -----------------------------------------------------------------------------
-- HoJ:RetNextRotationSkill
-- Takes a list of skills with the following structure:
-- {
--		spellInfo,				- the spell info
--		replacedByTalent		- if not nil, spell isn't usable if this talent is selected
--		test = { }				- test structure specified above
-- }
-- -----------------------------------------------------------------------------
function HoJ:RetNextRotationSkill( rotationSelectionList )
	local isTalentSelected = self.IsTalentSelected
	local isOffCooldown = self.IsOffCooldown
	local doRotSkillTests = self.DoRotSkillTests

	for i, v in ipairs( rotationSelectionList ) do
--		local debug = ( v.debugName ~= nil ) or ( HoJ._debugSpellName ~= nil and v.name == HoJ._debugSpellName );
--		if ( debug ) then
--			self:DebugMsg( "rotSkill.spellInfo = " .. v.spellInfo.name ) -- .. ".tests = " .. v.tests )
--		end
		if ( v.replacedByTalent == nil or not isTalentSelected( self, v.replacedByTalent ) ) then
--			if ( debug and v.replacedByTalent ~= nil ) then
--				self:DebugMsg( "Talent " .. v.debugName .. " is not replaced by " .. v.replacedByTalent.name )
--			end
			if ( ( v.spellInfo.spellType ~= HoJ.SPELL_TALENT and v.spellInfo.spellType ~= HoJ.SPELL_TALENT_BUFF )
				or isTalentSelected( self, v.spellInfo ) ) then
--				if ( debug ) then
--					self:DebugMsg( "Talent " .. v.debugName .. " is selected." )
--				end
				if ( isOffCooldown( self, v.spellInfo ) ) then
--					if ( debug ) then
--						self:DebugMsg( "Talent " .. v.debugName .. " is of cooldown." )
--					end
					if ( doRotSkillTests( self, v ) ) then
--						if ( v.debugName ~= nil ) then
--							self:DebugMsg( "(" .. HoJ._frameNum .. "):RotSkill " .. v.debugName .. " selected." )
--						end
						return v.spellInfo
					end
				end
			end
		end
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:RetNextSkillOffCooldown
-- -----------------------------------------------------------------------------
function HoJ:RetNextSkillOffCooldown( rotationSelectionList )
	local indexOfNextSkill = -1
	local isTalentSelected = self.IsTalentSelected
	local doRotSkillTests = self.DoRotSkillTests
	local getSpellCooldown = GetSpellCooldown

	local shortestTime = 999999.0
	for i, v in ipairs( rotationSelectionList ) do
		if ( v.replacedByTalent == nil or not isTalentSelected( self, v.replacedByTalent ) ) then
			if ( ( v.spellInfo.spellType ~= HoJ.SPELL_TALENT and v.spellInfo.spellType ~= HoJ.SPELL_TALENT_BUFF )
				or isTalentSelected( self, v.spellInfo ) ) then
				if ( doRotSkillTests( self, v ) ) then
					local start, duration, enabled = getSpellCooldown( v.spellInfo.id )
					local timeLeft = ( start + duration ) - HoJ._curTime
					if ( timeLeft < shortestTime ) then
						indexOfNextSkill = i
						shortestTime = timeLeft
					end
				end
			end
		end
	end
	if ( indexOfNextSkill < 0 ) then
		return nil
	end
	return rotationSelectionList[indexOfNextSkill].spellInfo
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation_Ret_8dot1
-- checks the DPS rotation twice, once for the current time, and if that fails
-- once for a future time
-- -----------------------------------------------------------------------------
function HoJ:DoRetDPSRotation_8dot1( skillsList, rotationSelectionList )
	local nextSkill = self:RetNextRotationSkill( rotationSelectionList )
	-- if nextSkill is nil, then we can find the next skill off cooldown and use that
	if ( nextSkill == nil ) then
		nextSkill = self:RetNextSkillOffCooldown( rotationSelectionList )
	end
	if ( nextSkill ~= nil ) then
		self:AddIcon( skillsList, nextSkill )
	end
end


-- -----------------------------------------------------------------------------
-- HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes
-- -----------------------------------------------------------------------------
function HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( predictAhead, curHoPo, maxHoPo )
	--HoJ:DebugMsg( "GetNextHoPoGeneratorWithoutWakeOfAshes" );

	if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.bladeOfJustice, predictAhead, curHoPo, maxHoPo ) ) then
		return HoJ.SpellInfo.bladeOfJustice;
	end

	if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.judgment, predictAhead, curHoPo, maxHoPo ) ) then
		return HoJ.SpellInfo.judgment;
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.hammerOfWrath ) ) then
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.hammerOfWrath, predictAhead, curHoPo, maxHoPo ) ) then
			if ( HoJ._targetHealthRatio <= 0.20
				or HoJ:IsBuffActive( self.SpellInfo.crusade.id )
				or HoJ:IsBuffActive( self.SpellInfo.avengingWrath.id ) ) then
				return HoJ.SpellInfo.hammerOfWrath;
			end
		end
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.retConsecration ) ) then
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.retConsecration, predictAhead, curHoPo, maxHoPo ) ) then
			return HoJ.SpellInfo.retConsecration;
		end
	end

	-- do not use crusader strike if we have >= 3 holy power and the charges are < 2
	currentCharges = GetSpellCharges( HoJ.SpellInfo.crusaderStrike.id );
	if ( HoJ:GetHolyPower() < 3 or currentCharges > 1 ) then
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.crusaderStrike, predictAhead, curHoPo, maxHoPo ) ) then
			return HoJ.SpellInfo.crusaderStrike;
		end
	end

	--HoJ:DebugMsg( "No HoPo gen available" );
	return nil;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetNextHoPoGenerator
-- -----------------------------------------------------------------------------
function HoJ:GetNextHoPoGenerator( predictAhead, curHoPo )
	--HoJ:DebugMsg( "GetNextHoPoGenerator" );

	local wakeOfAshesSelected = HoJ:IsTalentSelected( HoJ.SpellInfo.wakeOfAshes );
	local wakeOfAshesUsable = HoJ:IsSkillUsable( HoJ.SpellInfo.wakeOfAshes, predictAhead );
	local amountOfWakeOfAshesOverflow = HoJ:WillOverflowHoPo( HoJ.SpellInfo.wakeOfAshes, curHoPo, HoJ.MAX_HOLY_POWER );
	-- HoJ:DebugMsg( "Wake sel = " .. HoJ:BoolToStr( wakeOfAshesSelected ) .. ", use = " .. HoJ:BoolToStr( wakeOfAshesUsable ) .. ", over = " .. amountOfWakeOfAshesOverflow );
	if ( wakeOfAshesSelected ) then
		if ( wakeOfAshesUsable and amountOfWakeOfAshesOverflow <= 0 ) then
			return HoJ.SpellInfo.wakeOfAshes;
		end
	end

	-- if Wake of Ashes is usable but will overflow, then, if possible, choose a skill that will
	-- take us to exactly 3 HoPo so we can use it, then use Wake of Ashes without overflow next.
	local nextSkill = nil;
	if ( wakeOfAshesSelected and wakeOfAshesUsable ) then
		-- if we have exactly 3 HoPo then don't choose a skill here, which will result in the
		-- next chosen skill being Templar's Verdict or Divine Storm, allowing us to use Wake
		-- next without any HoPo lost
		if ( curHoPo == 3 ) then
			return nil;
		end
		-- try to get a skill that won't overflow 3 HoPo
		nextSkill = HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( predictAhead, curHoPo, 3 );
		if ( nextSkill ~= nil ) then
			--HoJ:DebugMsg( "Gen: " .. nextSkill.name .. " (without Wake)." );
			return nextSkill;
		end
	end

	nextSkill = HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( predictAhead, curHoPo, HoJ.MAX_HOLY_POWER );
--	if ( nextSkill ~= nil ) then
--		HoJ:DebugMsg( "Gen: " .. nextSkill.name );
--	end
	return nextSkill;
end

-- -----------------------------------------------------------------------------
-- HoJ:ChooseHolyPowerSkill
-- -----------------------------------------------------------------------------
function HoJ:ChooseHolyPowerSkill( skillsList )
	if ( HoJ:GetHolyPower() + HoJ.SpellInfo.executionSentence.hoPoGen >= 0 ) then
		if ( HoJ:IsTalentSelected( HoJ.SpellInfo.executionSentence ) ) then
			return HoJ.SpellInfo.executionSentence;
		end
	end

	-- TODO: figure out a way to tell if we should use Divine Storm instead
	if ( HoJ:GetHolyPower() + HoJ.SpellInfo.templarsVerdict.hoPoGen >= 0 ) then
		return HoJ.SpellInfo.templarsVerdict;
	end

	return nil;
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation_Ret_8dot1_Internal

-- DPS rotation for Retribution
--[[
Based on Icy-Veins rotation as of 12/27/2018:

The single target rotation consists of alternating Holy Power generation and Holy Power expenditure.

We begin with the priority of Holy Power generation, which you should utilize until you are either at 5 Holy Power, or until the next generator you use would put you over 5 Holy Power (thus wasting the excess).

1. Cast Wake of Ashes.
2. Cast Blade of Justice.
3. Cast Judgment.
4. Cast Hammer of Wrath.
5. Cast Consecration.
6. Cast Crusader Strike at 2 charges.
7. Cast Crusader Strike at 1 charge.
8. Cast Arcane Torrent (if Blood Elf).

To spend your Holy Power, use the following priority.

1. Cast Inquisition with at least 2 Holy Power.
2. Cast Execution Sentence.
3. Cast Templar's Verdict.
4. Cast Divine Storm (Empyrean Power).
It is important to note that, when generating Holy Power, if you have only 1 Crusader Strike charge available, you should immediately stop generating Holy Power and go to the Holy Power-spending priority. Stop immediately if you have 2 Crusader Strike charges available and you are using the ammer of Wrath talent, and one of the following two conditions is true:

1. Avenging Wrath is active;
2. Your target has less than 20% Health.

Always cast Avenging Wrath on cooldown.

--]]
-- -----------------------------------------------------------------------------
function HoJ:CheckDPSRotation_Ret_8dot1_Internal( skillsList, predictAhead )
	HoJ._useHolyPower = false;

	-- check if we should renew inquisition
	if ( HoJ:GetHolyPower() >= HoJ._minHolyPowerForInquisition and HoJ._inqState <= HOJ_INQ_NEARLY_EXPIRED ) then
		local _, _, _, inquisitionSelected, inquisitionAvailable = GetTalentInfo( 7, 3, 1 );
		if ( inquisitionAvailable and inquisitionSelected ) then
			--HoJ:DebugMsg( "Renew Inquisition" );
			HoJ:AddIcon( skillsList, HoJ.SpellInfo.inquisition );
			return true;
		end
	end

	-- if empyrean power buff is up and the target is in range, always use it first
	if ( HoJ._targetInRange[HoJ.SpellInfo.empyreanPower.range] ) then
		local epRemaining, epStart, epDuration = HoJ:GetBuffTimeRemaining( HoJ.SpellInfo.empyreanPower.id );
		if ( epRemaining > 0.05 ) then
			--HoJ:DebugMsg( "Use Empyrean Power" );
			HoJ:AddIcon( skillsList, HoJ.SpellInfo.empyreanPower );
			return true;
		end
	end

	if ( self:GetHolyPower() >= 5 ) then
		-- if Holy Power is maxed, don't bother looking for a generator -- this also includes Divine Purpose being active
		local holyPowerSkill = HoJ:ChooseHolyPowerSkill( skillsList );
		if ( holyPowerSkill ~= nil ) then
			HoJ:AddIcon( skillsList, holyPowerSkill )
			return true
		end
	end

	-- check for the next skill that is off cooldown AND won't overflow our holy power
	local nextHoPoGenerator = HoJ:GetNextHoPoGenerator( predictAhead, HoJ:GetHolyPower() );
	if ( nextHoPoGenerator == nil ) then
		-- we didn't find a skill to use, so we should try to use a holy power skill
		local holyPowerSkill = HoJ:ChooseHolyPowerSkill( skillsList );
		if ( holyPowerSkill ~= nil ) then
			--HoJ:DebugMsg( "Use Holy Power" );
			HoJ:AddIcon( skillsList, holyPowerSkill );
			HoJ._useHolyPower = true;
			return true;
		end

		return false;
	else
		--HoJ:DebugMsg( "Use generator " .. nextHoPoGenerator.name );
		HoJ:AddIconDebug( skillsList, nextHoPoGenerator, "4" );
		return true;
	end
end

HoJ._retRotationSkills = {
	[1] = { spellInfo = HoJ.SpellInfo.wakeOfAshes },
	[2] = { spellInfo = HoJ.SpellInfo.bladeOfJustice },
	[3] = { spellInfo = HoJ.SpellInfo.judgment },
	[4] = { spellInfo = HoJ.SpellInfo.hammerOfWrath, minTargetHealthRatio = 0.2 },
	[5] = { spellInfo = HoJ.SpellInfo.retConsecration },
	[6] = { spellInfo = HoJ.SpellInfo.crusaderStrike }
}

HoJ._protRotationSkills = {
	[1] = { spellInfo = HoJ.SpellInfo.consecration },
	[2] = { spellInfo = HoJ.SpellInfo.judgment },
	[3] = { spellInfo = HoJ.SpellInfo.avengersShield },
	[4] = { spellInfo = HoJ.SpellInfo.blessedHammer },
	[5] = { spellInfo = HoJ.SpellInfo.hammerOfTheRighteous, replacedByTalent = HoJ.SpellInfo.blessedHammer }
}


-- -----------------------------------------------------------------------------
-- HoJ:FindNextSkillOffCooldown
-- Takes a list of skills with the following structure:
-- {
--		spellInfo,				- the spell info
--		replacedByTalent		- if not nil, spell isn't usable if this talent is selected
--		maxTargetHealthRatio	- if not nil, target health must be >= this to use
-- }
-- -----------------------------------------------------------------------------
function HoJ:FindNextSkillOffCooldown( skillsList )
	local isTalentSelected = self.IsTalentSelected
	local doRotSkillTests = self.DoRotSkillTests
	local getSpellCooldown = GetSpellCooldown

	local indexOfNextSkill = -1
	local shortestTime = 999999.0
	for i, v in ipairs( skillsList ) do
		if ( v.replacedByTalent == nil or not isTalentSelected( self, v.replacedByTalent ) ) then
			if ( v.spellInfo.spellType ~= HoJ.SPELL_TALENT or isTalentSelected( self, v.spellInfo ) ) then
				if ( v.minTargetHealthRatio == nil or HoJ._targetHealthRatio <= v.minTargetHealthRatio ) then
					local start, duration, enabled = getSpellCooldown( v.spellInfo.id )
					local timeLeft = ( start + duration ) - HoJ._curTime
					if ( timeLeft < shortestTime ) then
						indexOfNextSkill = i
						shortestTime = timeLeft
					end
				end
			end
		end
	end
	if ( indexOfNextSkill < 0 ) then
		return nil
	end
	return skillsList[indexOfNextSkill].spellInfo
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation_Ret_8dot1
-- checks the DPS rotation twice, once for the current time, and if that fails
-- once for a future time
-- -----------------------------------------------------------------------------
function HoJ:CheckDPSRotation_Ret_8dot1( skillsList )
	if ( not HoJ._showRotationFrame ) then
		return false;
	end

	if ( not HoJ._inCombat and not HoJ._targetCanAttack ) then
		-- HoJ:DebugMsg( "not in combat or target can't attack" );
		return false;
	end
--[[
	if ( not self:CheckDPSRotation_Ret_8dot1_Internal( skillsList, 0.0 ) ) then
		if ( not self:CheckDPSRotation_Ret_8dot1_Internal( skillsList, HoJ._gcdRemaining + HoJ.SKILL_USE_PREDICTION_TIME ) ) then
			local nextSkill = self:FindNextSkillOffCooldown( self._retRotationSkills )
			if ( nextSkill ~= nil ) then
				HoJ:AddIcon( skillsList, nextSkill )
				-- HoJ:DebugMsg( "Chose next skill of CD: " .. nextSkill.name )
			end
		end
	end
]]
	self:DoRetDPSRotation_8dot1( skillsList, self._retRotationRules )
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckTankRotation_8_dot_1_Internal
-- -----------------------------------------------------------------------------
function HoJ:CheckTankRotation_8_dot_1_Internal( skillsList, predictAhead )
	local hasConsecrationBuff = HoJ:IsBuffActive( HoJ.SpellInfo.protConsecrationBuff.id )
	local consecrationCdRemaining = HoJ:GetCDRemainingAfterGCD( HoJ.SpellInfo.consecration.id )

	if ( not hasConsecrationBuff and consecrationCdRemaining <= 0 ) then
		HoJ:AddIcon( skillsList, HoJ.SpellInfo.consecration )
		return true;
	end

	if ( HoJ:IsSkillUsable( HoJ.SpellInfo.judgment, predictAhead ) ) then
		HoJ:AddIcon( skillsList, HoJ.SpellInfo.judgment )
		return true
	end

	if ( HoJ:IsSkillUsable( HoJ.SpellInfo.avengersShield, predictAhead ) ) then
		HoJ:AddIcon( skillsList, HoJ.SpellInfo.avengersShield )
		return true;
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.blessedHammer ) ) then
		if ( HoJ:IsSkillUsable( HoJ.SpellInfo.blessedHammer, predictAhead ) ) then
			HoJ:AddIcon( skillsList, HoJ.SpellInfo.blessedHammer )
			return true
		end
	elseif ( HoJ:IsSkillUsable( HoJ.SpellInfo.hammerOfTheRighteous, predictAhead ) ) then
		HoJ:AddIcon( skillsList, HoJ.SpellInfo.hammerOfTheRighteous )
		return true
	end

	if ( consecrationCdRemaining <= 0 ) then
		HoJ:AddIcon( skillsList, HoJ.SpellInfo.consecration )
		return true
	end

	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckTankRotation_8_dot_1
-- checks the rotation twice, once for the current time, and if that fails
-- once for a future time
-- -----------------------------------------------------------------------------
function HoJ:CheckTankRotation_8_dot_1( skillsList )
	if ( not HoJ._showRotationFrame ) then
		return
	end

	if ( not HoJ._inCombat and not HoJ._targetCanAttack ) then
		return
	end

	if ( not self:CheckTankRotation_8_dot_1_Internal( skillsList, 0.0 ) ) then
		if ( not self:CheckTankRotation_8_dot_1_Internal( skillsList, HoJ._gcdRemaining + HoJ.SKILL_USE_PREDICTION_TIME ) ) then
			local nextSkill = self:FindNextSkillOffCooldown( self._protRotationSkills )
			if ( nextSkill ~= nil ) then
				HoJ:AddIcon( skillsList, nextSkill )
				-- HoJ:DebugMsg( "Chose next skill of CD: " .. nextSkill.name )
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckInterrupts
-- -----------------------------------------------------------------------------
function HoJ:CheckInterrupts( skillsList )
	if ( not self._showCooldownsFrame ) then
		return;
	end

	if ( not self._inCombat and not self._targetCanAttack ) then
		return;
	end

	if ( self._targetInRange[self.RANGE_10_YARDS] ) then
		if ( self:IsTalentSelected( self.SpellInfo.blindingLight ) and self:IsOffCooldown( self.SpellInfo.blindingLight, 0.0 ) ) then
			self:AddIconExclusive( skillsList, self.SpellInfo.blindingLight )
		elseif ( not self._targetIsBoss and self:IsOffCooldown( self.SpellInfo.hammerOfJustice, 0.0 ) ) then
			self:AddIconExclusive( skillsList, self.SpellInfo.hammerOfJustice )
		end
	end

	if ( self._targetCanBeInterrupted ) then
		if ( self._targetInRange[HoJ.RANGE_MELEE] and HoJ:IsOffCooldown( self.SpellInfo.rebuke, 0.0 ) ) then
			self:AddIcon( skillsList, self.SpellInfo.rebuke )
		elseif ( self._role == HoJ.ROLE_TANK and self:IsOffCooldown( self.SpellInfo.avengersShield, 0.0 ) ) then
			self:AddIcon( skillsList, self.SpellInfo.avengersShield )
		elseif ( not self._targetInRange[HoJ.RANGE_10_YARDS] and not self._targetIsBoss and HoJ:IsOffCooldown( self.SpellInfo.hammerOfJustice ) ) then
			self:AddIconExclusive( skillsList, self.SpellInfo.hammerOfJustice )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:ADDON_LOADED( ... )
-- -----------------------------------------------------------------------------
function HoJ._events:ADDON_LOADED( arg1 )
	if ( arg1 == "HammerOfJustas" ) then
		HoJ_OnAddonLoaded( self );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LOGOUT( ... )
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_LOGOUT( arg1 )
	HoJ:WriteSavedVariables()
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTERING_WORLD
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_ENTERING_WORLD( ... )
	-- TODO: multiple greetings for the heck of it
	--HoJ:Print( "Welcome, " .. HoJ._playerName .. ". May the light favor your hammer." );

	-- HoJ:DebugMsg( "Entering world: level = " .. UnitLevel( "player" ) )

	HoJ._worldState = HoJ.WORLD_STATE_ENTERED
	HoJ._enteredWorldUpdateFrameNum = HoJ._updateFrameNum
	HoJ:ResetBlessings( HoJ._retActiveBlessingsList )
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LEAVING_WORLD
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_LEAVING_WORLD( ... )
	--HoJ:DebugMsg( "Leaving world: level = " .. UnitLevel( "player" ) )
	HoJ._worldState = HoJ.WORLD_STATE_LEFT
	HoJ._needsReset = true -- reset icons and such
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_TARGET_CHANGED
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_TARGET_CHANGED( ... )
	-- HoJ:DebugMsg( "PLAYER_TARGET_CHANGED" );
	--HoJ:UpdateSacredShieldIcon();
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_THREAT_LIST_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_THREAT_LIST_UPDATE( ... )
	--HoJ:DebugMsg( "UNIT_THREAT_LIST_UPDATE" );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTER_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_DISABLED( ... )
	-- HoJ:DebugMsg( "PLAYER_REGEN_DISABLED" )
	HoJ:EnterCombat( ... );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LEAVE_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_ENABLED( ... )
	-- HoJ:DebugMsg( "PLAYER_REGEN_ENABLED" )
	HoJ:LeaveCombat( ... );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_TALENT_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_TALENT_UPDATE( ... )
	-- HoJ:DebugMsg( "PLAYER_TALENT_UPDATE" );
	-- HoJ._needsReset = true;
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:HOLY_POWER
-- -----------------------------------------------------------------------------
function HoJ._events:HOLY_POWER(...)
	HoJ._holyPower = UnitPower( "player", 9 );
	--HoJ:DebugMsg( HoJ._playerName .. " has " .. HoJ._holyPower .. " holy power." )
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:UNIT_AURA
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_AURA( unitID )
	--HoJ:DebugMsg( "UNIT_AURA for " .. unitID )
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:UNIT_SPELLCAST_SUCCEEDED
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_SPELLCAST_SUCCEEDED( unitID, castGUID, spellID ) -- spell, rank, lineID, spellID )
	local spellName = HoJ:GetSpellName( spellID )
	--HoJ:DebugMsg( "UNIT_SPELLCAST_SUCCEEDED: " .. spellName .. " by " .. unitID )
	if ( unitID == "player" ) then
		local spellTarget = "target"
		if ( not UnitIsPlayer( "target" ) ) then
			spellTarget = "player"
		end
		--HoJ:DebugMsg( "spellTarget = " .. spellTarget )

		local name, realm = UnitName( spellTarget )
--[[
		if ( realm ~= nil ) then
			HoJ:DebugMsg( spellName .. " cast on player " .. name .. "-" .. realm )
		else
			HoJ:DebugMsg( spellName .. " cast on player " .. name )
		end

		HoJ:DebugMsg( "player UNIT_SPELLCAST_SUCCEEDED for " .. spell );
		HoJ:DebugMsg( "spellID = " .. spellID .. ", inquisition.id = " .. HoJ.SpellInfo.inquisition.id );
--]]
		if ( spellID == HoJ.SpellInfo.inquisition.id and HoJ._inCombat ) then
			HoJ:StartInquisitionCooldown()
		else
			local blessing = HoJ._retActiveBlessingsList[spellID]
			if ( blessing ~= nil ) then
				blessing.targetName = name
				blessing.start = HoJ._curTime
				blessing.duration = blessing.spellInfo.buffDurationSeconds
				blessing.found = false
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_HEALTH
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_HEALTH( unit )
--[[
	if ( unit == "player" ) then
		HoJ:DebugMsg( "Player damaged" )
		HoJ._showHealthTime = HoJ._curTime + 0.5
	end
]]
end

-- -----------------------------------------------------------------------------
-- HoJ:SetHPAlphas
-- -----------------------------------------------------------------------------
function HoJ:SetHPAlphas( alpha, numVisible )
	local holyPowerAlpha = alpha * self._holyPowerAlpha * self._globalAlpha
	local templarsVerdictAlpha = alpha * self._holyPowerAlpha * self._globalAlpha

	if ( not HoJ._showHolyPowerFrame ) then
		holyPowerAlpha = 0.0;
	end
	if ( not HoJ._showTemplarsVerdictFrame ) then
		templarsVerdictAlpha = 0.0;
	end

	for i = 0, numVisible - 1 do
		--HoJ:DebugMsg( "Setting alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( holyPowerAlpha );
		HoJ._HoPoGlow[i]:SetAlpha( holyPowerAlpha );
	end

	for i = numVisible, 4 do
		--HoJ:DebugMsg( "Clearing alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( 0.0 );
		HoJ._HoPoGlow[i]:SetAlpha( 0.0 );
	end

	-- set alpha for big graphic
	if ( numVisible < 3 ) then -- or not HoJ._useHolyPower ) then
		HoJFrame.HPMaxed:SetAlpha( 0.0 );
		HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );
	else
		HoJFrame.HPMaxed:SetAlpha( templarsVerdictAlpha );
		HoJFrame.HPMaxedGlow:SetAlpha( templarsVerdictAlpha );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:PredictDeathTime
-- -----------------------------------------------------------------------------
function HoJ:PredictDeathTime( secondsIntoFuture, secondsAlreadyPredicted, futureHealth, nearDeathHealth, damagePerSec, predictedDeathTime, predictedNearDeathTime )
	if ( predictedDeathTime == 0.0 or predictedNearDeathTime == 0.0 ) then
		-- we've already predicted HEALTH_SAMPLE_SECONDS into the future
		local numSeconds = secondsIntoFuture - secondsAlreadyPredicted

		for i = 1, numSeconds do
			local predictedHealth = futureHealth - ( damagePerSec * i )
			-- self:DebugMsg( "Predicting " .. predictedHealth .. " health in " .. ( i + secondsAlreadyPredicted ) )
			if ( predictedHealth < nearDeathHealth ) then
				if ( predictedNearDeathTime == 0.0 ) then
					predictedNearDeathTime = secondsAlreadyPredicted + i
				end
				if ( predictedHealth <= 1.0 and predictedDeathTime == 0.0 ) then
					predictedDeathTime = secondsAlreadyPredicted + i
					return predictedDeathTime, predictedNearDeathTime
				end
			end
		end
	end

	return predictedDeathTime, predictedNearDeathTime
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealthHistory
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealthHistory()
	if ( self._curTime >= self._nextHealthTime ) then
		if ( self._healthBuffer:isFull() ) then
			local h = self._healthBuffer:remove( 0 )
			if ( h == nil ) then
				self:DebugMsg( "ERROR! removed nil health value!" )
			end
		end

		self._healthBuffer:add( { time = HoJ._curTime, health = UnitHealth( "player" ) } )
		local SAMPLE_INTERVAL = ( 1.0 / HoJ.HEALTH_SAMPLES_PER_SEC )
		self._nextHealthTime =  self._curTime + SAMPLE_INTERVAL
		-- self:DebugMsg( "_nextHealthTime = " .. self._nextHealthTime )

		-- local show = self._curTime < self._showHealthTime

		local size = self._healthBuffer.maxSize -- maximum history samples
		if ( self._healthBuffer:count() < size ) then
			-- self:DebugMsg( "Waiting for health history to fill up." )
			return -- don't do any health prediction until the buffer is fill up
		end

		local damage = 0
		local heals = 0
		local largestDamage = 0 -- calculaed as a negative number
		local curHealth = self._playerHealth
		local futureHealth = curHealth
		local futureHealthTime = 0.0
		local predictedDeathTime = 0.0
		local predictedNearDeathTime = 0.0
		local playerMaxHealth = self._playerMaxHealth

		-- local healthBuffer will be a reference to the table only
		local healthBuffer = self._healthBuffer

		-- in the CircularBuffer, buffer[tail] = last item, so we start with an offset of 0 and
		-- we use size - 1 as the max offset
		local tail = self._healthBuffer.tail
		local index = tail % size
		local lastHealth = self._healthBuffer[index].health
		local count = size - 1
		for i = 1, count do
			index = ( tail + i ) % size
			local h = healthBuffer[index].health

			local delta = h - lastHealth
			futureHealth = futureHealth + delta
			lastHealth = h

			if ( delta ~= 0 ) then
				if ( delta < 0 ) then
					damage = damage - delta -- convert to positive
					if ( delta < largestDamage ) then
						largestDamage = delta
					end

					if ( predictedNearDeathTime == 0.0 and futureHealth <= playerMaxHealth * 0.1 ) then
						predictedNearDeathTime = i * SAMPLE_INTERVAL
					elseif ( predictedDeathTime == 0.0 and futureHealth <= 0 ) then
						predictedDeathTime = i * SAMPLE_INTERVAL
					end
				else
					heals = heals + delta
				end
			end
		end

		local SAMPLE_SECONDS = HoJ.HEALTH_SAMPLE_SECONDS
		local scale = 1.0 / SAMPLE_SECONDS
		local damagePerSec = damage * scale
		local healsPerSec = heals * scale
		self._damagePerSec = damagePerSec
		self._healsPerSec = healsPerSec

		self._damagePerSecRatio = damagePerSec / playerMaxHealth
		self._healsPerSecRatio = healsPerSec / playerMaxHealth

		self._largestRecentDamageRatio = largestDamage / self._playerMaxHealth

		self._predictedHealth = futureHealth
		self._predictedHealthTime = futureHealthTime

		if ( damagePerSec > 0 ) then
			self._predictedDeathTime, self._predictedNearDeathTime = HoJ:PredictDeathTime( 8, SAMPLE_SECONDS, futureHealth, playerMaxHealth * 0.1, damagePerSec, predictedDeathTime, predictedNearDeathTime )
		else
			self._predictedDeathTime = 0.0
			self._predictedNearDeathTime = 0.0
		end
--[[
		if ( self._predictedHealthTime > 0 and self._predictedHealth < self._playerHealth ) then
			self:DebugMsg( "FutureHealth = " .. self._predictedHealth .. " @ " .. self._predictedHealthTime )
		end
		if ( self._predictedNearDeathTime > 0.0 ) then
			self:DebugMsg( "Near death predicted in " .. self._predictedNearDeathTime .. " seconds" )
		end
		if ( self._predictedDeathTime > 0.0 ) then
			self:DebugMsg( "Death predicted in " .. self._predictedDeathTime .. " seconds" )
		end
		if ( damage > 0 or heals > 0 ) then
			self:DebugMsg( "Incoming DPS = " .. damage .. ", HPS = " .. heals )
		end
]]
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealth
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealth()
	self._playerHealth = UnitHealth( "player" )
	self._playerMaxHealth = UnitHealthMax( "player" )

	if ( self._playerMaxHealth == nil or self._playerMaxHealth <= 0 ) then
		self._playerMaxHealth = 1
	end

	self._playerHealthRatio = self._playerHealth / self._playerMaxHealth

	ProfileFunction( "UpdateHealthHistory", HoJ.UpdateHealthHistory, self )
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetHealth
-- -----------------------------------------------------------------------------
function HoJ:ResetHealth()
	self._healthBuffer:empty()
	self._predictedHealth = self._playerHealth
	self._predictedHealthTime = 0.0
	self._predictedDeathTime = 0.0
	self._predictedNearDeathTime = 0.0
	self._damagePerSec = 0.0
	self._damagePerSecRatio = 0.0
	self._healsPerSec = 0.0
	self._healsPerSecRatio = 0.0
	self._largestRecentDamageRatio = 0.0
--[[
	self._healthBuffer:add( { time = GetTime(), health = UnitHealth( "player" ) } )
	local h = self._healthBuffer:remove()
	if ( h ~= nil ) then
		self:DebugMsg( "h.time = " .. h.time )
		self:DebugMsg( "h.health = " .. h.health )
	else
		self:DebugMsg( "h == nil!" )
	end
]]
end

-- -----------------------------------------------------------------------------
-- HoJ:MouseOver
-- -----------------------------------------------------------------------------
function HoJ:MouseOver()
	-- At some point in Cata or Mists, EnableMouse became protected in combat.
	--if ( HoJ._inCombat ) then
	--	return;
	--end

	local mouseFocus = GetMouseFocus();
	for i, v in pairs( HoJ._windowFrames ) do
		if ( MouseIsOver( v ) ) then
			if ( IsControlKeyDown() ) then
				-- HoJ:DebugMsg( "Enabling mouse for frame '" .. v:GetName() .. "'" );
				v:EnableMouse( true );

				if ( IsMouseButtonDown( "LeftButton" ) ) then
					-- HoJ:DebugMsg( "Start moving frame '" .. v:GetName() .. "'" );
					v:StartMoving();
				end
			end
			if ( not IsMouseButtonDown( "LeftButton" ) ) then
				v:StopMovingOrSizing();
			end
		else
			--HoJ:DebugMsg( "Disabling mouse for frame '" .. v:GetName() .. "'" );
			v:EnableMouse( false );
			v:StopMovingOrSizing();
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnableFrame
-- -----------------------------------------------------------------------------
function HoJ:EnableFrame( self )
	if ( HoJ._isPaladin ) then
		table.insert( HoJ._windowFrames, self );
		self:Show();
		self:SetMovable( true );
		self:EnableMouse( true );
		self:SetPoint( "CENTER", UIParent );
	else
		self:Hide();
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnEvent
-- -----------------------------------------------------------------------------
function HoJ_OnEvent( self, event, ... )
	--HoJ:DebugMsg( "event = " .. event );

	-- do nothing for non-paladins
	if ( HoJ._isPaladin == false ) then
		return;
	end
	-- call event functions
	HoJ._events[ event ](self, ... );
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetFrames
-- -----------------------------------------------------------------------------

function HoJ:ResetFrames( frameList )
	-- clear all icon frames
	for key, value in pairs( frameList ) do
		value.frame:Hide()
--[[
		if ( value.texture ~= nil ) then
			value.texture:SetTexture( nil )
			value.texture:Hide()
		end
		if ( value.cooldown ~= nil ) then
			value.cooldown:Hide()
		end
		if ( value.charges ~= nil ) then
			value.charges:Hide()
		end
		value.spellInfo = nil
		--value.text:Hide();
]]
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Reset
-- -----------------------------------------------------------------------------
function HoJ:Reset()
	-- HoJ:DebugMsg( "Reset" );

	HoJ._level = UnitLevel( "player" ) -- handle party sync

	HoJ_InitSpellInfo()

	HoJ:ResetFrames( HoJ._buffFrames )
	HoJ:ResetFrames( HoJ._cooldownFrames )
	HoJ:ResetFrames( HoJ._rotationFrames )

	HoJ:ClearIconList( HoJ._buffs )
	HoJ:ClearIconList( HoJ._cooldowns )
	HoJ:ClearIconList( HoJ._rotationSkills )

	HoJ._glowAlpha = 0.0
	HoJ:SetHPAlphas( HoJ._glowAlpha, HoJ._holyPower )

	HoJ:ResetHealth()

	HoJ._needsReset = false
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBossUnit
-- returns:
-- - (bool) true if enemy boss
-- - (bool) true if friendly boss
-- -----------------------------------------------------------------------------
function HoJ:IsBossUnit( unit, targetClassification )
	if ( targetClassification == "worldboss" ) then
		return true, false
	end

	local isFriendlyBoss = UnitIsFriend( "player", unit ) and UnitLevel( unit ) == -1
	local isInstanceBoss

	for i = 1, 5 do
	    local bossGUID = UnitGUID( "boss" .. i )

	    if UnitGUID(unit) == bossGUID then
	        isInstanceBoss = bossGUID
	        break
	    end
	end
	return isInstanceBoss, isFriendlyBoss
end

-- -----------------------------------------------------------------------------
-- HoJ:Update
-- -----------------------------------------------------------------------------
function HoJ:Update()
	self._updateFrameNum = self._updateFrameNum + 1

	if ( self._level < 10 ) then
		return;
	end

	--HoJ:DebugMsg( "HoJ_Update" );
	-- update holy power amount
	local lastHolyPower = HoJ._holyPower;
	HoJ._holyPower = UnitPower( "player", 9 );

	-- we don't display our own cooldown text at some points when OmniCC is loaded
	if ( HoJ._omniCC == false ) then
		if ( IsAddOnLoaded( "OmniCC" ) ) then
			HoJ._omniCC = true;
			--HoJ:DebugMsg( "OmniCC is loaded. HammerOfJustas will hide OmniCC's inquisition timer." );
		else
			HoJ._omniCC = false;
		end
	end

	if ( HoJ._enabledForRole ) then
		ProfileFunction( "UpdateHealth", HoJ.UpdateHealth, self )
	end

	local updateTargetsStartMs = ProfileBegin2( "UpdateTargets" )

	local target = "target";

	HoJ._targetCanAttack = false;
	HoJ._targetCanBeInterrupted = false;
	HoJ._targetHealthRatio = 0.0;
	HoJ._targetHealth = 0;
	HoJ._targetLevel = 1;
	HoJ._targetClassification = "";
	HoJ._targetIsTough = false;

	HoJ._targetInRange[HoJ.RANGE_NONE] = true;
	HoJ._targetInRange[HoJ.RANGE_SELF] = true;
	HoJ._targetInRange[HoJ.RANGE_MELEE] = false;
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_20_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] = false;

	HoJ._targetExists = UnitExists( target );

	-- cache target stuff if we have a valid target
	if ( HoJ._targetExists and HoJ._enabledForRole ) then
		HoJ._targetReaction = UnitReaction( target, "player" );
		if ( HoJ._targetReaction ~= nil
			 and not UnitIsDead( target )
			 and UnitExists( target )
			 and HoJ._targetReaction < HoJ.REACTION_THRESHOLD ) then	-- 5 if you want to show the skill bar for yellow mobs
			HoJ._targetCanAttack = true;
		else
			HoJ._targetCanAttack = false;
		end

		if ( HoJ._targetCanAttack ) then
			local channelName, _, _, _, _, _, channelNotInterruptible = UnitChannelInfo( target )
			local castName, _, _, _, _, _, _, castNotInterruptible = UnitCastingInfo( target )
			if ( ( channelName ~= nil and not channelNotInterruptible )
				or ( castName ~= nil and not castNotInterruptible ) ) then
				HoJ._targetCanBeInterrupted = true;
			end
			HoJ._targetHealthRatio = HoJ:GetHealthRatio( target );
		end

		HoJ._targetHealth = UnitHealth( target )
		HoJ._targetLevel = UnitLevel( target )
		HoJ._targetClassification = UnitClassification( target )

		-- target range finding
		HoJ._targetInRange[HoJ.RANGE_MELEE] = HoJ:InRange( HoJ.SpellInfo.rebuke, target, "Rebuke" );

		HoJ._targetInRange[HoJ.RANGE_10_YARDS] = HoJ:InRange( HoJ.SpellInfo.hammerOfJustice, target, "Hammer of Justice" );

		if ( HoJ._role == HoJ.ROLE_DPS ) then
			HoJ._targetInRange[HoJ.RANGE_12_YARDS] = HoJ:InRange( HoJ.SpellInfo.bladeOfJustice, target, "Blade of Justice" )
		else
			-- only ret has a 12 yard spell to check with
			HoJ._targetInRange[HoJ.RANGE_12_YARDS] = HoJ._targetInRange[HoJ.RANGE_10_YARS]
		end

		if ( HoJ:IsTalentSelected( HoJ.SpellInfo.executionSentence ) ) then
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ:InRange( HoJ.SpellInfo.executionSentence, target, "Execution Sentence" )
		elseif ( HoJ._role == HoJ.ROLE_DPS ) then
			-- we have no other offensive 20 yards spells to test, so we use execution sentence as a stand-in
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ:InRange( HoJ.SpellInfo.bladeOfJustice, target, "Blade of Justice" )
		else
			-- prot has no 20 yard offensive spells
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ._targetInRange[HoJ.RANGE_10_YARDS]
		end

		HoJ._targetInRange[HoJ.RANGE_30_YARDS] = HoJ:InRange( HoJ.SpellInfo.judgment, target, "Judgment" );

		-- note LoH or any non-aggressive spell will simply fail for non-enemy targets, so we just
		-- set it to 30, since if you're within 30 you're within 40.
		if ( HoJ._targetReaction == nil or HoJ._targetReaction >= HoJ.REACTION_THRESHOLD ) then
			HoJ._targetInRange[HoJ.RANGE_40_YARDS] = HoJ._targetInRange[HoJ.RANGE_30_YARDS];
		else
			HoJ._targetInRange[HoJ.RANGE_40_YARDS] = HoJ:InRange( HoJ.SpellInfo.layOnHands, target, "Lay on Hands" );
		end
		--HoJ:DebugMsg( "40 yards = " .. HoJ:BoolToStr( HoJ._targetInRange[HoJ.RANGE_40_YARDS] ) );

		-- target toughtness rating
		HoJ._targetIsBoss = self:IsBossUnit( target, HoJ._targetClassification )
		HoJ._targetIsTough = HoJ._targetIsBoss
							 or HoJ._targetClassification == "rareelite"
							 or HoJ._targetClassification == "elite"
							 or HoJ._targetLevel > HoJ._level + 3
							 or HoJ._targetHealth / HoJ._playerHealth > 0.8;
	end

	ProfileEnd( "UpdateTargets", updateTargetsStartMs )

	local updateCooldowns = HoJ._updateType == HoJ.UPDATE_COOLDOWNS
	local updateBuffs = HoJ._updateType == HoJ.UPDATE_BUFFS
	local updateRotation = HoJ._updateType == HoJ.UPDATE_ROTATION

	-- rotate through the update types
	HoJ._updateType = HoJ._updateType + 1
	if ( HoJ._updateType > HoJ.UPDATE_MAX ) then
		HoJ._updateType = HoJ.UPDATE_START
	end

	-- sacred shield updates even if not in combat because we want to put it on
	-- before entering combat if we're targetting a hostile.
	if ( HoJ._role == HoJ.ROLE_TANK ) then
		if ( HoJ._enabledForRole ) then
			-- HoJ:DebugMsg( "Updating ROLE_TANK" );

			if ( updateCooldowns or updateRotation ) then
				HoJ:FindActiveBuffs( "player", HoJ._protActiveBuffList )
			end

			-- update cooldowns
			if ( updateCooldowns ) then
				local startCooldownsMs = ProfileBegin2( "Cooldowns" )

				HoJ:ClearIconList( HoJ._cooldowns );

				HoJ:CheckMitigationSkills();
				HoJ:CheckInterrupts( HoJ._cooldowns );

				HoJ:UpdateIcons( HoJ._cooldowns, HoJ._cooldownFrames, false );

				ProfileEnd( "Cooldowns", startCooldownsMs )
			end

			-- update rotation
			if ( updateRotation ) then
				local startRotationMs = ProfileBegin2( "Rotation" )

				HoJ:ClearIconList( HoJ._rotationSkills );

				-- HoJ:CheckConsecration();
				HoJ:CheckTankRotation_8_dot_1( HoJ._rotationSkills );

				HoJ:UpdateIcons( HoJ._rotationSkills, HoJ._rotationFrames, false );

				ProfileEnd( "Rotation", startRotationMs )
			end
		end
	elseif ( HoJ._role == HoJ.ROLE_DPS ) then
		-- HoJ:DebugMsg( "Updating ROLE_DPS" );

		if ( updateCooldowns or updateRotation or updateBuffs ) then
			HoJ:FindActiveBuffs( "player", HoJ._retActiveBuffList )
			-- don't treat Divine Purpose as normal Holy Power because we can hold it to
			-- wait for a dead spot in the rotation
--[[
			if ( HoJ:IsBuffActive( HoJ.SpellInfo.divinePurposeBuff.id ) ) then
				HoJ._fakeHolyPower = 5
			else
				HoJ._fakeHolyPower = HoJ._holyPower
			end
]]
			HoJ._fakeHolyPower = HoJ._holyPower
		end

		-- update buffs
		if ( updateBuffs ) then
			local startBuffsMs = ProfileBegin2( "Buffs" )

			HoJ:ClearIconList( HoJ._buffs );

			HoJ:CheckProcdBuffs();
			HoJ:CheckBlessings();
			-- HoJ:CheckSeals();
			ProfileFunction( "UpdateIcons:Buffs:", HoJ.UpdateIcons, self, HoJ._buffs, HoJ._buffFrames, true )

			ProfileEnd( "Buffs", startBuffsMs )
		end

		-- update cooldowns
		if ( updateCooldowns ) then
			local startCooldownsMs = ProfileBegin2( "Cooldowns" )

			ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )
			ProfileFunction( "CheckSelfHealing", HoJ.CheckSelfHealing, self )
			ProfileFunction( "CheckShieldOfVengeance", HoJ.CheckShieldOfVengeance, self )
			ProfileFunction( "CheckInquisition", HoJ.CheckInquisition, self )
			ProfileFunction( "CheckAvengingWrath", HoJ.CheckAvengingWrath, self )
			ProfileFunction( "CheckInterrupts", HoJ.CheckInterrupts, self, HoJ._cooldowns )
			ProfileFunction( "UpdateIcons:Cooldowns:", HoJ.UpdateIcons, self, HoJ._cooldowns, HoJ._cooldownFrames, true )

			ProfileEnd( "Cooldowns", startCooldownsMs )
		end

		-- update rotation
		if ( updateRotation ) then
			local startRotationMs = ProfileBegin2( "Rotation" )

			HoJ:ClearIconList( HoJ._rotationSkills );

			-- This is not working and may be causing issues.
			HoJ:CheckDPSRotation_Ret_8dot1( HoJ._rotationSkills );

			ProfileFunction( "UpdateIcons:Rotation:", HoJ.UpdateIcons, self, HoJ._rotationSkills, HoJ._rotationFrames, false );

			ProfileEnd( "Rotation", startRotationMs )
		end
	elseif ( HoJ._role == HoJ.ROLE_HEALING ) then
		if ( HoJ._enabledForRole ) then
			--HoJ:DebugMsg( "Updating ROLE_HEALING" );
		end
	else
		--HoJ:DebugMsg( "WE HAVE NO IDEA WHAT OUR ROLE IS!" );
	end

	if ( not HoJ._enabledForRole ) then
		HoJ:ResetFrames( HoJ._buffFrames );
		HoJ:ResetFrames( HoJ._cooldownFrames );
		HoJ:ResetFrames( HoJ._rotationFrames );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateGlow
-- -----------------------------------------------------------------------------
function HoJ:UpdateGlow( frame, elapsed )
	if ( self._glowAlpha > self._maxGlowAlpha ) then
		self._glowScale = 1.0 - ( 4.0 * elapsed )
	elseif ( self._glowAlpha < self._minGlowAlpha ) then
		self._glowScale = 1.0 + ( 4.0 * elapsed )
	end

	self._glowAlpha = self._glowAlpha * self._glowScale
	if ( self._glowAlpha <= 0.0 ) then
		self._glowAlpha = self._minGlowAlpha
	elseif ( self._glowAlpha > 1.0 ) then
		self._glowAlpha = self._maxGlowAlpha
	end
	--self:DebugMsg( "_glowAlpha = " .. self._glowAlpha )
	--self:DebugMsg( "_glowScale = " .. self._glowScale )

	-- update holy power icons
	if ( lastHolyPower ~= 3 and self._holyPower == 3 ) then
		--PlaySound( "LEVELUP" )
		--PlaySound( "igPVPUpdate" )
		-- BFAFIX
		-- self:DebugMsg( "Playing detected sound" )
		-- PlaySoundFile( "Interface\\AddOns\\HammerOfJustas\\Sounds\\detected-nearby.wav" )
	end

	finalAlpha = self._glowAlpha
	if ( IsMounted() ) then
		finalAlpha = finalAlpha * 0.1875
	end

	self:SetHPAlphas( finalAlpha, self._holyPower )

	HoJ_RotationFrame:SetAlpha( 0.666 )
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateGlobalCooldown
-- -----------------------------------------------------------------------------
function HoJ:UpdateGlobalCooldown()
	local start, duration, enabled = GetSpellCooldown( HoJ.SpellInfo.globalCooldown.id );
	if ( start > 0 and duration > 0 ) then
		HoJ._gcdDuration = duration;
		HoJ._gcdRemaining = ( start + duration ) - HoJ._curTime;
	else
		HoJ._gcdRemaining = 0;
	end

	--HoJ:DebugMsg( "Global Cooldown: duration = " .. HoJ._gcdDuration .. ", remaining = " .. HoJ._gcdRemaining );
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRender
-- -----------------------------------------------------------------------------
function HoJ:UpdateRender( frame, elapsed )
	self._renderFrameNum = self._renderFrameNum + 1

	if ( self._enabledForRole ) then
		HoJ:UpdateGlow( frame, elapsed )
	else
		HoJ_RotationFrame:SetAlpha( 0.0 )
	end

	-- inquisition stuff
	self:UpdateInquisitionFrames( frame )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnUpdate
-- -----------------------------------------------------------------------------
function HoJ_OnUpdate( frame, elapsed )
	ProfileFrameNum = ProfileFrameNum + 1
	HoJ._frameNum = HoJ._frameNum + 1;

	-- this prevents weird stuff from happening between the time when the player
	-- leaves the world and returns (for instance, taking a portal)
	if ( HoJ._worldState ~= HoJ.WORLD_STATE_ENTERED ) then
		return
	end

	if ( not HoJ._enabled ) then
		if ( HoJ._needsReset ) then
			HoJ:Reset();
		end
		return;
	end

	if ( HoJ._enableProfiling ) then
		ResetCPUUsage();
	end

	if ( HoJ._playerRaceId == 0 ) then
		_, _, HoJ._playerRaceId = UnitRace( "player" )
	end

	local lastRole = HoJ._role;
	HoJ:UpdateRole();
	if ( HoJ._role ~= lastRole ) then
		HoJ:Print( "HammerOfJustas: role changed to " .. HoJ.ROLE_NAMES[ HoJ._role ] .. "." )
		HoJ._needsReset = true
	end

	local level = UnitLevel( "player" )
	if ( HoJ._level ~= level ) then
		HoJ:Print( "HammerOfJustas: level changed from " .. HoJ._level .. " to " .. level .. "." )
		HoJ._level = level
		HoJ._needsReset = true
	end

	if ( HoJ._needsReset ) then
		HoJ:Reset();
	end

	if ( HoJ._role == HoJ.ROLE_HEALING ) then
		if ( HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: disabled for Holy." );
		end
		HoJ._enabledForRole = false;
	elseif ( HoJ._role == HoJ.ROLE_TANK ) then
		if ( not HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: enabled for Protection." );
		end
		HoJ._enabledForRole = true;
	else
		if ( not HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: enabled for Retribution." );
		end
		HoJ._enabledForRole = true;
	end

	--HoJ:DebugMsg( "HoJ_OnUpdate" );
	-- do nothing for non-paladins
	if ( HoJ._isPaladin == false ) then
		return;
	end

	-- update time
	HoJ._curTime = GetTime();
	HoJ:UpdateGlobalCooldown();
	HoJ:MouseOver();

	-- update general logic
	if ( HoJ._curTime >= HoJ._nextUpdateTime ) then
		HoJ._nextUpdateTime = HoJ._curTime + HoJ.UPDATE_INTERVAL

		ProfileFunction( "HoJ:Update", HoJ.Update, HoJ )
	end

	-- update render logic
	if ( HoJ._curTime >= HoJ._nextRenderTime ) then
		HoJ._nextRenderTime = HoJ._curTime + HoJ.RENDER_INTERVAL;

		ProfileFunction( "HoJ:UpdateRender", HoJ.UpdateRender, HoJ, frame, elapsed )
	end

	if ( HoJ._enableProfiling ) then
		UpdateAddOnCPUUsage();
		local addonMs = GetAddOnCPUUsage( "HammerOfJustas" )
		local updateMs = GetFunctionCPUUsage( HoJ.Update, true )
		local renderMs = GetFunctionCPUUsage( HoJ.UpdateRender, true )
		local eventMs = GetEventCPUUsage()
		--if ( addonMs ~= nil and addonMs > HoJ._profileThreshold ) then
			local metThreshold = HoJ._profileFunction == nil
			for key, value in pairs( ProfileTable ) do
				if ( value ~= nil and value.name ~= nil and value.timeMs >= HoJ._profileThreshold and value.frameNum == ProfileFrameNum and
					( HoJ._profileFunction == nil or ( string.find( value.name, HoJ._profileFunction ) ~= nil ) ) ) then
					HoJ:Msg( "func " .. value.name .. "( " .. value.count .. " ): " .. value.timeMs .. " ms" )
					metThreshold = true
				end
			end
			if ( metThreshold ) then
				HoJ:Msg( "Addon CPU: " .. addonMs .. "ms" )
				HoJ:Msg( "Update: " .. updateMs .. "ms" )
				HoJ:Msg( "Render: " .. renderMs .. "ms" )
			end
		--end
		if ( eventMs ~= nil and eventMs > HoJ._profileThreshold ) then
			HoJ:Msg( "Events: " .. eventMs .. "ms" )
		end
	end
end

function HoJ:FindChild( self, substr )
	--HoJ:DebugMsg( "FindChild( " .. self:GetName() .. " )" );
	local children = self:GetChildren();
	for i, child in ipairs( children ) do
		HoJ:DebugMsg( "Child " .. i .. " is " .. child.GetName() );
		if ( string.find( child:GetName(), substr ) ~= nil ) then
			return child;
		end
	end
	return nil;
end

function HoJ:FindRegion( self, substr )
	--HoJ:DebugMsg( "FindRegion( " .. self:GetName() .. " )" );
	local regions = { self:GetRegions() }
	for i, region in ipairs( regions ) do
		-- HoJ:DebugMsg( "Region " .. i .. " is " .. region.GetName() );
		-- if ( string.find( region:GetName(), substr ) ~= nil ) then
		if ( region and region:GetObjectType() == "FontString" ) then
			return region;
		end
	end
	return nil;
end

function HoJ:SetTitleTextColor( self, colorR, colorG, colorB, colorA )
	--HoJ:DebugMsg( "SetTitleTextColor( " .. self:GetName() .. " )" );
	local child = HoJ:FindChild( self, "TitleText" );
	if ( child ~= nil ) then
		child:SetTextColor( colorR, colorG, colorB, colorA );
	end
end

function HoJ:ShowTitle( self )
	local region = HoJ:FindRegion( self, "TitleText", self:GetRegions() );
	if ( region ~= nil ) then
		region:Show();
	end
end

function HoJ:HideTitle( self )
	local region = HoJ:FindRegion( self, "TitleText", self:GetRegions() );
	if ( region ~= nil ) then
		region:Hide();
	end
end

function HoJ:IsHoJFrame( self )
	for i, v in ipairs( HoJ._windowFrames ) do
		if ( v == self ) then
			return true;
		end
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ_OnEnter
-- -----------------------------------------------------------------------------
function HoJ_OnEnter( self )
	--HoJ:DebugMsg( "HoJ_OnEnter( " .. self:GetName() .. " )" );
	if ( not HoJ._enabledForRole ) then
		return;
	end

	if ( IsControlKeyDown() ) then
		-- if this isn't one of the parent windows, recurse until we get to one
		-- then, turn on background alpha for all the windows.
		if ( not HoJ:IsHoJFrame( self ) ) then
			HoJ_OnEnter( self:GetParent() );
		else
			for i, wf in ipairs( HoJ._windowFrames ) do
				wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.5 );
				--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.5 );
				HoJ:ShowTitle( wf );
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLeave
-- -----------------------------------------------------------------------------
function HoJ_OnLeave( self )
	--HoJ:DebugMsg( "HoJ_OnLeave( " .. self:GetName() .. " )" );
	if ( not HoJ._enabledForRole ) then
		return;
	end

	-- if this isn't one of the parent windows, recurse until we get to one
	-- then, turn on background alpha for all the windows.
	if ( not HoJ:IsHoJFrame( self ) ) then
		HoJ_OnLeave( self:GetParent() );
	else
		for i, wf in ipairs( HoJ._windowFrames ) do
			wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.0 );
			--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.0 );
			HoJ:HideTitle( wf );
		end
	end

	-- self:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStart
-- -----------------------------------------------------------------------------
function HoJ_OnDragStart( self, button )
	-- HoJ:DebugMsg( "HoJ_OnDragStart down for '" .. self:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStop
-- -----------------------------------------------------------------------------
function HoJ_OnDragStop( self )
	-- HoJ:DebugMsg( "HoJ_OnDragStop for '" .. self:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseDown
-- -----------------------------------------------------------------------------
function HoJ_OnMouseDown( self, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseDown for '" .. self:GetName() .. "'" )
	--if ( IsControlKeyDown() and not self.isLocked ) then
    -- 	self:StartMoving()
   	--end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseUp
-- -----------------------------------------------------------------------------
function HoJ_OnMouseUp( self, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseUp for '" .. self:GetName() .. "'" )
	--self:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ_Filter_Whisper
-- -----------------------------------------------------------------------------
function HoJ_Filter_Whisper( self, event, msg, player, _, _, _, gm )
	local blocked = false;
	if ( HoJ._blockedNames[ player ] ~= nil ) then
		blocked = true;
	end

	if ( blocked == true ) then
		-- HoJ:Print( type( HoJ._ignoreHistory[ player ] ) );
		if ( HoJ._ignoreHistory[ player ] == nil ) then
			HoJ._ignoreHistory[ player ] = { };
		end
		table.insert( HoJ._ignoreHistory[ player ], msg );
		if ( HoJ._silentIgnore == false ) then
			if ( HoJ._stealthIgnore == true ) then
				HoJ:Print( HoJ._stealthIgnoreMessage );
			else
				HoJ:Print( "Whisper from " .. player .. " blocked." );
			end
		end
		return true;
	end
	-- do default behavior
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnOK
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnOK()
	HoJ:WriteSavedVariables();
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnCancel
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnCancel()
	--HoJ:DebugMsg( "AddonOptions_OnCancel" );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnDefault
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnDefault()
	--HoJ:DebugMsg( "AddonOptions_OnDefault" );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnRefresh
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnRefresh()
	--HoJ:DebugMsg( "AddonOptions_OnRefresh" );
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowSpellOnBar
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowSpellOnBar( spell )
	spell.showOnBar = not spell.showOnBar
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowSkillSFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowSkillSFrame()
	if ( HoJ._showCooldownsFrame ) then
		--HoJ:DebugMsg( "_showCooldownsFrame = false" );
		HoJ._showCooldownsFrame = false;
	else
		--HoJ:DebugMsg( "_showCooldownsFrame = true" );
		HoJ._showCooldownsFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowBuffsFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowBuffsFrame()
	if ( HoJ._showBuffsFrame ) then
		--HoJ:DebugMsg( "_showBuffsFrame = false" );
		HoJ._showBuffsFrame = false;
	else
		--HoJ:DebugMsg( "_showBuffsFrame = true" );
		HoJ._showBuffsFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowRotationFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowRotationFrame()
	if ( HoJ._showRotationFrame ) then
		HoJ._showRotationFrame = false;
	else
		HoJ._showRotationFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:TogglehowTemplarsVerdictFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowTemplarsVerdictFrame()
	if ( HoJ._showTemplarsVerdictFrame ) then
		HoJ._showTemplarsVerdictFrame = false;
	else
		HoJ._showTemplarsVerdictFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowHolyPowerFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowHolyPowerFrame()
	if ( HoJ._showHolyPowerFrame ) then
		HoJ._showHolyPowerFrame = false;
	else
		HoJ._showHolyPowerFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowInquisitionFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowInquisitionFrame()
	if ( HoJ._showInquisitionFrame ) then
		HoJ._showInquisitionFrame = false;
	else
		HoJ._showInquisitionFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, checked, text )
	local cb = CreateFrame( "CheckButton", name, parentFrame, "OptionsCheckButtonTemplate" );
	cb:SetPoint( anchor1, anchorFrame, anchor2, xoffset, yoffset );
	--cb:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSpellOnBar( spellInfo ) end );
	cb:SetChecked( checked );
	-- the following is really pretty stupid and was a PITA to figure out
	getglobal( cb:GetName() .. 'Text' ):SetText( text );
	return cb;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupSpellCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, text )
	local cb = CreateFrame( "CheckButton", name, parentFrame, "OptionsCheckButtonTemplate" );
	cb:SetPoint( anchor1, anchorFrame, anchor2, xoffset, yoffset );
	cb:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSpellOnBar( spellInfo ) end );
	cb:SetChecked( spellInfo.showOnBar );
	-- the following is really pretty stupid and was a PITA to figure out
	if ( spellInfo.spellType == HoJ.SPELL_BUFF or spellInfo.spellType == HoJ.SPELL_TALENT_BUFF ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Buff)" );
	elseif ( spellInfo.spellType == HoJ.SPELL_SKILL ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Skill)" );
	elseif ( spellInfo.spellType == HoJ.SPELL_TALENT ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Talent)" );
	elseif ( spellInfo.spellType == HoJ.SPELL_SHAPESHIFT ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Seal)" );
	else
		getglobal( cb:GetName() .. 'Text' ):SetText( text );
	end
	return cb;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupShowSpellCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupShowSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, count )
	return HoJ:SetupSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, HOJ_SHOW .. spellInfo.name );
end

-- -----------------------------------------------------------------------------
-- HoJ:OnScrollValueChanged
-- -----------------------------------------------------------------------------
function HoJ:OnScrollValueChanged( slider, value )
	--HoJ:DebugMsg( "OnScrollValueChanged(): slider = " .. slider.name .. ", value = " .. value );
	slider.scrollView:SetVerticalScroll( value );
	slider.scrollView:UpdateScrollChildRect();
	local itemFrame = slider.scrollView:GetScrollChild();
	itemFrame:SetPoint( "TOPLEFT", slider.scrollView, "TOPLEFT", 0, value );
	itemFrame:SetPoint( "TOPRIGHT", slider.scrollView, "TOPRIGHT", 0, value );
end

-- -----------------------------------------------------------------------------
-- HoJ:scrollView_OnMouseWheel
-- -----------------------------------------------------------------------------
local function scrollView_OnMouseWheel( self, delta )
	local cur = self.slider:GetValue();
	local min, max = self.slider:GetMinMaxValues();
	local new = cur + ( delta * -1 * self.itemHeight );
	if ( new < min ) then
		new = min;
	elseif ( new > max ) then
		new = max;
	end
	self.slider:SetValue( new );
end

-- -----------------------------------------------------------------------------
-- Clamp
-- -----------------------------------------------------------------------------
 local function Clamp( value, min, max )
 	if ( value < min ) then
 		return min
 	elseif ( value > max ) then
 		return max
 	end

 	return value
 end

-- -----------------------------------------------------------------------------
-- HoJ:OnGlobalAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnGlobalAlphaChanged( alphaSlider, value )
	self._globalAlpha = Clamp( value / 10.0, 0.1, 1.0 )

	HoJFrame_Rotation:SetAlpha( self._globalAlpha * HoJ._rotationAlpha )
	HoJFrame_Cooldowns:SetAlpha( self._globalAlpha * HoJ._cooldownsAlpha )
	HoJFrame_Buffs:SetAlpha( self._globalAlpha * HoJ._buffsAlpha )
	-- holy power alpha is always set dynamically
	--HoJFrame_HoPo:SetAlpha( self._globalAlpha * HoJ._holyPowerAlpha )
	--HoJFrame:SetAlpha( self._globalAlpha * HoJ._holyPowerAlpha )
	HoJFrame_Inquisition:SetAlpha( self._globalAlpha * HoJ._inquisitionAlpha )

	--self:DebugMsg( "UI alpha = " .. self._globalAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:OnGlobalAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnRotationAlphaChanged( alphaSlider, value )
	self._rotationAlpha = Clamp( value / 10.0, 0.1, 1.0 )
	HoJFrame_Rotation:SetAlpha( self._globalAlpha * HoJ._rotationAlpha )
	--self:DebugMsg( "Rotation alpha = " .. self._rotationAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:OnCooldownsAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnCooldownsAlphaChanged( alphaSlider, value )
	self._cooldownsAlpha = Clamp( value / 10.0, 0.1, 1.0 )
	HoJFrame_Cooldowns:SetAlpha( self._globalAlpha * HoJ._cooldownsAlpha )
	--self:DebugMsg( "Cooldowns alpha = " .. self._cooldownsAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:OnBuffsAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnBuffsAlphaChanged( alphaSlider, value )
	self._buffsAlpha = Clamp( value / 10.0, 0.1, 1.0 )
	HoJFrame_Buffs:SetAlpha( self._globalAlpha * HoJ._buffsAlpha )
	--self:DebugMsg( "Buffs alpha = " .. self._buffsAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:OnHolyPowerAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnHolyPowerAlphaChanged( alphaSlider, value )
	self._holyPowerAlpha = Clamp( value / 10.0, 0.1, 1.0 )
	-- holy power alphas are updated dynamically each frame
	--HoJFrame_HoPo:SetAlpha( self._globalAlpha * HoJ._holyPowerAlpha )
	--HoJFrame:SetAlpha( self._globalAlpha * HoJ._holyPowerAlpha )
	--self:DebugMsg( "Holy Power alpha = " .. self._holyPowerAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:OnInquisitionAlphaChanged
-- -----------------------------------------------------------------------------
function HoJ:OnInquisitionAlphaChanged( alphaSlider, value )
	self._inquisitionAlpha = Clamp( value / 10.0, 0.1, 1.0 )
	HoJFrame_Inquisition:SetAlpha( self._globalAlpha * HoJ._inquisitionAlpha )
	--self:DebugMsg( "value = " .. value .. ", Inquisition alpha = " .. self._inquisitionAlpha )
end

-- -----------------------------------------------------------------------------
-- HoJ:AddSlider
-- -----------------------------------------------------------------------------
function HoJ:AddSlider( name, title, parent, relativeFrame, relativeTo, ofsx, ofsy, width, height, min, max, minText, maxText, value, callbackFn )
	local sl = CreateFrame( "Slider", name, parent, "OptionsSliderTemplate")
	--sl.name = "UI Alpha"
	sl:SetOrientation( "HORIZONTAL" )
	sl:SetHitRectInsets( 0, 0, -10, 0 )
	sl:SetBackdrop( sliderBackdrop )
	sl:SetWidth( SCROLLBAR_WIDTH )
	sl:SetThumbTexture( "Interface\\Buttons\\UI-SliderBar-Button-Horizontal" )
	--sl.scrollView = scrollView;
	sl.slider = sl
	sl:SetPoint( "TOPLEFT", relativeFrame, relativeTo, ofsx, ofsy )
	sl:SetWidth( width )
	sl:SetHeight( height )
	sl:SetScript( "OnValueChanged", callbackFn )
	sl:SetMinMaxValues( min, max )
	sl:SetValueStep( 1 )
	sl:SetValue( value )
	getglobal( sl:GetName() .. "Low" ):SetText( minText )
 	getglobal( sl:GetName() .. "High" ):SetText( maxText )
 	getglobal( sl:GetName() .. "Text" ):SetText( title )

 	return sl
 end

-- -----------------------------------------------------------------------------
-- HoJ:SetupAddOnOptions
-- -----------------------------------------------------------------------------
function HoJ:SetupAddOnOptions()
	--HoJ:DebugMsg( "HoJ:SetupAddOnOptions" );
	local panel = CreateFrame( "FRAME", "HoJOptions" );
	--panel.name = "Hammer of Justas";
	panel.name = GetAddOnMetadata( "HammerOfJustas", "Title" );
	panel.parent = nil;
	panel.okay = function ( self ) HoJ:AddonOptions_OnOK( self ) end;
	panel.cancel = function ( self ) HoJ:AddonOptions_OnCancel( self ) end;
	panel.defalt = function ( self ) HoJ:AddonOptions_OnDefault( self ) end;
	panel.refresh = function ( self ) HoJ:AddonOptions_OnRefresh( self ) end;
	InterfaceOptions_AddCategory( panel );

	local panelHeader = panel:CreateFontString( nil, "ARTWORK" );
	panelHeader:SetFont( "Fonts\\FRIZQT__.TTF", 24 );
	panelHeader:SetJustifyH( "LEFT" );
	panelHeader:SetJustifyV( "TOP" );
	panelHeader:ClearAllPoints();
	panelHeader:SetPoint( "TOPLEFT", 16, -16 );
	panelHeader:SetText( "Hammer of Justas v. " .. HoJ:GetVersionString() );

	-- TODO: make this check list a class
	local checkListHeader = panel:CreateFontString( nil, "ARTWORK" );
	checkListHeader:SetFont( "Fonts\\FRIZQT__.TTF", 18 );
	checkListHeader:SetJustifyH( "LEFT" );
	checkListHeader:SetJustifyV( "TOP" );
	checkListHeader:ClearAllPoints();
	checkListHeader:SetTextColor( 1, 1, 1, 1 );
	checkListHeader:SetPoint( "TOPLEFT", panelHeader, "BOTTOMLEFT", 0, -16 );
	checkListHeader:SetText( "Skill Selection" );

	local checkListDescription = panel:CreateFontString( nil, "ARTWORK" );
	checkListDescription:SetFont( "Fonts\\FRIZQT__.TTF", 11 );
	checkListDescription:SetJustifyH( "LEFT" );
	checkListDescription:SetJustifyV( "TOP" );
	checkListDescription:ClearAllPoints();
	checkListDescription:SetTextColor( 1, 1, 1, 1 );
	checkListDescription:SetPoint( "TOPLEFT", panelHeader, "BOTTOMLEFT", 0, -40 );
	checkListDescription:SetText( 	"Display of individual cooldowns, talents and buffs can be toggled in the rotation, buffs and cooldowns frames.\n" ..
									"Note that turning off a skill in the rotation frame will likely result in an 'empty' rotation action\n" ..
									"when that skill is chosen for the rotation." );

	local SCROLL_BORDER_HEIGHT = 24 * 8 + EDGE_SIZE * 2;
	local scrollBorder = CreateFrame( "FRAME", "scrollBorder", panel );
	scrollBorder:SetPoint( "TOPLEFT", checkListDescription, "BOTTOMLEFT", 0, 0 );
	scrollBorder:SetPoint( "TOPRIGHT", panel, "TOPRIGHT", -16, 0 );
	scrollBorder:SetPoint( "BOTTOMLEFT", checkListDescription, "BOTTOMLEFT", 0, -SCROLL_BORDER_HEIGHT );
	scrollBorder:SetBackdrop( listBackdrop );

	local scrollView = CreateFrame( "SCROLLFRAME", "scrollView", scrollBorder );
	scrollView:EnableMouseWheel( true );
	scrollView:SetScript( "OnMouseWheel", scrollView_OnMouseWheel );
	scrollView:SetPoint( "TOPLEFT", scrollBorder, "TOPLEFT", EDGE_SIZE, -EDGE_SIZE );
	scrollView:SetPoint( "TOPRIGHT", scrollBorder, "TOPRIGHT", -EDGE_SIZE, -EDGE_SIZE );
	scrollView:SetPoint( "BOTTOMLEFT", scrollBorder, "BOTTOMLEFT", EDGE_SIZE,
	EDGE_SIZE );
	scrollView:SetPoint( "BOTTOMRIGHT", scrollBorder, "BOTTOMRIGHT", -EDGE_SIZE, EDGE_SIZE );

	local checkFrame = CreateFrame( "FRAME", "itemFrame", scrollView );
	scrollView:SetScrollChild( checkFrame );

	checkFrame:SetPoint( "TOPLEFT", scrollView, "TOPLEFT", 0, 0 );
	--checkFrame:SetPoint( "TOPRIGHT", scrollView, "TOPRIGHT", -EDGE_SIZE, -EDGE_SIZE * 2 );
	--checkFrame:SetPoint( "BOTTOMLEFT", scrollView, "BOTTOMLEFT", EDGE_SIZE, EDGE_SIZE );
	--checkFrame:SetPoint( "BOTTOMRIGHT", scrollView, "BOTTOMRIGHT", -EDGE_SIZE, EDGE_SIZE );
	checkFrame:Lower();

	local YPAD = 2;

	local prev = checkFrame;
	local anchor = "TOPLEFT";
	local count = 0;
	local spellIndex = 0;
	for key, value in pairs( HoJ.SpellInfo ) do
		if ( value.showCheckbox ) then
			--HoJ:DebugMsg( key .. " is " .. value.name );
			if ( key == nil ) then
				HoJ:DebugMsg( "got nil name for spellInfo id " .. value.id );
			elseif ( value == nil ) then
				HoJ:DebugMsg( "got nil value for spellInfo id " .. value.id );
			elseif ( value.name == nil ) then
				HoJ:DebugMsg( "got nil value.name for spellInfo id " .. value.id );
			else
				prev = HoJ:SetupShowSpellCheckBox( checkFrame, prev, key, "TOPLEFT", anchor, 0, YPAD, value, count );
				if ( count == 0 ) then
					anchor = "BOTTOMLEFT";
				end
			end
			count = count + 1;
		end
		spellIndex = spellIndex + 1;
	end

	-- There must be a better way to calculate the distance from the top of the first item to the bottom of the last item
	scrollView.itemHeight = prev:GetHeight() - YPAD;
	local itemsInScrollView = 4;
	local totalHeight = ( scrollView.itemHeight ) * ( count - itemsInScrollView );
	checkFrame:SetSize( 400, totalHeight );
	scrollView:SetSize( 400, totalHeight );

	local slider = CreateFrame( "SLIDER", "slider", scrollView );
	slider.name = "slider";
	slider:SetOrientation( "VERTICAL" );
	slider:SetHitRectInsets( 0, 0, -10, 0 );
	slider:SetBackdrop( sliderBackdrop );
	slider:SetWidth( SCROLLBAR_WIDTH );
	slider:SetThumbTexture( "Interface\\Buttons\\UI-SliderBar-Button-Vertical" );
	slider.scrollView = scrollView;
	slider.slider = slider;
	slider:SetPoint( "TOPLEFT", scrollView, "TOPRIGHT", -SCROLLBAR_WIDTH, 0 );
	slider:SetPoint( "BOTTOMLEFT", scrollView, "BOTTOMRIGHT", -SCROLLBAR_WIDTH, 0 );
	slider:SetScript( "OnValueChanged", function ( slider, value ) HoJ:OnScrollValueChanged( slider, value ) end );
	slider:SetMinMaxValues( 0, totalHeight );
	slider:SetValueStep( 1 );
	slider:SetValue( 0 );
	scrollView.slider = slider;


	-- TODO: make this check list a class
	local frameToggleHeader = panel:CreateFontString( nil, "ARTWORK" );
	frameToggleHeader:SetFont( "Fonts\\FRIZQT__.TTF", 18 );
	frameToggleHeader:SetJustifyH( "LEFT" );
	frameToggleHeader:SetJustifyV( "TOP" );
	frameToggleHeader:ClearAllPoints();
	frameToggleHeader:SetTextColor( 1, 1, 1, 1 );
	frameToggleHeader:SetPoint( "TOPLEFT", scrollBorder, "BOTTOMLEFT", 0, -16 );
	frameToggleHeader:SetText( "Frame Toggles" );

	-- add check box for skills bar
	local checkShowCooldownsFrame = HoJ:SetupCheckBox( panel, frameToggleHeader, "checkCooldowns", "TOPLEFT", "BOTTOMLEFT", 0, 0, HoJ._showCooldownsFrame, "Show Cooldowns" );
	checkShowCooldownsFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSkillSFrame() end );
	checkShowCooldownsFrame:Show();

	-- add check box for buffs bar
	local checkShowBuffsFrame = HoJ:SetupCheckBox( panel, checkShowCooldownsFrame, "checkBuffs", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showBuffsFrame, "Show Buffs" );
	checkShowBuffsFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowBuffsFrame() end );
	checkShowBuffsFrame:Show();

	local checkShowRotationFrame = HoJ:SetupCheckBox( panel, checkShowBuffsFrame, "checkRotation", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showRotationFrame, "Show Rotation" );
	checkShowRotationFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowRotationFrame() end );
	checkShowRotationFrame:Show();

	local checkShowTemplarsVerdictFrame = HoJ:SetupCheckBox( panel, checkShowCooldownsFrame, "checkShowTemplarsVerdict", "TOPLEFT", "BOTTOMLEFT", 0, 0, HoJ._showTemplarsVerdictFrame, "Show Templars Verdict" );
	checkShowTemplarsVerdictFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowTemplarsVerdictFrame() end );
	checkShowTemplarsVerdictFrame:Show();

	local checkShowHolyPowerFrame = HoJ:SetupCheckBox( panel, checkShowTemplarsVerdictFrame, "checkShowHolyPower", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showHolyPowerFrame, "Show Holy Power" );
	checkShowHolyPowerFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowHolyPowerFrame() end );
	checkShowHolyPowerFrame:Show();

	local checkShowInquisitionFrame = HoJ:SetupCheckBox( panel, checkShowHolyPowerFrame, "checkShowInquisition", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showHolyPowerFrame, "Show Inquisition" );
	checkShowInquisitionFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowInquisitionFrame() end );
	checkShowInquisitionFrame:Show();

	local alphaSlider = self:AddSlider( "uiAlphaSlider", "Global Alpha", panel, checkShowTemplarsVerdictFrame,
										"BOTTOMLEFT", 0, -16, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._globalAlpha * 10, 1, 10 ),
										function ( alphaSlider, value ) HoJ:OnGlobalAlphaChanged( alphaSlider, value ) end )

	local buffsSlider = self:AddSlider( "uiBuffsSlider", "Buffs Frame Alpha", panel, alphaSlider,
										"TOPLEFT", 200, 0, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._buffsAlpha * 10, 1, 10 ),
										function ( buffsSlider, value ) HoJ:OnBuffsAlphaChanged( buffsSlider, value ) end )

	local rotationSlider = self:AddSlider( "uiRotationSlider", "Rotation Frame Alpha", panel, buffsSlider,
										"TOPLEFT", 200, 0, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._rotationAlpha * 10, 1, 10 ),
										function ( rotationSlider, value ) HoJ:OnRotationAlphaChanged( rotationSlider, value ) end )

	local cooldownsSlider = self:AddSlider( "uiCooldownsSlider", "Cooldowns Frame Alpha", panel, alphaSlider,
										"BOTTOMLEFT", 0, -32, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._cooldownsAlpha * 10, 1, 10 ),
										function ( cooldownsSlider, value ) HoJ:OnCooldownsAlphaChanged( cooldownsSlider, value ) end )

	local hopoSlider = self:AddSlider( "uiHoPoSlider", "Holy Power Frame Alpha", panel, cooldownsSlider,
										"TOPLEFT", 200, 0, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._holyPowerAlpha * 10, 1, 10 ),
										function ( hopoSlider, value ) HoJ:OnHolyPowerAlphaChanged( hopoSlider, value ) end )

	local inquisitionSlider = self:AddSlider( "uiInqSlider", "Inquisition Frame Alpha", panel, hopoSlider,
										"TOPLEFT", 200, 0, 150, 16, 1, 10, "0.1", "1.0", Clamp( self._inquisitionAlpha * 10, 1, 10 ),
										function ( inquisitionSlider, value ) HoJ:OnInquisitionAlphaChanged( inquisitionSlider, value ) end )

	scrollView:Show()
	checkFrame:Show()
	slider:Show()
	alphaSlider:Show()
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad
-- -----------------------------------------------------------------------------
function HoJ_OnLoad( self )
	--HoJ:DebugMsg( "HoJ_OnLoad : " .. self:GetName() );
	HoJ._curTime = GetTime();

	-- CircularBuffer:test()
	HoJ._healthBuffer = CircularBuffer:new( HoJ.HEALTH_SAMPLES_PER_SEC * HoJ.HEALTH_SAMPLE_SECONDS );

	local _, className = UnitClass( "player" );
	if ( className == "PALADIN" ) then
		HoJ._isPaladin = true;
	else
		HoJ._isPaladin = false;
	end

	HoJ:EnableFrame( self );

	local _, className = UnitClass( "player" );
	if ( HoJ._isPaladin ~= true ) then
		HoJ:Print( "The Hammer of Justas addon is intended for paladins only. The addon is disabled." );
		return;
	end

	HoJ._level = UnitLevel( "player" );
	HoJ._playerName = UnitName( "player" );

	-- variables loaded
	HoJFrame:RegisterEvent( "ADDON_LOADED" );

	HoJFrame:RegisterEvent( "PLAYER_LOGOUT" )

	-- holy power gained
	-- BFAFIX
	-- HoJFrame:RegisterEvent( "HOLY_POWER" );

	-- handle player entering the world
	HoJFrame:RegisterEvent( "PLAYER_ENTERING_WORLD" )

	-- handle player leaving the world
	HoJFrame:RegisterEvent( "PLAYER_LEAVING_WORLD" )

	-- check for changing into a bg or arena, etc.
	-- HoJFrame:RegisterEvent( "ZONE_CHANGED_NEW_AREA" );
	-- check for an attackable target
	HoJFrame:RegisterEvent( "PLAYER_TARGET_CHANGED" );

	-- check for entering combat
	--HoJFrame:RegisterEvent( "UNIT_COMBAT" );
	--HoJFrame:RegisterEvent( "PLAYER_ENTER_COMBAT" );
	HoJFrame:RegisterEvent( "PLAYER_REGEN_DISABLED" );

	-- check for leaving combat
	--HoJFrame:RegisterEvent( "PLAYER_LEAVE_COMBAT" );
	HoJFrame:RegisterEvent( "PLAYER_REGEN_ENABLED" );

	-- HoJFrame:RegisterEvent( "PLAYER_TALENT_UPDATE" );

	-- check for successful spellcasts
	HoJFrame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" );

	-- check for combat?
	-- HoJFrame:RegisterEvent( "UNIT_COMBAT" );
	-- ???
	-- HoJFrame:RegisterEvent( "VARIABLES_LOADED" );
	-- check auras
	HoJFrame:RegisterEvent( "UNIT_AURA" );
	-- check mana changes
	-- HoJFrame:RegisterEvent( "UNIT_MANA" );
	-- check health changes
	HoJFrame:RegisterEvent( "UNIT_HEALTH" );

	--HoJ:Print( HoJ.ADDON_NAME .. " version " .. HoJ:GetVersionString() .. " loaded." );
end

-- -----------------------------------------------------------------------------
-- HoJ_OnAddonLoad
-- -----------------------------------------------------------------------------
function HoJ_OnAddonLoaded( self )
	HoJ:Print( HoJ.ADDON_NAME .. " version " .. HoJ:GetVersionString() .. " loaded." );

	if ( HoJ._debugSpellName ~= nil and HoJ._debugSpellName ~= '' ) then
		HoJ:DebugMsg( "_debugSpellName = '" .. HoJ._debugSpellName .. "'" )
	end

	--HoJ:DebugMsg( "HoJ_OnAddonLoaded : " .. self:GetName() );

	HoJ:InitSavedVariables();

	-- hook the chat handler
	ChatFrame_AddMessageEventFilter( "CHAT_MSG_WHISPER", HoJ_Filter_Whisper );

	HoJ:UpdateRole();

	HoJ:SetupAddOnOptions();

	HoJFrame_Cooldowns:SetAlpha( HoJ._cooldownsAlpha * HoJ._globalAlpha )
	HoJFrame_Rotation:SetAlpha( HoJ._rotationAlpha * HoJ._globalAlpha )
	HoJFrame_Buffs:SetAlpha( HoJ._buffsAlpha * HoJ._globalAlpha )
	HoJFrame_Inquisition:SetAlpha( HoJ._inquisitionAlpha * HoJ._globalAlpha )
	-- holy power alphas are updated dynamically each frame
	--HoJFrame_HoPo:SetAlpha( HoJ._holyPowerAlpha * HoJ._globalAlpha )
	--HoJFrame:SetAlpha( HoJ._holyPowerAlpha * HoJ._globalAlpha )

--[[-- debug.. show spells in spellbook
	for i=1,1000 do
	   local spellName, spellSubName = GetSpellBookItemName(i, BOOKTYPE_SPELL)
	   if spellName == nil then
		  break
	   end
	   -- use spellName and spellSubName here
	   DEFAULT_CHAT_FRAME:AddMessage( spellName .. '(' .. spellSubName .. ')' )
	end
--]]

	HoJ:Print( "Welcome, " .. HoJ._playerName .. ". May the light favor your hammer." );
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_HoPo
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_HoPo( self )
	--HoJ:DebugMsg( "Set Up HoPo" );
	HoJ:EnableFrame( self );

	HoJ._HoPo[0] = self.HP1;
	HoJ._HoPoGlow[0] = self.HP1Glow;
	HoJ._HoPo[1] = self.HP2;
	HoJ._HoPoGlow[1] = self.HP2Glow;
	HoJ._HoPo[2] = self.HP3;
	HoJ._HoPoGlow[2] = self.HP3Glow;
	HoJ._HoPo[3] = self.HP4;
	HoJ._HoPoGlow[3] = self.HP4Glow;
	HoJ._HoPo[4] = self.HP5;
	HoJ._HoPoGlow[4] = self.HP5Glow;
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Buffs
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Buffs( self )
	--HoJ:DebugMsg( "Set Up Buffs" );
	HoJ:EnableFrame( self );

	HoJ._buffFrames = {
		{
			frame = HoJ_Buff1Frame,
			texture = HoJ_Buff1Texture,
			spellInfo = nil,
			text = HoJ_Buff1Text
		},
		{
			frame = HoJ_Buff2Frame,
			texture = HoJ_Buff2Texture,
			spellInfo = nil,
			text = HoJ_Buff2Text
		},
		{
			frame = HoJ_Buff3Frame,
			texture = HoJ_Buff3Texture,
			spellInfo = nil,
			text = HoJ_Buff3Text
		},
		{
			frame = HoJ_Buff4Frame,
			texture = HoJ_Buff4Texture,
			spellInfo = nil,
			text = HoJ_Buff4Text
		},
		{
			frame = HoJ_Buff5Frame,
			texture = HoJ_Buff5Texture,
			spellInfo = nil,
			text = HoJ_Buff5Text
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Cooldowns
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Cooldowns( self )
	--HoJ:DebugMsg( "Set Up Skills" );
	HoJ:EnableFrame( self );

	HoJ._cooldownFrames = {
		{
			frame = HoJ_Cooldown1Frame,
			texture = HoJ_Cooldown1Texture,
			spellInfo = nil,
			text = HoJ_Cooldown1Text
		},
		{
			frame = HoJ_Cooldown2Frame,
			texture = HoJ_Cooldown2Texture,
			spellInfo = nil,
			text = HoJ_Cooldown2Text
		},
		{
			frame = HoJ_Cooldown3Frame,
			texture = HoJ_Cooldown3Texture,
			spellInfo = nil,
			text = HoJ_Cooldown3Text
		},
		{
			frame = HoJ_Cooldown4Frame,
			texture = HoJ_Cooldown4Texture,
			spellInfo = nil,
			text = HoJ_Cooldown4Text
		},
		{
			frame = HoJ_Cooldown5Frame,
			texture = HoJ_Cooldown5Texture,
			spellInfo = nil,
			text = HoJ_Cooldown5Text
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Rotation
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Rotation( self )
	HoJ:EnableFrame( self );

	HoJ._rotationFrames = {
		{
			frame = HoJ_RotationFrame,
			texture = HoJ_RotationTexture,
			spellInfo = nil,
			text = HoJ_RotationText,
			cooldown = HoJ_RotationCooldown,
			charges = HoJ_RotationCharges
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Inquisition
-- this is pre-WoD stuff
-- FIXME: remove the call to this from window template
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Inquisition( self )
	-- HoJ:DebugMsg( "Set Up Inquisition" );
	HoJ:EnableFrame( self );

	HoJ_InqTexture:SetTexture( HoJ.SpellInfo.inquisition.icon );
	HoJ_InqTexture:Show();
	local fontName, fontHeight, fontFlags = HoJ_InqTimer:GetFont();
	HoJ_InqTimer:SetFont( fontName, fontHeight, "THICKOUTLINE" );
	HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );

	if ( HoJ._role ~= HoJ.ROLE_DPS ) then
		HoJ_InqFrameCooldown:SetAlpha( 0.0 );
		HoJ_InqTexture:SetAlpha( 0.0 );
		HoJ_InqTimer:SetAlpha( 0.0 );
	end
end
